Package: statgen
Type: Package
Title: Statistical Genetics Data Objects and Loaders
Version: 0.3.2
Authors@R: c(
    person("Oleksandr", "Frei", role = c("aut", "cph"),
        email = "oleksandr.frei@gmail.com"),
    person("Andrew", "Morris", role = c("cre", "ctb"),
        email = "a.h.morris@mn.uio.no")
    )
Description: Loads and manages statistical genetics data objects including reference
    panels, genotypes, LD matrices, annotations, and summary statistics. Follows
    the 'statgen' specification for use across 'Python', 'R', and 'MATLAB'/'Octave'
    runtimes.
License: MIT + file LICENSE
URL: https://github.com/precimed/statgen
BugReports: https://github.com/precimed/statgen/issues
Encoding: UTF-8
Depends: R (>= 4.1)
Imports: Matrix, jsonlite, digest, bit64, methods, data.table
Suggests: R.utils, testthat (>= 3.0.0), knitr, rmarkdown
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-20 18:59:34 UTC; andrewmorris
Author: Oleksandr Frei [aut, cph],
  Andrew Morris [cre, ctb]
Maintainer: Andrew Morris <a.h.morris@mn.uio.no>
Repository: CRAN
Date/Publication: 2026-05-27 20:10:17 UTC
Built: R 4.6.0; ; 2026-05-27 21:46:03 UTC; unix

S3method(a1,ReferencePanel)
S3method(a1,ReferenceShard)
S3method(a1_hash64,ReferencePanel)
S3method(a1_hash64,ReferenceShard)
S3method(a2,ReferencePanel)
S3method(a2,ReferenceShard)
S3method(a2_hash64,ReferencePanel)
S3method(a2_hash64,ReferenceShard)
S3method(a1freq,LDPanel)
S3method(a1freq,LDShard)
S3method(annomat,AnnotationPanel)
S3method(annomat,AnnotationShard)
S3method(annonames,AnnotationPanel)
S3method(beta_vec,Sumstats)
S3method(beta_vec,SumstatsShard)
S3method(bp,ReferencePanel)
S3method(bp,ReferenceShard)
S3method(chr,ReferencePanel)
S3method(chr,ReferenceShard)
S3method(eaf_vec,Sumstats)
S3method(eaf_vec,SumstatsShard)
S3method(father_id,GenotypePanel)
S3method(fetch_genotypes,GenotypePanel)
S3method(fetch_genotypes_int8,GenotypePanel)
S3method(fid,GenotypePanel)
S3method(default_chrX_sex,LDPanel)
S3method(info_vec,Sumstats)
S3method(info_vec,SumstatsShard)
S3method(iid,GenotypePanel)
S3method(is_female,GenotypePanel)
S3method(is_male,GenotypePanel)
S3method(is_object_compatible,ReferencePanel)
S3method(is_object_compatible,default)
S3method(is_present,GenotypePanel)
S3method(is_present,GenotypeShard)
S3method(is_present,Sumstats)
S3method(is_present,SumstatsShard)
S3method(is_single_nucleotide_variant,ReferencePanel)
S3method(is_single_nucleotide_variant,ReferenceShard)
S3method(is_strand_ambiguous,ReferencePanel)
S3method(is_strand_ambiguous,ReferenceShard)
S3method(is_subject_present,GenotypePanel)
S3method(logpvec,Sumstats)
S3method(logpvec,SumstatsShard)
S3method(mother_id,GenotypePanel)
S3method(multiply_r2,LDPanel)
S3method(nvec,Sumstats)
S3method(nvec,SumstatsShard)
S3method(num_annot,AnnotationPanel)
S3method(num_annot,AnnotationShard)
S3method(num_sample,GenotypePanel)
S3method(num_snp,LDPanel)
S3method(num_snp,LDShard)
S3method(num_snp,ReferencePanel)
S3method(num_snp,ReferenceShard)
S3method(num_snp,AnnotationPanel)
S3method(num_snp,AnnotationShard)
S3method(num_snp,GenotypePanel)
S3method(num_snp,GenotypeShard)
S3method(num_snp,Sumstats)
S3method(num_snp,SumstatsShard)
S3method(ploidy_female,GenotypePanel)
S3method(ploidy_female,GenotypeShard)
S3method(ploidy_male,GenotypePanel)
S3method(ploidy_male,GenotypeShard)
S3method(print,GenotypePanel)
S3method(print,ReferencePanel)
S3method(print,ReferenceShard)
S3method(print,AnnotationPanel)
S3method(print,LDPanel)
S3method(print,Sumstats)
S3method(reference,LDPanel)
S3method(save_cache,AnnotationPanel)
S3method(save_cache,GenotypePanel)
S3method(save_cache,ReferencePanel)
S3method(save_cache,Sumstats)
S3method(se_vec,Sumstats)
S3method(se_vec,SumstatsShard)
S3method(select_annotations,AnnotationPanel)
S3method(select_shards,AnnotationPanel)
S3method(select_shards,GenotypePanel)
S3method(select_shards,LDPanel)
S3method(select_shards,ReferencePanel)
S3method(select_shards,Sumstats)
S3method(sex,GenotypePanel)
S3method(shard_offsets,ReferencePanel)
S3method(shard_offsets,AnnotationPanel)
S3method(shard_offsets,GenotypePanel)
S3method(shard_offsets,LDPanel)
S3method(shard_offsets,Sumstats)
S3method(shards,AnnotationPanel)
S3method(shards,GenotypePanel)
S3method(shards,LDPanel)
S3method(shards,ReferencePanel)
S3method(shards,Sumstats)
S3method(shards,default)
S3method(snp,ReferencePanel)
S3method(snp,ReferenceShard)
S3method(source_layout,GenotypePanel)
S3method(source_row0,GenotypePanel)
S3method(source_row0,GenotypeShard)
S3method(union_annotations,AnnotationPanel)
S3method(validate_checksums,ReferencePanel)
S3method(zvec,Sumstats)
S3method(zvec,SumstatsShard)
export(a1)
export(a1_hash64)
export(a2)
export(a2_hash64)
export(a1freq)
export(annomat)
export(annonames)
export(beta_vec)
export(bp)
export(chr)
export(create_annotation)
export(create_annotations)
export(create_sumstats)
export(default_chrX_sex)
export(eaf_vec)
export(father_id)
export(fast_prune)
export(fetch_genotypes)
export(fetch_genotypes_int8)
export(fid)
export(get_verbosity)
export(info_vec)
export(iid)
export(is_female)
export(is_male)
export(is_object_compatible)
export(is_present)
export(is_single_nucleotide_variant)
export(is_strand_ambiguous)
export(is_subject_present)
export(load_annotations)
export(load_annotations_cache)
export(load_genotype)
export(load_genotype_cache)
export(load_ld)
export(load_ld_reference)
export(load_reference)
export(load_reference_cache)
export(load_sumstats)
export(load_sumstats_cache)
export(logpvec)
export(mother_id)
export(multiply_r2)
export(nvec)
export(num_annot)
export(num_sample)
export(num_snp)
export(ploidy_female)
export(ploidy_male)
export(prepare_ld_npz_for_r)
export(reference)
export(save_cache)
export(save_annotations_cache)
export(save_genotype_cache)
export(save_reference_cache)
export(save_sumstats_cache)
export(se_vec)
export(select_annotations)
export(select_shards)
export(sex)
export(set_verbosity)
export(shard_offsets)
export(shards)
export(snp)
export(source_layout)
export(source_row0)
export(union_annotations)
export(validate_checksums)
export(validate_ld_distribution)
export(version)
export(zvec)
importFrom(Matrix,Matrix)
importFrom(Matrix,diag)
importFrom(Matrix,sparseMatrix)
importFrom(Matrix,t)
importFrom(bit64,integer64)
importFrom(digest,digest)
importFrom(jsonlite,fromJSON)
importFrom(methods,as)

# statgen R Package

Statistical genetics data objects and loaders for R. This package is part of the larger `statgen` repository, which also includes Python and MATLAB/Octave implementations.

## Installation

### From GitHub (development)

```r
# Install devtools if needed
if (!require("devtools")) install.packages("devtools")

# Install from the R-package directory
devtools::install()

# Or from the repository root
devtools::install("R-package")
```

### From CRAN (release)

```r
install.packages("statgen")
```

## Documentation

- [Main statgen repository](https://github.com/precimed/statgen)
- Package vignette: `vignette("statgen")`

Tiny CRAN-safe fixtures for R package tests are included here in built package
tarballs.

Portable-format fixtures copied into this directory must be small and derived
from the repository-level canonical fixtures under `tests/fixtures/` when
possible. They are generated working-tree copies, not independent committed
sources. Broader cross-runtime parity tests belong in repository-level pytest
tests.

R-native fixtures with no portable counterpart may be committed directly here.

Current copied fixtures:

- `reference_chr1.bim` is a byte-for-byte copy of
  `tests/fixtures/reference/sharded/1.bim`.
- `traits.tsv.gz` is a byte-for-byte copy of
  `tests/fixtures/sumstats/traits.tsv.gz`; it intentionally exercises
  warning paths for missing optional summary-statistics values.
- `traits_complete.tsv.gz` is a byte-for-byte copy of
  `tests/fixtures/sumstats/traits_complete.tsv.gz`; it is the warning-free
  summary-statistics fixture used by examples and vignettes.
- `ld/` is an R-ready copy of `tests/fixtures/ld/python/` with an added
  `reference_cache.rds` and manifest `r_reference_cache` fields. It is used by
  examples and vignettes so installed-package examples can call `load_ld()`
  directly without mutating files under `inst/extdata`.

Prepare copied R fixtures with:

```sh
make prepare-r-fixtures
```

Repository pytest checks that copied fixtures remain synchronized with their
canonical source files.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## -----------------------------------------------------------------------------
library(statgen)
set_verbosity("quiet")

## -----------------------------------------------------------------------------
reference_template <- file.path(
  dirname(system.file("extdata", "reference_chr1.bim", package = "statgen")),
  "reference_chr@.bim"
)
reference <- load_reference(reference_template)

sumstats_path <- system.file("extdata", "traits_complete.tsv.gz", package = "statgen")
sumstats <- load_sumstats(sumstats_path, reference)

annotation_paths <- system.file(
  "extdata",
  c("anno1.bed", "anno2.bed"),
  package = "statgen"
)
annotations <- load_annotations(annotation_paths, reference)

num_snp(reference)
head(logpvec(sumstats))
colnames(annomat(annotations))

## -----------------------------------------------------------------------------
sumstats_cache <- tempfile(fileext = ".rds")
save_sumstats_cache(sumstats, sumstats_cache)
sumstats_cached <- load_sumstats_cache(sumstats_cache)
identical(is_present(sumstats_cached), is_present(sumstats))

## -----------------------------------------------------------------------------
ld_root <- system.file("extdata", "ld", package = "statgen")
ld <- load_ld(ld_root)

num_snp(ld)
a1freq(ld)

## -----------------------------------------------------------------------------
scores <- seq_len(num_snp(ld))
multiply_r2(ld, scores)

logp <- rev(seq_len(num_snp(ld)))
fast_prune(logp, ld, r2_threshold = 0.35)

## -----------------------------------------------------------------------------
ref_chr1 <- load_reference(system.file("extdata", "reference_chr1.bim", package = "statgen"))
bed_chr1 <- system.file("extdata", "genotype_1.bed", package = "statgen")
genotype <- load_genotype(sub("\\.bed$", "", bed_chr1), ref_chr1)

geno <- fetch_genotypes_int8(genotype, c(1, 3))
dim(geno)
head(geno)

.allele_hash_p <- 2147483647
.allele_hash_base1 <- 257
.allele_hash_base2 <- 263
.allele_hash_max_chars <- 150L
.hash64_shift <- bit64::as.integer64("4294967296")
.single_base_hash64 <- bit64::as.integer64(c(
  "1387274436937",
  "1395864371531",
  "1413044240719",
  "1468878815580"
))
names(.single_base_hash64) <- c("A", "C", "G", "T")

.allele_hash64 <- function(alleles) {
  alleles <- as.character(alleles)
  if (!length(alleles)) {
    return(bit64::integer64(0L))
  }
  if (anyNA(alleles)) {
    stop("Alleles must be non-missing character values", call. = FALSE)
  }

  lengths <- nchar(alleles, type = "chars", allowNA = FALSE)
  long <- lengths > .allele_hash_max_chars
  if (any(long)) {
    warning("Allele length exceeds 150 characters; hashing uses first 150 characters", call. = FALSE)
    alleles[long] <- substr(alleles[long], 1L, .allele_hash_max_chars)
    lengths <- pmin(lengths, .allele_hash_max_chars)
  }

  out <- bit64::integer64(length(alleles))
  single_base <- lengths == 1L
  if (any(single_base)) {
    single_idx <- which(single_base)
    assigned <- rep(FALSE, length(single_idx))
    single_alleles <- alleles[single_idx]
    for (base in names(.single_base_hash64)) {
      matches <- single_alleles == base
      if (any(matches)) {
        out[single_idx[matches]] <- .single_base_hash64[[base]]
        assigned[matches] <- TRUE
      }
    }
    if (any(!assigned)) {
      other_idx <- single_idx[!assigned]
      out[other_idx] <- .allele_hash64_utf8(alleles[other_idx])
    }
  }
  if (all(single_base)) {
    return(out)
  }

  multi_idx <- which(!single_base)
  out[multi_idx] <- .allele_hash64_utf8(alleles[multi_idx])
  out
}

.allele_hash64_utf8 <- function(alleles) {
  h1 <- rep.int(1, length(alleles))
  h2 <- rep.int(1, length(alleles))
  bytes <- lapply(alleles, charToRaw)
  byte_lengths <- lengths(bytes)
  max_len <- max(byte_lengths)
  if (max_len > 0L) {
    for (j in seq_len(max_len)) {
      active <- byte_lengths >= j
      byte <- vapply(bytes[active], function(x) as.integer(x[[j]]), integer(1))
      x <- byte + 1
      h1[active] <- (h1[active] * .allele_hash_base1 + x) %% .allele_hash_p
      h2[active] <- (h2[active] * .allele_hash_base2 + x) %% .allele_hash_p
    }
  }
  bit64::as.integer64(h1) * .hash64_shift + bit64::as.integer64(h2)
}

.is_single_nucleotide_variant_hash <- function(a1_hash, a2_hash) {
  a1_hash %in% .single_base_hash64 & a2_hash %in% .single_base_hash64
}

.is_strand_ambiguous_hash <- function(a1_hash, a2_hash) {
  h <- .single_base_hash64
  (a1_hash == h[["A"]] & a2_hash == h[["T"]]) |
    (a1_hash == h[["T"]] & a2_hash == h[["A"]]) |
    (a1_hash == h[["C"]] & a2_hash == h[["G"]]) |
    (a1_hash == h[["G"]] & a2_hash == h[["C"]])
}

.genotype_cache_schema <- "genotype_cache/0.1"
.bed_missing_int <- -1L
.haploid_modes <- c("raw", "ploidy_scaled")

num_sample <- function(x, ...) UseMethod("num_sample")
fid <- function(x, ...) UseMethod("fid")
iid <- function(x, ...) UseMethod("iid")
father_id <- function(x, ...) UseMethod("father_id")
mother_id <- function(x, ...) UseMethod("mother_id")
sex <- function(x, ...) UseMethod("sex")
is_male <- function(x, ...) UseMethod("is_male")
is_female <- function(x, ...) UseMethod("is_female")
ploidy_male <- function(x, ...) UseMethod("ploidy_male")
ploidy_female <- function(x, ...) UseMethod("ploidy_female")
source_layout <- function(x, ...) UseMethod("source_layout")
source_row0 <- function(x, ...) UseMethod("source_row0")
is_subject_present <- function(x, shard, ...) UseMethod("is_subject_present")
fetch_genotypes_int8 <- function(x, snp_indices, bed_path = NULL, ...) UseMethod("fetch_genotypes_int8")
fetch_genotypes <- function(x, snp_indices, bed_path = NULL, haploid_mode = NULL, ...) UseMethod("fetch_genotypes")

load_genotype <- function(bfile_prefix, reference) {
  prefix <- .validate_path_scalar(bfile_prefix, "bfile_prefix")
  if (!inherits(reference, "ReferencePanel")) {
    stop("reference must be a ReferencePanel", call. = FALSE)
  }
  ref_shards <- shards(reference)
  if (!length(ref_shards)) {
    stop("load_genotype requires a reference with at least one shard", call. = FALSE)
  }
  ref_labels <- vapply(ref_shards, function(s) s$label, character(1))
  .validate_requested_shards(ref_labels, ref_labels, "load_genotype reference")

  if (grepl("@", prefix, fixed = TRUE)) {
    layout <- "sharded"
    sources <- stats::setNames(vector("list", length(ref_labels)), ref_labels)
    for (label in ref_labels) {
      sources[[label]] <- .load_genotype_source_record(gsub("@", label, prefix, fixed = TRUE), label)
    }
  } else {
    layout <- "non_sharded"
    shared <- .load_genotype_source_record(prefix, NULL)
    if ("X" %in% ref_labels && any(shared$bim$chr == "X") && !file.exists(paste0(prefix, ".ploidy"))) {
      warning(
        sprintf("%s: chrX genotype source has no .ploidy sidecar; defaulting chrX rows to male/female ploidy (1, 2)", prefix),
        call. = FALSE
      )
    }
    sources <- stats::setNames(rep(list(shared), length(ref_labels)), ref_labels)
  }

  subject_maps <- stats::setNames(vector("list", length(ref_labels)), ref_labels)
  if (identical(layout, "non_sharded")) {
    panel_fam <- sources[[1]]$fam
    for (label in ref_labels) {
      if (!any(sources[[label]]$bim$chr == label)) {
        warning(
          sprintf("%s: no source BIM rows for requested reference shard %s; marking shard absent", prefix, sQuote(label)),
          call. = FALSE
        )
      }
      subject_maps[[label]] <- .all_subjects_present(sources[[label]]$source_num_sample)
    }
  } else {
    autosomal <- ref_labels[ref_labels != "X"]
    if (length(autosomal)) {
      panel_fam <- sources[[autosomal[[1]]]]$fam
      if (length(autosomal) > 1L) {
        for (label in autosomal[-1L]) {
          if (!.fam_equal(sources[[label]]$fam, panel_fam)) {
            stop(sprintf("load_genotype: autosomal FAM mismatch for shard %s", sQuote(label)), call. = FALSE)
          }
        }
      }
      for (label in autosomal) {
        subject_maps[[label]] <- .all_subjects_present(sources[[label]]$source_num_sample)
      }
      if ("X" %in% ref_labels) {
        subject_maps[["X"]] <- .map_chrx_subjects(sources[["X"]]$fam, panel_fam, "load_genotype")
      }
    } else {
      panel_fam <- sources[["X"]]$fam
      subject_maps[["X"]] <- .all_subjects_present(sources[["X"]]$source_num_sample)
    }
  }

  out <- vector("list", length(ref_shards))
  for (i in seq_along(ref_shards)) {
    label <- ref_shards[[i]]$label
    out[[i]] <- .build_genotype_shard(ref_shards[[i]], sources[[label]], subject_maps[[label]])
  }
  .new_genotype_panel(
    out,
    fid = panel_fam$fid,
    iid = panel_fam$iid,
    father_id = panel_fam$father_id,
    mother_id = panel_fam$mother_id,
    sex = panel_fam$sex,
    source_layout = layout
  )
}

save_genotype_cache <- function(panel, path) {
  if (!inherits(panel, "GenotypePanel")) {
    stop("panel must be a GenotypePanel", call. = FALSE)
  }
  path <- .validate_path_scalar(path, "path")
  offsets <- shard_offsets(panel)
  metadata <- list(
    schema = .genotype_cache_schema,
    n_shards = length(panel$shards),
    shard_labels = vapply(panel$shards, function(s) s$label, character(1)),
    shard_checksums = vapply(panel$shards, function(s) s$reference_checksum, character(1)),
    shard_start0 = offsets$start0,
    shard_stop0 = offsets$stop0,
    source_layout = panel$source_layout,
    bed_paths = vapply(panel$shards, function(s) s$bed_path, character(1)),
    bed_file_sizes = vapply(panel$shards, function(s) s$bed_file_size, numeric(1)),
    source_num_snp = vapply(panel$shards, function(s) s$source_num_snp, integer(1)),
    source_num_sample = vapply(panel$shards, function(s) s$source_num_sample, integer(1)),
    num_sample = panel$num_sample
  )
  payload <- list(
    metadata = metadata,
    is_present = is_present(panel),
    ploidy_male = ploidy_male(panel),
    ploidy_female = ploidy_female(panel),
    source_row0 = source_row0(panel),
    subject_present = do.call(cbind, lapply(panel$shards, function(s) s$subject_present)),
    source_subject_row0 = do.call(cbind, lapply(panel$shards, function(s) s$source_subject_row0)),
    fid = fid(panel),
    iid = iid(panel),
    father_id = father_id(panel),
    mother_id = mother_id(panel),
    sex = sex(panel),
    is_male = is_male(panel),
    is_female = is_female(panel)
  )
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  saveRDS(payload, path)
  invisible(NULL)
}

load_genotype_cache <- function(path, shards = NULL) {
  path <- .validate_path_scalar(path, "path")
  payload <- readRDS(path)
  .validate_genotype_cache_payload(payload)
  meta <- payload$metadata
  selected <- .validate_requested_shards(shards, meta$shard_labels, "load_genotype_cache")
  label_to_index <- stats::setNames(seq_along(meta$shard_labels), meta$shard_labels)

  out <- vector("list", length(selected))
  for (i in seq_along(selected)) {
    label <- selected[[i]]
    idx <- label_to_index[[label]]
    start <- meta$shard_start0[[idx]] + 1L
    stop <- meta$shard_stop0[[idx]]
    out[[i]] <- .new_genotype_shard(
      label = label,
      bed_path = meta$bed_paths[[idx]],
      bed_file_size = meta$bed_file_sizes[[idx]],
      source_num_snp = meta$source_num_snp[[idx]],
      source_num_sample = meta$source_num_sample[[idx]],
      source_row0 = payload$source_row0[start:stop],
      subject_present = payload$subject_present[, idx],
      source_subject_row0 = payload$source_subject_row0[, idx],
      is_present = payload$is_present[start:stop],
      ploidy_male = payload$ploidy_male[start:stop],
      ploidy_female = payload$ploidy_female[start:stop],
      reference_checksum = meta$shard_checksums[[idx]]
    )
  }
  .new_genotype_panel(
    out,
    fid = payload$fid,
    iid = payload$iid,
    father_id = payload$father_id,
    mother_id = payload$mother_id,
    sex = payload$sex,
    source_layout = meta$source_layout
  )
}

num_snp.GenotypeShard <- function(x, ...) x$num_snp
num_snp.GenotypePanel <- function(x, ...) x$num_snp
num_sample.GenotypePanel <- function(x, ...) x$num_sample
shards.GenotypePanel <- function(x, ...) x$shards
shard_offsets.GenotypePanel <- function(x, ...) x$shard_offsets
fid.GenotypePanel <- function(x, ...) x$fid
iid.GenotypePanel <- function(x, ...) x$iid
father_id.GenotypePanel <- function(x, ...) x$father_id
mother_id.GenotypePanel <- function(x, ...) x$mother_id
sex.GenotypePanel <- function(x, ...) x$sex
is_male.GenotypePanel <- function(x, ...) x$sex == 1L
is_female.GenotypePanel <- function(x, ...) x$sex == 2L
is_present.GenotypeShard <- function(x, ...) x$is_present
is_present.GenotypePanel <- function(x, ...) unlist(lapply(x$shards, is_present), use.names = FALSE)
ploidy_male.GenotypeShard <- function(x, ...) x$ploidy_male
ploidy_male.GenotypePanel <- function(x, ...) unlist(lapply(x$shards, ploidy_male), use.names = FALSE)
ploidy_female.GenotypeShard <- function(x, ...) x$ploidy_female
ploidy_female.GenotypePanel <- function(x, ...) unlist(lapply(x$shards, ploidy_female), use.names = FALSE)
source_layout.GenotypePanel <- function(x, ...) x$source_layout
source_row0.GenotypeShard <- function(x, ...) x$source_row0
source_row0.GenotypePanel <- function(x, ...) unlist(lapply(x$shards, source_row0), use.names = FALSE)

is_subject_present.GenotypePanel <- function(x, shard, ...) {
  label <- .validate_char_scalar(shard, "shard")
  for (s in x$shards) {
    if (identical(s$label, label)) {
      return(s$subject_present)
    }
  }
  stop(sprintf("GenotypePanel.is_subject_present: requested shard %s is not loaded", sQuote(label)), call. = FALSE)
}

select_shards.GenotypePanel <- function(x, shards, ...) {
  available <- vapply(x$shards, function(s) s$label, character(1))
  selected <- .validate_requested_shards(shards, available, "GenotypePanel.select_shards")
  by_label <- stats::setNames(x$shards, available)
  .new_genotype_panel(
    unname(by_label[selected]),
    fid = x$fid,
    iid = x$iid,
    father_id = x$father_id,
    mother_id = x$mother_id,
    sex = x$sex,
    source_layout = x$source_layout
  )
}

fetch_genotypes_int8.GenotypePanel <- function(x, snp_indices, bed_path = NULL, ...) {
  indices <- .normalize_genotype_indices(snp_indices, x$num_snp)
  out <- matrix(.bed_missing_int, nrow = x$num_sample, ncol = length(indices))
  if (!length(indices)) {
    storage.mode(out) <- "integer"
    return(out)
  }
  addressed <- .addressed_genotype_shards(x, indices)
  bed_paths <- .resolve_fetch_bed_paths(x, addressed, bed_path)
  .validate_requested_genotype_present(x, addressed)

  for (entry in addressed) {
    shard <- entry$shard
    effective_bed <- bed_paths[[shard$label]]
    .validate_bed_file(effective_bed, shard$source_num_sample, shard$source_num_snp)
    source_rows <- shard$source_row0[entry$local_indices]
    source_geno <- .read_bed_rows_int8_codes(effective_bed, source_rows, shard$source_num_sample)
    panel_rows <- which(shard$subject_present)
    if (length(panel_rows)) {
      source_subject_rows <- shard$source_subject_row0[panel_rows] + 1L
      out[panel_rows, entry$cols] <- source_geno[source_subject_rows, , drop = FALSE]
    }
  }
  storage.mode(out) <- "integer"
  out
}

fetch_genotypes.GenotypePanel <- function(x, snp_indices, bed_path = NULL, haploid_mode = NULL, ...) {
  mode <- if (is.null(haploid_mode)) "raw" else as.character(haploid_mode)
  if (length(mode) != 1L || is.na(mode) || !(mode %in% .haploid_modes)) {
    stop("GenotypePanel.fetch_genotypes: haploid_mode must be 'raw' or 'ploidy_scaled'", call. = FALSE)
  }
  indices <- .normalize_genotype_indices(snp_indices, x$num_snp)
  gi <- fetch_genotypes_int8(x, indices, bed_path = bed_path)
  out <- matrix(as.numeric(gi), nrow = nrow(gi), ncol = ncol(gi))
  out[gi == .bed_missing_int] <- NaN
  if (identical(mode, "ploidy_scaled")) {
    out <- .apply_ploidy_scaled(x, out, indices)
  }
  out
}

save_cache.GenotypePanel <- function(x, path, ...) save_genotype_cache(x, path)

print.GenotypePanel <- function(x, ...) {
  labels <- vapply(x$shards, function(s) s$label, character(1))
  cat(sprintf(
    "<GenotypePanel: %d SNPs x %d samples across %d shard(s): %s>\n",
    x$num_snp, x$num_sample, length(labels), paste(labels, collapse = ", ")
  ))
  invisible(x)
}

.new_genotype_shard <- function(label, bed_path, bed_file_size, source_num_snp,
                                source_num_sample, source_row0, subject_present,
                                source_subject_row0, is_present, ploidy_male,
                                ploidy_female, reference_checksum) {
  label <- as.character(label)
  source_num_snp <- as.integer(source_num_snp)
  source_num_sample <- as.integer(source_num_sample)
  bed_file_size <- as.numeric(bed_file_size)
  if (is.na(source_num_snp) || is.na(source_num_sample) || source_num_snp < 0L || source_num_sample < 0L) {
    stop("source_num_snp and source_num_sample must be non-negative", call. = FALSE)
  }
  expected <- .expected_bed_size(source_num_sample, source_num_snp)
  if (!identical(as.numeric(expected), bed_file_size)) {
    stop(sprintf("Invalid genotype metadata for shard %s: bed_file_size mismatch", label), call. = FALSE)
  }

  source_row0 <- as.integer(source_row0)
  is_present <- as.logical(is_present)
  ploidy_male <- as.numeric(ploidy_male)
  ploidy_female <- as.numeric(ploidy_female)
  .validate_equal_lengths(
    list(source_row0 = source_row0, is_present = is_present, ploidy_male = ploidy_male, ploidy_female = ploidy_female),
    sprintf("GenotypeShard %s: SNP-axis vector lengths must match", label)
  )
  if (!identical(is_present, source_row0 >= 0L)) {
    stop(sprintf("GenotypeShard %s: is_present must match source_row0 >= 0", label), call. = FALSE)
  }
  bad_source <- source_row0 < -1L | source_row0 >= source_num_snp
  if (any(bad_source)) {
    stop(sprintf("GenotypeShard %s: source_row0 out of source BIM bounds", label), call. = FALSE)
  }
  absent <- !is_present
  if (any(!is.nan(ploidy_male[absent])) || any(!is.nan(ploidy_female[absent]))) {
    stop(sprintf("GenotypeShard %s: absent SNPs must have NaN ploidy", label), call. = FALSE)
  }
  present_ploidy <- c(ploidy_male[is_present], ploidy_female[is_present])
  if (length(present_ploidy) && any(!(present_ploidy %in% c(0, 1, 2)))) {
    stop(sprintf("GenotypeShard %s: present ploidy values must be 0, 1, or 2", label), call. = FALSE)
  }

  subject_present <- as.logical(subject_present)
  source_subject_row0 <- as.integer(source_subject_row0)
  .validate_equal_lengths(
    list(subject_present = subject_present, source_subject_row0 = source_subject_row0),
    sprintf("GenotypeShard %s: sample-axis vector lengths must match", label)
  )
  if (!identical(subject_present, source_subject_row0 >= 0L)) {
    stop(sprintf("GenotypeShard %s: subject_present must match source_subject_row0 >= 0", label), call. = FALSE)
  }
  bad_subject <- source_subject_row0 < -1L | source_subject_row0 >= source_num_sample
  if (any(bad_subject)) {
    stop(sprintf("GenotypeShard %s: source_subject_row0 out of source FAM bounds", label), call. = FALSE)
  }

  structure(
    list(
      chr = label,
      label = label,
      bed_path = .validate_path_scalar(bed_path, "bed_path"),
      bed_file_size = bed_file_size,
      source_num_snp = source_num_snp,
      source_num_sample = source_num_sample,
      source_row0 = source_row0,
      subject_present = subject_present,
      source_subject_row0 = source_subject_row0,
      is_present = is_present,
      ploidy_male = ploidy_male,
      ploidy_female = ploidy_female,
      reference_checksum = as.character(reference_checksum),
      num_snp = length(is_present)
    ),
    class = "GenotypeShard"
  )
}

.new_genotype_panel <- function(shard_list, fid, iid, father_id, mother_id, sex, source_layout) {
  if (!length(shard_list)) {
    stop("GenotypePanel requires at least one shard", call. = FALSE)
  }
  if (!(source_layout %in% c("non_sharded", "sharded"))) {
    stop("source_layout must be 'non_sharded' or 'sharded'", call. = FALSE)
  }
  fid <- as.character(fid)
  iid <- as.character(iid)
  father_id <- as.character(father_id)
  mother_id <- as.character(mother_id)
  sex <- as.integer(sex)
  .validate_equal_lengths(
    list(fid = fid, iid = iid, father_id = father_id, mother_id = mother_id, sex = sex),
    "GenotypePanel FAM vector lengths must match"
  )
  if (any(!(sex %in% c(0L, 1L, 2L)))) {
    stop("GenotypePanel sex values must be 0, 1, or 2", call. = FALSE)
  }
  num_sample <- length(fid)
  pos <- 0L
  offsets <- data.frame(
    shard_label = character(length(shard_list)),
    start0 = integer(length(shard_list)),
    stop0 = integer(length(shard_list)),
    stringsAsFactors = FALSE
  )
  for (i in seq_along(shard_list)) {
    if (length(shard_list[[i]]$subject_present) != num_sample) {
      stop(sprintf("Genotype shard %s: subject_present length does not match panel sample axis", shard_list[[i]]$label), call. = FALSE)
    }
    offsets$shard_label[[i]] <- shard_list[[i]]$label
    offsets$start0[[i]] <- pos
    pos <- pos + shard_list[[i]]$num_snp
    offsets$stop0[[i]] <- pos
  }
  structure(
    list(
      shards = shard_list,
      fid = fid,
      iid = iid,
      father_id = father_id,
      mother_id = mother_id,
      sex = sex,
      source_layout = source_layout,
      num_sample = num_sample,
      num_snp = pos,
      shard_offsets = offsets
    ),
    class = "GenotypePanel"
  )
}

.load_genotype_source_record <- function(prefix, label) {
  paths <- .require_genotype_source_paths(prefix, label)
  bim <- .parse_bim(paths$bim)
  source_num_snp <- nrow(bim)
  bim$line <- seq_len(source_num_snp)
  bim$source_row0 <- if (source_num_snp) seq.int(0L, source_num_snp - 1L) else integer()
  if (!is.null(label) && any(bim$chr != label)) {
    idx <- which(bim$chr != label)[[1]]
    stop(
      sprintf("%s:%d: sharded genotype BIM for %s contains chr %s", paths$bim, idx, sQuote(label), sQuote(bim$chr[[idx]])),
      call. = FALSE
    )
  }
  fam <- .parse_fam(paths$fam)
  if (identical(label, "X") && is.null(paths$ploidy)) {
    warning(
      sprintf("%s: chrX genotype source has no .ploidy sidecar; defaulting chrX rows to male/female ploidy (1, 2)", prefix),
      call. = FALSE
    )
  }
  ploidy <- .parse_ploidy(paths$ploidy, source_num_snp, bim$chr)
  bed_size <- .validate_bed_file(paths$bed, nrow(fam), source_num_snp)
  list(
    bed_path = paths$bed,
    bim_path = paths$bim,
    bed_file_size = bed_size,
    source_num_snp = source_num_snp,
    source_num_sample = nrow(fam),
    bim = bim,
    fam = fam,
    ploidy_male = ploidy$male,
    ploidy_female = ploidy$female
  )
}

.require_genotype_source_paths <- function(prefix, label) {
  out <- list(
    bed = paste0(prefix, ".bed"),
    bim = paste0(prefix, ".bim"),
    fam = paste0(prefix, ".fam")
  )
  for (path in out) {
    if (!file.exists(path)) {
      where <- if (is.null(label)) "" else sprintf(" for requested shard %s", sQuote(label))
      stop(sprintf("Missing genotype source%s: %s", where, path), call. = FALSE)
    }
  }
  ploidy <- paste0(prefix, ".ploidy")
  out$ploidy <- if (file.exists(ploidy)) ploidy else NULL
  out
}

.all_subjects_present <- function(source_num_sample) {
  list(
    subject_present = rep.int(TRUE, source_num_sample),
    source_subject_row0 = if (source_num_sample) seq.int(0L, source_num_sample - 1L) else integer()
  )
}

.fam_equal <- function(lhs, rhs) {
  identical(lhs[c("fid", "iid", "father_id", "mother_id", "sex")], rhs[c("fid", "iid", "father_id", "mother_id", "sex")])
}

.map_chrx_subjects <- function(chrx_fam, panel_fam, where) {
  if (.fam_equal(chrx_fam, panel_fam)) {
    return(.all_subjects_present(nrow(panel_fam)))
  }
  panel_key <- paste(panel_fam$fid, panel_fam$iid, sep = "\r")
  src_key <- paste(chrx_fam$fid, chrx_fam$iid, sep = "\r")
  panel_row <- match(src_key, panel_key)
  missing <- is.na(panel_row)
  if (any(missing)) {
    idx <- which(missing)[[1]]
    stop(sprintf("%s: chrX FAM subject (%s, %s) is absent from the autosomal sample axis", where, sQuote(chrx_fam$fid[[idx]]), sQuote(chrx_fam$iid[[idx]])), call. = FALSE)
  }
  for (field in c("father_id", "mother_id", "sex")) {
    bad <- chrx_fam[[field]] != panel_fam[[field]][panel_row]
    if (any(bad)) {
      idx <- which(bad)[[1]]
      stop(sprintf("%s: chrX FAM metadata mismatch for subject (%s, %s)", where, sQuote(chrx_fam$fid[[idx]]), sQuote(chrx_fam$iid[[idx]])), call. = FALSE)
    }
  }
  subject_present <- rep.int(FALSE, nrow(panel_fam))
  source_subject_row0 <- rep.int(-1L, nrow(panel_fam))
  subject_present[panel_row] <- TRUE
  source_subject_row0[panel_row] <- seq.int(0L, nrow(chrx_fam) - 1L)
  list(subject_present = subject_present, source_subject_row0 = source_subject_row0)
}

.build_genotype_shard <- function(ref_shard, source, subject_map) {
  source_bim <- source$bim[source$bim$chr == ref_shard$label, , drop = FALSE]
  if (!nrow(source_bim)) {
    local_match <- rep.int(NA_integer_, ref_shard$num_snp)
  } else {
    local_match <- .match_variant_keys(
      bp(ref_shard), a1_hash64(ref_shard), a2_hash64(ref_shard),
      source_bim$bp, source_bim$a1_hash64, source_bim$a2_hash64,
      ref_shard$label, "genotype"
    )
    has_match <- !is.na(local_match)
    matched_source_mask <- rep.int(FALSE, nrow(source_bim))
    matched_source_mask[local_match[has_match]] <- TRUE
    unmatched <- source_bim[!matched_source_mask, , drop = FALSE]
    swapped <- .count_variant_key_intersections(
      bp(ref_shard), a1_hash64(ref_shard), a2_hash64(ref_shard),
      unmatched$bp, unmatched$a2_hash64, unmatched$a1_hash64
    )
    if (swapped > 0L) {
      warning(
        sprintf(
          "%s: shard %s: %d unmatched genotype variant(s) would match the reference if a1/a2 were swapped; variants remain unmatched",
          source$bim_path, ref_shard$label, swapped
        ),
        call. = FALSE
      )
    }
  }

  is_present <- !is.na(local_match)
  source_row0_vec <- rep.int(-1L, ref_shard$num_snp)
  ploidy_male_vec <- rep.int(NaN, ref_shard$num_snp)
  ploidy_female_vec <- rep.int(NaN, ref_shard$num_snp)
  if (any(is_present)) {
    source_rows <- source_bim$source_row0[local_match[is_present]]
    source_row0_vec[is_present] <- source_rows
    ploidy_male_vec[is_present] <- source$ploidy_male[source_rows + 1L]
    ploidy_female_vec[is_present] <- source$ploidy_female[source_rows + 1L]
  }

  .new_genotype_shard(
    label = ref_shard$label,
    bed_path = source$bed_path,
    bed_file_size = source$bed_file_size,
    source_num_snp = source$source_num_snp,
    source_num_sample = source$source_num_sample,
    source_row0 = source_row0_vec,
    subject_present = subject_map$subject_present,
    source_subject_row0 = subject_map$source_subject_row0,
    is_present = is_present,
    ploidy_male = ploidy_male_vec,
    ploidy_female = ploidy_female_vec,
    reference_checksum = ref_shard$checksum
  )
}

.normalize_genotype_indices <- function(snp_indices, num_snp_value) {
  if (is.null(snp_indices)) {
    stop("GenotypePanel.fetch_genotypes: snp_indices must be integer indices", call. = FALSE)
  }
  if (!length(snp_indices)) {
    return(integer())
  }
  arr <- as.numeric(snp_indices)
  if (length(arr) != length(snp_indices) || any(is.na(arr) | !is.finite(arr) | floor(arr) != arr)) {
    stop("GenotypePanel.fetch_genotypes: snp_indices must be integer indices", call. = FALSE)
  }
  arr <- as.integer(arr)
  bad <- arr < 1L | arr > num_snp_value
  if (any(bad)) {
    stop(
      sprintf("GenotypePanel.fetch_genotypes: SNP index %d is out of bounds for num_snp=%d", arr[which(bad)[[1]]], num_snp_value),
      call. = FALSE
    )
  }
  arr
}

.addressed_genotype_shards <- function(panel, indices) {
  out <- list()
  offsets <- shard_offsets(panel)
  for (i in seq_along(panel$shards)) {
    start <- offsets$start0[[i]] + 1L
    stop <- offsets$stop0[[i]]
    cols <- which(indices >= start & indices <= stop)
    if (length(cols)) {
      out[[length(out) + 1L]] <- list(
        shard = panel$shards[[i]],
        cols = cols,
        local_indices = indices[cols] - offsets$start0[[i]]
      )
    }
  }
  out
}

.resolve_fetch_bed_paths <- function(panel, addressed, bed_path) {
  if (is.null(bed_path)) {
    out <- lapply(addressed, function(entry) entry$shard$bed_path)
    names(out) <- vapply(addressed, function(entry) entry$shard$label, character(1))
    return(out)
  }
  override <- .validate_path_scalar(bed_path, "bed_path")
  if (grepl("@", override, fixed = TRUE)) {
    if (identical(panel$source_layout, "non_sharded")) {
      stop("GenotypePanel.fetch_genotypes: @ override incompatible with non-sharded panel metadata", call. = FALSE)
    }
    out <- lapply(addressed, function(entry) gsub("@", entry$shard$label, override, fixed = TRUE))
    names(out) <- vapply(addressed, function(entry) entry$shard$label, character(1))
    return(out)
  }
  if (identical(panel$source_layout, "sharded") && length(addressed)) {
    nsnp <- unique(vapply(addressed, function(entry) entry$shard$source_num_snp, integer(1)))
    nsample <- unique(vapply(addressed, function(entry) entry$shard$source_num_sample, integer(1)))
    if (length(nsnp) != 1L || length(nsample) != 1L) {
      stop("GenotypePanel.fetch_genotypes: flat bed_path override requires equal source_num_snp and source_num_sample across addressed shards", call. = FALSE)
    }
  }
  out <- rep(list(override), length(addressed))
  names(out) <- vapply(addressed, function(entry) entry$shard$label, character(1))
  out
}

.validate_requested_genotype_present <- function(panel, addressed) {
  start_by_label <- stats::setNames(shard_offsets(panel)$start0, shard_offsets(panel)$shard_label)
  for (entry in addressed) {
    present <- entry$shard$is_present[entry$local_indices]
    if (!all(present)) {
      j <- which(!present)[[1]]
      panel_index <- start_by_label[[entry$shard$label]] + entry$local_indices[[j]]
      stop(
        sprintf("GenotypePanel.fetch_genotypes: requested SNP %d in shard %s is not present in the genotype source", panel_index, sQuote(entry$shard$label)),
        call. = FALSE
      )
    }
  }
}

.apply_ploidy_scaled <- function(panel, geno, indices) {
  if (!length(indices)) {
    return(geno)
  }
  unknown <- panel$sex == 0L
  male <- panel$sex == 1L
  female <- panel$sex == 2L
  addressed <- .addressed_genotype_shards(panel, indices)
  for (entry in addressed) {
    shard <- entry$shard
    pm <- shard$ploidy_male[entry$local_indices]
    pf <- shard$ploidy_female[entry$local_indices]
    differs <- pm != pf
    unknown_present <- unknown & shard$subject_present
    if (any(unknown_present) && any(differs) && any(is.finite(geno[unknown_present, entry$cols[differs], drop = FALSE]))) {
      stop("GenotypePanel.fetch_genotypes: haploid_mode='ploidy_scaled' requires known FAM sex when male and female ploidy differ", call. = FALSE)
    }
    if (any(male)) {
      geno[male, entry$cols] <- sweep(geno[male, entry$cols, drop = FALSE], 2L, pm / 2, `*`)
    }
    if (any(female)) {
      geno[female, entry$cols] <- sweep(geno[female, entry$cols, drop = FALSE], 2L, pf / 2, `*`)
    }
    same_ploidy <- !differs
    if (any(unknown) && any(same_ploidy)) {
      cols <- entry$cols[same_ploidy]
      geno[unknown, cols] <- sweep(geno[unknown, cols, drop = FALSE], 2L, pf[same_ploidy] / 2, `*`)
    }
  }
  geno
}

.validate_genotype_cache_payload <- function(payload) {
  cache <- .validate_cache_payload_metadata(
    payload,
    .genotype_cache_schema,
    "genotype",
    extra_fields = c("bed_paths", "bed_file_sizes", "source_num_snp", "source_num_sample")
  )
  meta <- cache$meta
  n_shards <- cache$n_shards
  total <- cache$total
  .validate_requested_shards(meta$shard_labels, meta$shard_labels, "load_genotype_cache")
  if (!(meta$source_layout %in% c("non_sharded", "sharded"))) {
    stop("Invalid genotype cache: source_layout must be 'non_sharded' or 'sharded'", call. = FALSE)
  }
  for (field in c("is_present", "ploidy_male", "ploidy_female", "source_row0", "subject_present", "source_subject_row0", "fid", "iid", "father_id", "mother_id", "sex", "is_male", "is_female")) {
    if (is.null(payload[[field]])) {
      stop(sprintf("Invalid genotype cache: missing field %s", sQuote(field)), call. = FALSE)
    }
  }
  if (length(payload$is_present) != total || length(payload$source_row0) != total) {
    stop("Invalid genotype cache: SNP-axis vector lengths mismatch", call. = FALSE)
  }
  if (length(payload$ploidy_male) != total || length(payload$ploidy_female) != total) {
    stop("Invalid genotype cache: ploidy vector lengths mismatch", call. = FALSE)
  }
  num_sample_value <- as.integer(meta$num_sample)
  if (is.na(num_sample_value) || num_sample_value < 0L) {
    stop("Invalid genotype cache: num_sample must be non-negative", call. = FALSE)
  }
  for (field in c("fid", "iid", "father_id", "mother_id", "sex", "is_male", "is_female")) {
    if (length(payload[[field]]) != num_sample_value) {
      stop(sprintf("Invalid genotype cache: %s length mismatch", field), call. = FALSE)
    }
  }
  sex_vec <- as.integer(payload$sex)
  if (any(!(sex_vec %in% c(0L, 1L, 2L)))) {
    stop("Invalid genotype cache: sex values must be 0, 1, or 2", call. = FALSE)
  }
  if (!identical(as.logical(payload$is_male), sex_vec == 1L)) {
    stop("Invalid genotype cache: is_male does not match sex", call. = FALSE)
  }
  if (!identical(as.logical(payload$is_female), sex_vec == 2L)) {
    stop("Invalid genotype cache: is_female does not match sex", call. = FALSE)
  }
  if (!identical(as.integer(dim(payload$subject_present)), c(num_sample_value, n_shards))) {
    stop("Invalid genotype cache: subject_present shape mismatch", call. = FALSE)
  }
  if (!identical(as.integer(dim(payload$source_subject_row0)), c(num_sample_value, n_shards))) {
    stop("Invalid genotype cache: source_subject_row0 shape mismatch", call. = FALSE)
  }
  for (i in seq_len(n_shards)) {
    expected <- .expected_bed_size(meta$source_num_sample[[i]], meta$source_num_snp[[i]])
    if (!identical(as.numeric(expected), as.numeric(meta$bed_file_sizes[[i]]))) {
      stop(sprintf("Invalid genotype cache: bed_file_size mismatch for shard %s", sQuote(meta$shard_labels[[i]])), call. = FALSE)
    }
  }
  invisible(NULL)
}

.prepare_sharded_genotype_fixture <- function() {
  root <- tempfile("statgen-genotype-")
  dir.create(root)
  file.copy(system.file("extdata", "reference_chr1.bim", package = "statgen", mustWork = TRUE), file.path(root, "1.bim"))
  file.copy(system.file("extdata", "reference_chrX.bim", package = "statgen", mustWork = TRUE), file.path(root, "X.bim"))
  for (label in c("1", "X")) {
    for (suffix in c(".bed", ".bim", ".fam")) {
      file.copy(
        system.file("extdata", paste0("genotype_", label, suffix), package = "statgen", mustWork = TRUE),
        file.path(root, paste0("g", label, suffix))
      )
    }
  }
  file.copy(
    system.file("extdata", "genotype_X.ploidy", package = "statgen", mustWork = TRUE),
    file.path(root, "gX.ploidy")
  )
  root
}

test_that("load_genotype reads bundled PLINK fixture and fetches slices", {
  ref_path <- system.file("extdata", "reference_chr1.bim", package = "statgen", mustWork = TRUE)
  bed_path <- system.file("extdata", "genotype_1.bed", package = "statgen", mustWork = TRUE)
  prefix <- sub("\\.bed$", "", bed_path)

  ref <- load_reference(ref_path)
  g <- load_genotype(prefix, ref)

  expect_s3_class(g, "GenotypePanel")
  expect_equal(num_snp(g), 5L)
  expect_equal(num_sample(g), 4L)
  expect_equal(source_layout(g), "non_sharded")
  expect_equal(fid(g), c("FAM1", "FAM1", "FAM2", "FAM2"))
  expect_equal(sex(g), c(1L, 2L, 1L, 2L))
  expect_equal(source_row0(g), 0:4)

  gi <- fetch_genotypes_int8(g, c(1, 5))
  expect_equal(dim(gi), c(4L, 2L))
  expect_equal(as.integer(gi), rep(2L, 8L))
  gi_empty <- fetch_genotypes_int8(g, integer(0))
  expect_equal(dim(gi_empty), c(4L, 0L))
  expect_equal(typeof(gi_empty), "integer")

  gf <- fetch_genotypes(g, c(1, 5))
  expect_equal(dim(gf), c(4L, 2L))
  expect_equal(as.numeric(gf), rep(2, 8L))
})

test_that("sharded genotype exposes chrX ploidy, subject presence, and shard subset", {
  root <- .prepare_sharded_genotype_fixture()
  ref <- load_reference(file.path(root, "@.bim"))
  g <- load_genotype(file.path(root, "g@"), ref)

  expect_equal(source_layout(g), "sharded")
  expect_equal(ploidy_male(g), c(rep(2, 5), rep(1, 3)))
  expect_equal(ploidy_female(g), rep(2, 8))
  expect_equal(is_subject_present(g, "X"), rep(TRUE, 4))

  gx <- select_shards(g, "X")
  expect_equal(num_snp(gx), 3L)
  expect_equal(num_sample(gx), 4L)
  expect_equal(vapply(shards(gx), function(s) s$chr, character(1)), "X")
})

test_that("genotype metadata cache round-trips", {
  root <- .prepare_sharded_genotype_fixture()
  cache <- tempfile(fileext = ".rds")
  override <- file.path(root, "override@.bed")

  ref <- load_reference(file.path(root, "@.bim"))
  g <- load_genotype(file.path(root, "g@"), ref)
  save_genotype_cache(g, cache)
  file.copy(file.path(root, "g1.bed"), file.path(root, "override1.bed"))
  file.copy(file.path(root, "gX.bed"), file.path(root, "overrideX.bed"))
  unlink(file.path(root, "g1.bed"))
  unlink(file.path(root, "gX.bed"))
  loaded <- load_genotype_cache(cache)
  loaded_x <- load_genotype_cache(cache, shards = "X")

  expect_equal(source_layout(loaded), source_layout(g))
  expect_equal(source_row0(loaded), source_row0(g))
  expect_equal(fid(loaded), fid(g))
  expect_error(fetch_genotypes_int8(loaded, 1), "File not found")
  expect_equal(fetch_genotypes_int8(loaded, c(1, 6), bed_path = override), fetch_genotypes_int8(g, c(1, 6), bed_path = override))
  expect_equal(num_snp(loaded_x), 3L)
  expect_equal(source_layout(loaded_x), "sharded")
  expect_equal(fetch_genotypes_int8(loaded_x, 1, bed_path = override), fetch_genotypes_int8(g, 6, bed_path = override))
})

test_that("genotype validation rejects bad sex and malformed cache", {
  root <- .prepare_sharded_genotype_fixture()
  writeLines("FAM1\tIND1\t0\t0\t3\t-9", file.path(root, "bad.fam"))
  file.copy(file.path(root, "g1.bim"), file.path(root, "bad.bim"))
  file.copy(file.path(root, "g1.bed"), file.path(root, "bad.bed"))
  ref1 <- load_reference(file.path(root, "1.bim"))

  expect_error(load_genotype(file.path(root, "bad"), ref1), "FAM sex")

  bad_cache <- tempfile(fileext = ".rds")
  saveRDS(list(metadata = list(schema = "genotype_cache/0.1", n_shards = 0L)), bad_cache)
  expect_error(load_genotype_cache(bad_cache), "n_shards")
})

.load_ld_manifest_reference <- function(root, manifest, shards) {
  reference_cache_path <- .ld_r_reference_cache_path(root, manifest)
  .require_ld_file(reference_cache_path)
  load_reference_cache(reference_cache_path, shards = shards)
}

.ld_r_reference_cache_path <- function(root, manifest) {
  file.path(root, .validate_plain_relative_filename(manifest$r_reference_cache, "r_reference_cache"))
}

.save_reference_cache_atomic <- function(panel, path) {
  tmp <- tempfile(".reference_cache_", tmpdir = dirname(path), fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  save_reference_cache(panel, tmp)
  if (!file.rename(tmp, path)) {
    stop(sprintf("Failed to replace reference cache: %s", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_ld_reference_compatibility <- function(shard, ref_shard, path) {
  if (!identical(shard$chr, ref_shard$label)) {
    stop(sprintf("%s: LD chr %s does not match reference shard %s", path, sQuote(shard$chr), sQuote(ref_shard$label)), call. = FALSE)
  }
  if (!identical(as.integer(shard$num_snp), as.integer(ref_shard$num_snp))) {
    stop(sprintf("%s: LD num_snp does not match reference shard", path), call. = FALSE)
  }
  if (!identical(shard$reference_checksum, ref_shard$checksum)) {
    stop(sprintf("%s: LD reference_checksum does not match reference shard", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_ld_reference_entry <- function(ref_shard, entry, path) {
  if (!identical(as.integer(ref_shard$num_snp), as.integer(entry$num_snp))) {
    stop(sprintf("%s: reference cache num_snp does not match manifest", path), call. = FALSE)
  }
  if (!identical(ref_shard$checksum, entry$reference_checksum)) {
    stop(sprintf("%s: reference cache reference_checksum does not match manifest", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_bundled_reference_bim <- function(path, entry) {
  .require_ld_file(path)
  bim <- .parse_bim(path)
  keep <- bim$chr %in% .canonical_chr_order
  bim <- bim[keep, , drop = FALSE]
  if (!nrow(bim)) {
    stop(sprintf("%s: bundled reference BIM has no supported rows", path), call. = FALSE)
  }
  if (any(bim$chr != entry$chr)) {
    stop(sprintf("%s: bundled reference BIM chr does not match manifest", path), call. = FALSE)
  }
  checksum <- .checksum_from_arrays(bim$chr, bim$bp, bim$a1, bim$a2)
  if (!identical(nrow(bim), as.integer(entry$num_snp))) {
    stop(sprintf("%s: bundled reference BIM num_snp does not match manifest", path), call. = FALSE)
  }
  if (!identical(checksum, entry$reference_checksum)) {
    stop(sprintf("%s: bundled reference BIM checksum does not match manifest", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_ld_manifest_reference_panel <- function(reference_panel, entries, root) {
  by_label <- stats::setNames(shards(reference_panel), vapply(shards(reference_panel), function(s) s$label, character(1)))
  seen <- character()
  for (entry in entries) {
    ref_shard <- by_label[[entry$chr]]
    if (is.null(ref_shard)) {
      stop(sprintf("%s: reference cache missing shard %s", root, entry$chr), call. = FALSE)
    }
    .validate_ld_reference_entry(ref_shard, entry, root)
    if (!(entry$chr %in% seen)) {
      .require_ld_file(file.path(root, entry$reference_bim))
      seen <- c(seen, entry$chr)
    }
  }
  invisible(NULL)
}

.load_reference_from_ld_bims <- function(root, entries, selected_labels) {
  by_chr <- list()
  for (entry in entries) {
    if (is.null(by_chr[[entry$chr]])) {
      by_chr[[entry$chr]] <- entry$reference_bim
    }
  }
  out <- vector("list", length(selected_labels))
  for (i in seq_along(selected_labels)) {
    label <- selected_labels[[i]]
    bim_path <- file.path(root, by_chr[[label]])
    .require_ld_file(bim_path)
    bim <- .parse_bim(bim_path)
    keep <- bim$chr == label
    out[[i]] <- .new_reference_shard(
      label, bim$chr[keep], bim$snp[keep], bim$bp[keep], bim$a1[keep], bim$a2[keep],
      bim$a1_hash64[keep], bim$a2_hash64[keep]
    )
  }
  .new_reference_panel(out)
}

.parse_bim <- function(path) {
  if (!file.exists(path)) {
    stop(sprintf("File not found: %s", path), call. = FALSE)
  }
  df <- tryCatch(
    utils::read.table(
      path,
      header = FALSE,
      sep = "",
      quote = "",
      comment.char = "",
      stringsAsFactors = FALSE,
      colClasses = rep("character", 6L),
      fill = FALSE
    ),
    error = function(e) {
      stop(sprintf("%s: expected 6 whitespace-delimited columns", path), call. = FALSE)
    }
  )
  if (ncol(df) != 6L) {
    stop(sprintf("%s: expected 6 whitespace-delimited columns, got %d", path, ncol(df)), call. = FALSE)
  }
  names(df) <- c("chr", "snp", "cm", "bp", "a1", "a2")
  bad_chr <- is.na(df$chr) | df$chr == ""
  if (any(bad_chr)) {
    stop(sprintf("%s:%d: chr must be non-empty", path, which(bad_chr)[[1]]), call. = FALSE)
  }
  chr_style <- startsWith(tolower(df$chr), "chr")
  if (any(chr_style)) {
    stop(sprintf("%s:%d: chr-style labels (e.g., chr1/chrX) are not allowed", path, which(chr_style)[[1]]), call. = FALSE)
  }
  known_chr <- df$chr %in% c(.canonical_chr_order, .ignored_chr)
  if (!all(known_chr)) {
    idx <- which(!known_chr)[[1]]
    stop(sprintf("%s:%d: unsupported chr label %s; expected 1-22, X (Y/MT are ignored)", path, idx, sQuote(df$chr[[idx]])), call. = FALSE)
  }
  bad_allele <- is.na(df$a1) | is.na(df$a2) | df$a1 == "" | df$a2 == ""
  if (any(bad_allele)) {
    stop(sprintf("%s:%d: a1 and a2 must be non-empty", path, which(bad_allele)[[1]]), call. = FALSE)
  }
  bad_a1 <- !grepl("^[ACGT]+$", df$a1)
  if (any(bad_a1)) {
    idx <- which(bad_a1)[[1]]
    stop(sprintf("%s:%d: a1 must be uppercase DNA bases (A/C/G/T): %s", path, idx, sQuote(df$a1[[idx]])), call. = FALSE)
  }
  bad_a2 <- !grepl("^[ACGT]+$", df$a2)
  if (any(bad_a2)) {
    idx <- which(bad_a2)[[1]]
    stop(sprintf("%s:%d: a2 must be uppercase DNA bases (A/C/G/T): %s", path, idx, sQuote(df$a2[[idx]])), call. = FALSE)
  }
  same_allele <- df$a1 == df$a2
  if (any(same_allele)) {
    stop(sprintf("%s:%d: a1 and a2 must differ", path, which(same_allele)[[1]]), call. = FALSE)
  }
  cm <- suppressWarnings(as.numeric(df$cm))
  if (any(is.na(cm) | !is.finite(cm))) {
    idx <- which(is.na(cm) | !is.finite(cm))[[1]]
    stop(sprintf("%s:%d: cm is not a number: %s", path, idx, sQuote(df$cm[[idx]])), call. = FALSE)
  }
  bp_num <- suppressWarnings(as.numeric(df$bp))
  bad_bp <- is.na(bp_num) | !is.finite(bp_num) | floor(bp_num) != bp_num
  if (any(bad_bp)) {
    idx <- which(bad_bp)[[1]]
    stop(sprintf("%s:%d: bp is not an integer: %s", path, idx, sQuote(df$bp[[idx]])), call. = FALSE)
  }
  df$bp <- as.integer(bp_num)
  df$a1_hash64 <- .allele_hash64(df$a1)
  df$a2_hash64 <- .allele_hash64(df$a2)
  .validate_bim_order_and_duplicates(df[df$chr %in% .canonical_chr_order, , drop = FALSE], path)
  df
}

.validate_bim_order_and_duplicates <- function(df, path) {
  if (!nrow(df)) {
    return(invisible(NULL))
  }
  rank <- unname(.chr_rank[df$chr])
  bad_order <- c(FALSE, rank[-1L] < rank[-length(rank)] | (rank[-1L] == rank[-length(rank)] & df$bp[-1L] < df$bp[-nrow(df)]))
  if (any(bad_order)) {
    stop(sprintf("%s:%d: rows must be sorted by (chr_rank, bp) in canonical contig order", path, which(bad_order)[[1]]), call. = FALSE)
  }

  for (label in unique(df$chr)) {
    keep <- df$chr == label
    .check_unique_variant_keys(
      df$bp[keep],
      df$a1_hash64[keep],
      df$a2_hash64[keep],
      sprintf("%s: shard %s", path, label),
      "variant"
    )
  }
  invisible(NULL)
}

.parse_fam <- function(path) {
  df <- tryCatch(
    utils::read.table(
      path,
      header = FALSE,
      sep = "",
      quote = "",
      comment.char = "",
      stringsAsFactors = FALSE,
      colClasses = rep("character", 6L),
      fill = FALSE
    ),
    error = function(e) {
      stop(sprintf("%s: expected 6 whitespace-delimited columns", path), call. = FALSE)
    }
  )
  if (ncol(df) != 6L) {
    stop(sprintf("%s: expected 6 whitespace-delimited columns, got %d", path, ncol(df)), call. = FALSE)
  }
  names(df) <- c("fid", "iid", "father_id", "mother_id", "sex", "pheno")
  bad_required <- df == ""
  if (any(bad_required)) {
    stop(sprintf("%s:%d: FAM fields must be non-empty; 6 whitespace-delimited columns are required", path, which(rowSums(bad_required) > 0L)[[1]]), call. = FALSE)
  }
  sex_num <- suppressWarnings(as.numeric(df$sex))
  bad_sex <- is.na(sex_num) | !is.finite(sex_num) | floor(sex_num) != sex_num | !(sex_num %in% c(0, 1, 2))
  if (any(bad_sex)) {
    stop(sprintf("%s:%d: FAM sex must be one of PLINK values 1, 2, or 0", path, which(bad_sex)[[1]]), call. = FALSE)
  }
  out <- df[c("fid", "iid", "father_id", "mother_id")]
  out$sex <- as.integer(sex_num)
  dup <- duplicated(paste(out$fid, out$iid, sep = "\r"))
  if (any(dup)) {
    stop(sprintf("%s:%d: duplicate FAM subject pair (fid, iid)", path, which(dup)[[1]]), call. = FALSE)
  }
  out
}

.parse_ploidy <- function(path, source_num_snp, source_chr = NULL) {
  if (is.null(path)) {
    male <- rep.int(2, source_num_snp)
    female <- rep.int(2, source_num_snp)
    if (!is.null(source_chr)) {
      if (length(source_chr) != source_num_snp) {
        stop("source_chr length must match source_num_snp", call. = FALSE)
      }
      male[as.character(source_chr) == "X"] <- 1
    }
    return(list(male = as.numeric(male), female = as.numeric(female)))
  }
  df <- tryCatch(
    utils::read.table(
      path,
      header = FALSE,
      sep = "\t",
      quote = "",
      comment.char = "",
      stringsAsFactors = FALSE,
      colClasses = rep("character", 2L),
      fill = FALSE
    ),
    error = function(e) {
      stop(sprintf("%s: expected 2 tab-delimited columns", path), call. = FALSE)
    }
  )
  if (ncol(df) != 2L) {
    stop(sprintf("%s: expected 2 tab-delimited columns, got %d", path, ncol(df)), call. = FALSE)
  }
  if (nrow(df) != source_num_snp) {
    stop(sprintf("%s: PLOIDY row count mismatch: expected %d, got %d", path, source_num_snp, nrow(df)), call. = FALSE)
  }
  values <- matrix(
    c(
      suppressWarnings(as.numeric(df[[1L]])),
      suppressWarnings(as.numeric(df[[2L]]))
    ),
    ncol = 2L
  )
  bad <- is.na(values) | !is.finite(values) | floor(values) != values | !(values %in% c(0, 1, 2))
  if (any(bad)) {
    stop(sprintf("%s:%d: ploidy values must be integers 0, 1, or 2", path, which(rowSums(bad) > 0L)[[1]]), call. = FALSE)
  }
  list(male = as.numeric(values[, 1]), female = as.numeric(values[, 2]))
}

.expected_bed_size <- function(source_num_sample, source_num_snp) {
  sample_count <- as.numeric(source_num_sample)
  snp_count <- as.numeric(source_num_snp)
  3 + floor((sample_count + 3) / 4) * snp_count
}

.validate_bed_file <- function(path, source_num_sample, source_num_snp) {
  if (!file.exists(path)) {
    stop(sprintf("File not found: %s", path), call. = FALSE)
  }
  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)
  magic <- readBin(con, what = "raw", n = 3L)
  if (length(magic) != 3L || !identical(magic[1:2], as.raw(c(0x6c, 0x1b)))) {
    stop(sprintf("%s: invalid PLINK BED magic bytes", path), call. = FALSE)
  }
  if (!identical(magic[[3]], as.raw(0x01))) {
    stop(sprintf("%s: PLINK BED must be SNP-major mode", path), call. = FALSE)
  }
  size <- file.info(path)$size
  expected <- .expected_bed_size(source_num_sample, source_num_snp)
  if (!identical(as.numeric(size), as.numeric(expected))) {
    stop(sprintf("%s: BED file size mismatch: expected %.0f, got %.0f", path, expected, size), call. = FALSE)
  }
  as.numeric(size)
}

.bed_lookup_int <- local({
  values <- 0:255
  out <- matrix(0L, nrow = 256L, ncol = 4L)
  for (j in 0:3) {
    two_bit <- bitwAnd(bitwShiftR(values, 2L * j), 3L)
    decoded <- integer(256L)
    decoded[two_bit == 0L] <- 2L
    decoded[two_bit == 1L] <- -1L
    decoded[two_bit == 2L] <- 1L
    decoded[two_bit == 3L] <- 0L
    out[, j + 1L] <- decoded
  }
  out
})

.bed_max_read_bytes <- 64 * 1024^2

.read_bed_rows_int8_codes <- function(path, source_rows0, source_num_sample) {
  source_rows0 <- as.integer(source_rows0)
  sample_count <- as.numeric(source_num_sample)
  bytes_per_snp <- floor((sample_count + 3) / 4)
  if (!length(source_rows0)) {
    return(matrix(0L, nrow = source_num_sample, ncol = 0L))
  }
  unique_rows0 <- sort(unique(source_rows0))
  inverse <- match(source_rows0, unique_rows0)
  unique_decoded <- matrix(0L, nrow = source_num_sample, ncol = length(unique_rows0))
  rows_per_read <- if (bytes_per_snp == 0) {
    length(unique_rows0)
  } else {
    max(1L, as.integer(floor(.bed_max_read_bytes / bytes_per_snp)))
  }
  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)

  breaks <- c(1L, which(diff(unique_rows0) != 1L) + 1L, length(unique_rows0) + 1L)
  for (g in seq_len(length(breaks) - 1L)) {
    group_cols <- seq.int(breaks[[g]], breaks[[g + 1L]] - 1L)
    chunk_starts <- seq.int(1L, length(group_cols), by = rows_per_read)
    for (chunk_start in chunk_starts) {
      chunk_stop <- min(chunk_start + rows_per_read - 1L, length(group_cols))
      block_cols <- group_cols[chunk_start:chunk_stop]
      block_rows0 <- unique_rows0[block_cols]
      byte_offset <- 3 + as.numeric(block_rows0[[1]]) * bytes_per_snp
      expected_bytes <- bytes_per_snp * length(block_rows0)
      if (expected_bytes > .Machine$integer.max) {
        stop(sprintf("%s: BED read block is too large: %.0f bytes", path, expected_bytes), call. = FALSE)
      }
      seek(con, where = byte_offset, origin = "start", rw = "read")
      packed <- readBin(con, what = "integer", n = as.integer(expected_bytes), size = 1L, signed = FALSE, endian = "little")
      if (length(packed) != expected_bytes) {
        stop(
          sprintf("%s: short BED read for source SNP rows %d-%.0f: expected %.0f bytes, got %d", path, block_rows0[[1]], block_rows0[[length(block_rows0)]], expected_bytes, length(packed)),
          call. = FALSE
        )
      }
      decoded_flat <- as.vector(t(.bed_lookup_int[packed + 1L, , drop = FALSE]))
      decoded_block <- matrix(decoded_flat, nrow = bytes_per_snp * 4, ncol = length(block_rows0))
      unique_decoded[, block_cols] <- decoded_block[seq_len(sample_count), , drop = FALSE]
    }
  }
  unique_decoded[, inverse, drop = FALSE]
}

.read_ld_npz_shard <- function(path, check_payload_structure = FALSE, retain_ld_r = TRUE, expected_file_md5 = NULL) {
  .require_ld_file(path)
  payload <- .read_npz_ld_payload(path, expected_file_md5 = expected_file_md5)
  meta <- payload$metadata
  .validate_ld_shard_metadata(meta, path, expected_format = .ld_npz_format)
  .validate_csc_payload_dimensions(path, payload, meta)
  if (isTRUE(check_payload_structure)) {
    .validate_csc_payload_structure(path, payload$data, payload$indices, payload$indptr, meta$num_snp, meta$nnz)
  }
  ld_r <- .csc_payload_to_dgCMatrix(payload$data, payload$indices, payload$indptr, meta$num_snp)
  if (isTRUE(check_payload_structure)) {
    .validate_sparse_matrix_values(path, ld_r)
  }
  shard <- .new_ld_shard(
    chr = meta$chr,
    sex = meta$sex,
    num_snp = meta$num_snp,
    ld_r = ld_r,
    a1freq = payload$a1freq,
    reference_checksum = meta$reference_checksum,
    num_monomorphic_snps = meta$num_monomorphic_snps,
    retain_ld_r = retain_ld_r
  )
  if (shard$num_monomorphic_snps > 0L) {
    warning(
      sprintf(
        "%s: LD shard contains %d monomorphic SNPs; LD involving those SNPs is undefined and represented by omitted off-diagonal entries",
        path, shard$num_monomorphic_snps
      ),
      call. = FALSE
    )
  }
  list(shard = shard, metadata = meta)
}

.csc_payload_to_dgCMatrix <- function(data, indices, indptr, num_snp) {
  .new_dgCMatrix_from_slots(
    i = as.integer(indices),
    p = as.integer(indptr),
    x = as.numeric(data),
    dim = as.integer(c(num_snp, num_snp))
  )
}

.new_dgCMatrix_from_slots <- function(i, p, x, dim, dimnames = list(NULL, NULL)) {
  loadNamespace("Matrix")
  methods::new(
    "dgCMatrix",
    i = as.integer(i),
    p = as.integer(p),
    Dim = as.integer(dim),
    Dimnames = dimnames,
    x = as.numeric(x),
    factors = list()
  )
}

.validate_csc_payload_dimensions <- function(path, payload, meta) {
  .validate_numeric_1d(payload$data, path, "data")
  .validate_integer_1d(payload$indices, path, "indices")
  .validate_integer_1d(payload$indptr, path, "indptr")
  .validate_numeric_1d(payload$a1freq, path, "a1freq")
  shape <- as.integer(payload$shape)
  num_snp <- as.integer(meta$num_snp)
  nnz <- as.integer(meta$nnz)
  if (length(shape) != 2L || any(shape != num_snp)) {
    stop(sprintf("%s: shape must equal [num_snp, num_snp]", path), call. = FALSE)
  }
  if (length(payload$indptr) != num_snp + 1L) {
    stop(sprintf("%s: indptr length must be num_snp + 1", path), call. = FALSE)
  }
  if (length(payload$data) != length(payload$indices) || length(payload$data) != nnz) {
    stop(sprintf("%s: data, indices, and metadata nnz length mismatch", path), call. = FALSE)
  }
  if (as.integer(payload$indptr[[length(payload$indptr)]]) != nnz) {
    stop(sprintf("%s: indptr[-1] must equal metadata nnz", path), call. = FALSE)
  }
  if (length(payload$a1freq) != num_snp) {
    stop(sprintf("%s: a1freq length must equal num_snp", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_csc_payload_structure <- function(path, data, indices, indptr, num_snp, nnz) {
  if (!identical(as.integer(indptr[[1L]]), 0L)) {
    stop(sprintf("%s: indptr[0] must be 0", path), call. = FALSE)
  }
  if (length(indptr) > 1L && any(indptr[-1L] < indptr[-length(indptr)])) {
    stop(sprintf("%s: indptr must be monotonic nondecreasing", path), call. = FALSE)
  }
  if (length(indices) && (min(indices) < 0L || max(indices) >= as.integer(num_snp))) {
    stop(sprintf("%s: sparse row indices out of bounds", path), call. = FALSE)
  }
  if (any(!is.finite(data))) {
    stop(sprintf("%s: sparse data must be finite", path), call. = FALSE)
  }
  if (as.integer(nnz) < as.integer(num_snp)) {
    stop(sprintf("%s: sparse matrix must include explicit unit diagonal", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_sparse_matrix_values <- function(path, mat) {
  diag <- Matrix::diag(mat)
  if (length(diag) != nrow(mat) || any(abs(diag - 1) > 1e-7)) {
    stop(sprintf("%s: ld_r diagonal must be explicit unit", path), call. = FALSE)
  }
  diff <- mat - Matrix::t(mat)
  if (length(diff@x) && max(abs(diff@x)) > 1e-6) {
    stop(sprintf("%s: ld_r must be symmetric", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_numeric_1d <- function(value, path, name) {
  if (!is.numeric(value) || !is.null(dim(value))) {
    stop(sprintf("%s: %s must be a one-dimensional numeric vector", path, name), call. = FALSE)
  }
  invisible(NULL)
}

.validate_integer_1d <- function(value, path, name) {
  if (!is.integer(value) || !is.null(dim(value))) {
    stop(sprintf("%s: %s must be a one-dimensional integer vector", path, name), call. = FALSE)
  }
  invisible(NULL)
}

.statgen_state <- new.env(parent = emptyenv())
.statgen_state$verbosity <- "info"

set_verbosity <- function(level) {
  if (!is.character(level) || length(level) != 1L || is.na(level)) {
    stop("level must be a character scalar", call. = FALSE)
  }
  if (!(level %in% c("quiet", "info"))) {
    stop("level must be one of 'quiet' or 'info'", call. = FALSE)
  }
  .statgen_state$verbosity <- level
  invisible(level)
}

get_verbosity <- function() {
  .statgen_state$verbosity
}

test_that("package load exposes version and verbosity controls", {
  expect_identical(version(), as.character(utils::packageVersion("statgen")))
  expect_identical(get_verbosity(), "info")

  expect_invisible(set_verbosity("quiet"))
  expect_identical(get_verbosity(), "quiet")

  expect_invisible(set_verbosity("info"))
  expect_identical(get_verbosity(), "info")

  expect_error(set_verbosity("verbose"), "quiet.*info")
})

.canonical_chr_order <- c(as.character(seq_len(22L)), "X")
.chr_rank <- stats::setNames(seq_along(.canonical_chr_order), .canonical_chr_order)
.ignored_chr <- c("Y", "MT")

is_present <- function(x, ...) UseMethod("is_present")

.validate_requested_shards <- function(requested, available_labels, where) {
  available_labels <- as.character(available_labels)
  if (is.null(requested)) {
    if (!length(available_labels)) {
      stop(sprintf("%s: no supported shards are present", where), call. = FALSE)
    }
    return(available_labels)
  }
  if (!is.character(requested) || !length(requested) || anyNA(requested)) {
    stop(sprintf("%s: shards must be a non-empty character vector of unique canonical contig labels", where), call. = FALSE)
  }
  seen <- character()
  prev_rank <- 0L
  for (label in requested) {
    if (!(label %in% names(.chr_rank))) {
      stop(sprintf("%s: unsupported shard label %s; expected canonical labels 1-22 or X", where, sQuote(label)), call. = FALSE)
    }
    if (label %in% seen) {
      stop(sprintf("%s: duplicate shard label %s in shards", where, sQuote(label)), call. = FALSE)
    }
    rank <- .chr_rank[[label]]
    if (rank <= prev_rank) {
      stop(sprintf("%s: shards must be in canonical subsequence order", where), call. = FALSE)
    }
    if (!(label %in% available_labels)) {
      stop(sprintf("%s: requested shard %s is not present", where, sQuote(label)), call. = FALSE)
    }
    seen <- c(seen, label)
    prev_rank <- rank
  }
  requested
}

.validate_char_scalar <- function(value, name) {
  if (!is.character(value) || length(value) != 1L || is.na(value)) {
    stop(sprintf("%s must be a character scalar", name), call. = FALSE)
  }
  value
}

.validate_path_scalar <- function(path, name) {
  .validate_char_scalar(path, name)
}

.validate_equal_lengths <- function(values, message) {
  lengths <- vapply(values, length, integer(1))
  if (length(unique(lengths)) != 1L) {
    stop(message, call. = FALSE)
  }
}

.validate_cache_payload_metadata <- function(payload, schema, cache_name, extra_fields = character()) {
  if (!is.list(payload) || is.null(payload$metadata)) {
    stop(sprintf("Invalid %s cache: expected an RDS list with metadata", cache_name), call. = FALSE)
  }
  meta <- payload$metadata
  if (!identical(meta$schema, schema)) {
    stop(sprintf("Unsupported %s cache schema: %s", cache_name, sQuote(as.character(meta$schema))), call. = FALSE)
  }
  n_shards <- suppressWarnings(as.integer(meta$n_shards))
  if (length(n_shards) != 1L || is.na(n_shards) || n_shards < 1L) {
    stop(sprintf("Invalid %s cache: n_shards must be at least 1", cache_name), call. = FALSE)
  }
  for (field in c("shard_labels", "shard_checksums", "shard_start0", "shard_stop0", extra_fields)) {
    if (is.null(meta[[field]]) || length(meta[[field]]) != n_shards) {
      stop(sprintf("Invalid %s cache: metadata.%s length mismatch", cache_name, field), call. = FALSE)
    }
  }
  total <- meta$shard_stop0[[n_shards]]
  if (!identical(as.integer(meta$shard_start0[[1]]), 0L) || any(meta$shard_stop0 < meta$shard_start0)) {
    stop(sprintf("Invalid %s cache: shard offsets are invalid", cache_name), call. = FALSE)
  }
  if (n_shards > 1L && any(meta$shard_start0[-1L] != meta$shard_stop0[-n_shards])) {
    stop(sprintf("Invalid %s cache: shard offsets are not contiguous", cache_name), call. = FALSE)
  }
  list(meta = meta, n_shards = n_shards, total = total)
}

.annotations_cache_schema <- "annotations_cache/0.1"

annomat <- function(x, ...) UseMethod("annomat")
annonames <- function(x, ...) UseMethod("annonames")
num_annot <- function(x, ...) UseMethod("num_annot")
select_annotations <- function(x, names, ...) UseMethod("select_annotations")
union_annotations <- function(x, other, mode = "by_name", ...) UseMethod("union_annotations")

load_annotations <- function(bed_paths, reference) {
  if (!inherits(reference, "ReferencePanel")) {
    stop("reference must be a ReferencePanel", call. = FALSE)
  }
  paths <- .coerce_bed_paths(bed_paths)
  for (path in paths) {
    if (!file.exists(path)) {
      stop(sprintf("BED file not found: %s", path), call. = FALSE)
    }
  }
  names <- .annotation_names_from_paths(paths)
  if (anyDuplicated(names)) {
    stop("duplicate annotation names derived from BED basenames", call. = FALSE)
  }

  columns <- lapply(paths, function(path) {
    Matrix::Matrix(.paint_bed_to_reference(.parse_bed(path), reference), ncol = 1L, sparse = TRUE)
  })
  create_annotations(reference, annotation_matrix = do.call(cbind, columns), annotation_names = names)
}

create_annotation <- function(reference, annovec, annoname) {
  name <- as.character(annoname)
  if (length(name) != 1L || is.na(name) || identical(name, "")) {
    stop("annoname must be a non-empty character scalar", call. = FALSE)
  }
  vec <- .coerce_annovec(annovec, num_snp(reference))
  create_annotations(reference, annotation_matrix = Matrix::Matrix(vec, ncol = 1L, sparse = TRUE), annotation_names = name)
}

create_annotations <- function(reference, annotation_matrix, annotation_names) {
  if (!inherits(reference, "ReferencePanel")) {
    stop("reference must be a ReferencePanel", call. = FALSE)
  }
  names <- .coerce_annonames(annotation_names)
  mat <- .as_lgC_binary_matrix(annotation_matrix, "annotation_matrix")
  expected <- c(num_snp(reference), length(names))
  if (!identical(as.integer(dim(mat)), as.integer(expected))) {
    stop(sprintf(
      "annotation_matrix shape mismatch: expected (%d, %d), got (%d, %d)",
      expected[[1]], expected[[2]], dim(mat)[[1]], dim(mat)[[2]]
    ), call. = FALSE)
  }

  out <- vector("list", length(shards(reference)))
  offsets <- shard_offsets(reference)
  for (i in seq_along(shards(reference))) {
    ref_shard <- shards(reference)[[i]]
    start <- offsets$start0[[i]] + 1L
    stop <- offsets$stop0[[i]]
    out[[i]] <- .new_annotation_shard(
      ref_shard$label,
      ref_shard$checksum,
      mat[start:stop, , drop = FALSE]
    )
  }
  .new_annotation_panel(out, names)
}

save_annotations_cache <- function(panel, path) {
  if (!inherits(panel, "AnnotationPanel")) {
    stop("panel must be an AnnotationPanel", call. = FALSE)
  }
  path <- .validate_path_scalar(path, "path")
  offsets <- shard_offsets(panel)
  payload <- list(
    metadata = list(
      schema = .annotations_cache_schema,
      n_shards = length(panel$shards),
      shard_labels = vapply(panel$shards, function(s) s$label, character(1)),
      shard_checksums = vapply(panel$shards, function(s) s$reference_checksum, character(1)),
      shard_start0 = offsets$start0,
      shard_stop0 = offsets$stop0
    ),
    annomat = annomat(panel),
    annonames = annonames(panel)
  )
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  saveRDS(payload, path)
  invisible(NULL)
}

load_annotations_cache <- function(path, shards = NULL) {
  path <- .validate_path_scalar(path, "path")
  payload <- readRDS(path)
  cache <- .validate_annotations_cache_payload(payload)
  meta <- payload$metadata
  selected <- .validate_requested_shards(shards, meta$shard_labels, "load_annotations_cache")
  label_to_index <- stats::setNames(seq_along(meta$shard_labels), meta$shard_labels)
  names <- .coerce_annonames(payload$annonames)

  out <- vector("list", length(selected))
  for (i in seq_along(selected)) {
    label <- selected[[i]]
    idx <- label_to_index[[label]]
    start <- meta$shard_start0[[idx]] + 1L
    stop <- meta$shard_stop0[[idx]]
    out[[i]] <- .new_annotation_shard(
      label,
      meta$shard_checksums[[idx]],
      cache$annomat[start:stop, , drop = FALSE]
    )
  }
  .new_annotation_panel(out, names)
}

num_snp.AnnotationShard <- function(x, ...) x$num_snp
num_snp.AnnotationPanel <- function(x, ...) x$num_snp
num_annot.AnnotationShard <- function(x, ...) x$num_annot
num_annot.AnnotationPanel <- function(x, ...) x$num_annot
shards.AnnotationPanel <- function(x, ...) x$shards
shard_offsets.AnnotationPanel <- function(x, ...) x$shard_offsets
annomat.AnnotationShard <- function(x, ...) x$annomat
annomat.AnnotationPanel <- function(x, ...) {
  mats <- lapply(x$shards, annomat)
  mat <- do.call(rbind, mats)
  .set_annotation_colnames(mat, x$annonames)
}
annonames.AnnotationPanel <- function(x, ...) x$annonames

select_shards.AnnotationPanel <- function(x, shards, ...) {
  available <- vapply(x$shards, function(s) s$label, character(1))
  selected <- .validate_requested_shards(shards, available, "AnnotationPanel.select_shards")
  by_label <- stats::setNames(x$shards, available)
  .new_annotation_panel(unname(by_label[selected]), x$annonames)
}

select_annotations.AnnotationPanel <- function(x, names, ...) {
  requested <- .coerce_annonames(names)
  idx <- match(requested, x$annonames)
  missing <- requested[is.na(idx)]
  if (length(missing)) {
    stop(sprintf("unknown annotation name(s): %s", paste(missing, collapse = ", ")), call. = FALSE)
  }
  out <- lapply(x$shards, function(s) {
    .new_annotation_shard(s$label, s$reference_checksum, s$annomat[, idx, drop = FALSE])
  })
  .new_annotation_panel(out, requested)
}

union_annotations.AnnotationPanel <- function(x, other, mode = "by_name", ...) {
  if (!identical(mode, "by_name")) {
    stop("union_annotations supports only mode = 'by_name'", call. = FALSE)
  }
  if (!inherits(other, "AnnotationPanel")) {
    stop("other must be an AnnotationPanel", call. = FALSE)
  }
  overlap <- intersect(x$annonames, other$annonames)
  if (length(overlap)) {
    stop(sprintf("annotation name collision(s): %s", paste(overlap, collapse = ", ")), call. = FALSE)
  }
  .check_annotation_reference_alignment(x, other, "union_annotations")
  out <- vector("list", length(x$shards))
  for (i in seq_along(x$shards)) {
    a <- x$shards[[i]]
    b <- other$shards[[i]]
    out[[i]] <- .new_annotation_shard(
      a$label,
      a$reference_checksum,
      do.call(cbind, list(a$annomat, b$annomat))
    )
  }
  .new_annotation_panel(out, c(x$annonames, other$annonames))
}

.check_annotation_reference_alignment <- function(x, other, context) {
  if (length(x$shards) != length(other$shards)) {
    stop(sprintf(
      "%s requires compatible reference alignment: shard count mismatch",
      context
    ), call. = FALSE)
  }
  for (i in seq_along(x$shards)) {
    a <- x$shards[[i]]
    b <- other$shards[[i]]
    if (!identical(a$label, b$label)) {
      stop(sprintf(
        "%s requires compatible reference alignment: shard label mismatch: %s vs %s",
        context, sQuote(a$label), sQuote(b$label)
      ), call. = FALSE)
    }
    if (!identical(as.integer(a$num_snp), as.integer(b$num_snp))) {
      stop(sprintf(
        "%s requires compatible reference alignment: shard %s row count mismatch",
        context, sQuote(a$label)
      ), call. = FALSE)
    }
    if (!identical(a$reference_checksum, b$reference_checksum)) {
      stop(sprintf(
        "%s requires compatible reference alignment: shard %s reference_checksum mismatch",
        context, sQuote(a$label)
      ), call. = FALSE)
    }
  }
  invisible(NULL)
}

save_cache.AnnotationPanel <- function(x, path, ...) save_annotations_cache(x, path)

print.AnnotationPanel <- function(x, ...) {
  labels <- vapply(x$shards, function(s) s$label, character(1))
  cat(sprintf(
    "<AnnotationPanel: %d SNPs x %d annotations across %d shard(s): %s>\n",
    x$num_snp, x$num_annot, length(labels), paste(labels, collapse = ", ")
  ))
  invisible(x)
}

.new_annotation_shard <- function(label, reference_checksum, mat) {
  mat <- .as_lgC_binary_matrix(mat)
  structure(
    list(
      label = as.character(label),
      reference_checksum = as.character(reference_checksum),
      num_snp = as.integer(dim(mat)[[1]]),
      num_annot = as.integer(dim(mat)[[2]]),
      annomat = mat
    ),
    class = "AnnotationShard"
  )
}

.new_annotation_panel <- function(shard_list, names) {
  if (!length(shard_list)) {
    stop("AnnotationPanel requires at least one shard", call. = FALSE)
  }
  names <- .coerce_annonames(names)
  pos <- 0L
  offsets <- data.frame(
    shard_label = character(length(shard_list)),
    start0 = integer(length(shard_list)),
    stop0 = integer(length(shard_list)),
    stringsAsFactors = FALSE
  )
  for (i in seq_along(shard_list)) {
    if (!identical(as.integer(shard_list[[i]]$num_annot), as.integer(length(names)))) {
      stop("all annotation shards must share the same annotation columns", call. = FALSE)
    }
    offsets$shard_label[[i]] <- shard_list[[i]]$label
    offsets$start0[[i]] <- pos
    pos <- pos + shard_list[[i]]$num_snp
    offsets$stop0[[i]] <- pos
  }
  structure(
    list(
      shards = shard_list,
      annonames = names,
      num_snp = pos,
      num_annot = length(names),
      shard_offsets = offsets
    ),
    class = "AnnotationPanel"
  )
}

.set_annotation_colnames <- function(mat, names) {
  colnames(mat) <- names
  mat
}

.coerce_bed_paths <- function(bed_paths) {
  if (!is.character(bed_paths) || !length(bed_paths) || anyNA(bed_paths) || any(bed_paths == "")) {
    stop("bed_paths must be a non-empty character vector of BED files", call. = FALSE)
  }
  if (is.character(bed_paths) && length(bed_paths) == 1L) {
    return(bed_paths)
  }
  as.character(bed_paths)
}

.annotation_names_from_paths <- function(paths) {
  tools::file_path_sans_ext(basename(paths))
}

.coerce_annonames <- function(names) {
  if (!is.character(names) || !length(names) || anyNA(names) || any(names == "")) {
    stop("annonames must be a non-empty character vector of unique strings", call. = FALSE)
  }
  if (anyDuplicated(names)) {
    stop("annonames must be unique", call. = FALSE)
  }
  as.character(names)
}

.coerce_annovec <- function(annovec, n) {
  vec <- as.vector(annovec)
  if (length(vec) != n) {
    stop(sprintf("annovec length mismatch: expected %d, got %d", n, length(vec)), call. = FALSE)
  }
  if (any(is.na(vec)) || any(!(vec %in% c(FALSE, TRUE, 0, 1)))) {
    stop("annovec must be binary (0/1)", call. = FALSE)
  }
  as.logical(vec)
}

.as_lgC_binary_matrix <- function(mat, name = "annomat") {
  if (inherits(mat, "lgCMatrix")) {
    return(mat)
  }
  if (inherits(mat, "Matrix")) {
    dims <- dim(mat)
    entries <- Matrix::summary(mat)
    if (nrow(entries)) {
      nonzero <- entries$x != 0
      bad <- !(entries$x[nonzero] %in% c(TRUE, 1))
      if (any(bad)) {
        stop(sprintf("%s contains non-binary values", name), call. = FALSE)
      }
      return(Matrix::sparseMatrix(
        i = entries$i[nonzero],
        j = entries$j[nonzero],
        x = rep.int(TRUE, sum(nonzero)),
        dims = dims
      ))
    }
    return(Matrix::sparseMatrix(i = integer(), j = integer(), x = logical(), dims = dims))
  }
  arr <- as.matrix(mat)
  if (length(dim(arr)) != 2L) {
    stop(sprintf("%s must be a 2D matrix", name), call. = FALSE)
  }
  if (any(is.na(arr)) || any(!(arr %in% c(FALSE, TRUE, 0, 1)))) {
    stop(sprintf("%s contains non-binary values", name), call. = FALSE)
  }
  Matrix::Matrix(arr != 0, sparse = TRUE)
}

.parse_bed <- function(path) {
  # First pass: scan line-by-line to validate structure and count header lines to
  # skip. Avoids seek(), which R documents as unreliable on Windows text connections.
  con <- file(path, open = "rt")
  on.exit(close(con), add = TRUE)

  skip_count <- 0L
  first_data <- NULL
  found_data <- FALSE

  repeat {
    line <- readLines(con, n = 1L, warn = FALSE)
    if (!length(line)) {
      if (!found_data) {
        stop(sprintf("%s: BED file is empty", path), call. = FALSE)
      }
      break
    }
    if (!found_data) {
      if (identical(line, "") || startsWith(line, "#")) {
        skip_count <- skip_count + 1L
      } else {
        first_data <- line
        found_data <- TRUE
      }
    } else {
      if (identical(line, "") || startsWith(line, "#")) {
        stop(sprintf("%s: blank or comment line after BED data row", path), call. = FALSE)
      }
    }
  }
  close(con)
  on.exit(NULL)

  if (length(strsplit(first_data, "\t", fixed = TRUE)[[1L]]) < 3L) {
    stop(sprintf("%s: BED must have at least 3 tab-separated columns", path), call. = FALSE)
  }

  # Second pass: read.table opens the file fresh, no seek() needed.
  df <- tryCatch(
    utils::read.table(
      file = path,
      skip = skip_count,
      header = FALSE,
      sep = "\t",
      quote = "",
      comment.char = "",
      stringsAsFactors = FALSE,
      colClasses = "character",
      fill = FALSE
    ),
    error = function(e) {
      stop(sprintf("%s: malformed BED", path), call. = FALSE)
    }
  )
  if (ncol(df) < 3L) {
    stop(sprintf("%s: BED must have at least 3 tab-separated columns", path), call. = FALSE)
  }
  chr <- df[[1]]
  bad_chr <- is.na(chr) | chr == ""
  if (any(bad_chr)) {
    stop(sprintf("%s: row %d: chromosome label must be non-empty", path, which(bad_chr)[[1]]), call. = FALSE)
  }
  starts <- suppressWarnings(as.numeric(df[[2]]))
  ends <- suppressWarnings(as.numeric(df[[3]]))
  bad_start <- is.na(starts) | !is.finite(starts) | floor(starts) != starts | starts < 0
  if (any(bad_start)) {
    stop(sprintf("%s: row %d: BED start must be a non-negative integer", path, which(bad_start)[[1]]), call. = FALSE)
  }
  bad_end <- is.na(ends) | !is.finite(ends) | floor(ends) != ends | ends < 0
  if (any(bad_end)) {
    stop(sprintf("%s: row %d: BED end must be a non-negative integer", path, which(bad_end)[[1]]), call. = FALSE)
  }
  if (any(ends < starts)) {
    stop(sprintf("%s: row %d: BED interval end must be >= start", path, which(ends < starts)[[1]]), call. = FALSE)
  }
  split_idx <- split(seq_along(chr), chr)
  lapply(split_idx, function(idx) .merge_intervals(as.integer(starts[idx]), as.integer(ends[idx])))
}

.merge_intervals <- function(starts, ends) {
  if (!length(starts)) {
    return(matrix(integer(), ncol = 2L, dimnames = list(NULL, c("start", "end"))))
  }
  ord <- order(starts, ends)
  starts <- starts[ord]
  ends <- ends[ord]

  running_max_end <- cummax(ends)
  new_group <- c(TRUE, starts[-1L] > running_max_end[-length(running_max_end)])
  group_id <- cumsum(new_group)

  cbind(
    start = starts[new_group],
    end = as.integer(unname(tapply(ends, group_id, max)))
  )
}

.paint_bed_to_reference <- function(intervals_by_chr, reference) {
  out <- rep.int(FALSE, num_snp(reference))
  offsets <- shard_offsets(reference)
  for (i in seq_along(shards(reference))) {
    ref_shard <- shards(reference)[[i]]
    intervals <- intervals_by_chr[[ref_shard$label]]
    if (is.null(intervals) || !nrow(intervals)) {
      next
    }
    start <- offsets$start0[[i]] + 1L
    stop <- offsets$stop0[[i]]
    out[start:stop] <- .paint_mask(bp(ref_shard), intervals)
  }
  out
}

.paint_mask <- function(bp, intervals) {
  pos0 <- as.integer(bp) - 1L
  idx <- findInterval(pos0, intervals[, "start"])
  valid <- idx > 0L
  out <- rep.int(FALSE, length(bp))
  out[valid] <- pos0[valid] < intervals[idx[valid], "end"]
  out
}

.validate_annotations_cache_payload <- function(payload) {
  cache <- .validate_cache_payload_metadata(payload, .annotations_cache_schema, "annotations")
  total <- cache$total
  if (is.null(payload$annomat) || is.null(payload$annonames)) {
    stop("Invalid annotations cache: missing annomat or annonames", call. = FALSE)
  }
  annomat <- .as_lgC_binary_matrix(payload$annomat)
  if (!identical(as.integer(dim(annomat)[[1]]), as.integer(total))) {
    stop("Invalid annotations cache: annomat row count mismatch", call. = FALSE)
  }
  if (!identical(as.integer(dim(annomat)[[2]]), as.integer(length(payload$annonames)))) {
    stop("Invalid annotations cache: annomat column count mismatch", call. = FALSE)
  }
  list(annomat = annomat)
}

.ld_manifest_schema <- "1.0"
.ld_runtime_format <- "python_npz_csc32"
.ld_npz_format <- "statgen_ld_npz_csc32"
.ld_valid_chrx_sex <- c("female", "male", "combined")
.ld_required_arrays <- c("data", "indices", "indptr", "shape", "a1freq", "metadata")

.read_ld_manifest <- function(path, require_r_reference = FALSE) {
  .require_ld_file(path)
  manifest <- jsonlite::fromJSON(path, simplifyVector = FALSE)
  if (!identical(manifest$object_type, "ld_panel_manifest")) {
    stop(sprintf("%s: object_type must be 'ld_panel_manifest'", path), call. = FALSE)
  }
  if (!identical(manifest$schema_version, .ld_manifest_schema)) {
    stop(sprintf("%s: unsupported LD manifest schema_version %s", path, sQuote(as.character(manifest$schema_version))), call. = FALSE)
  }
  if (!identical(manifest$runtime_format, .ld_runtime_format)) {
    stop(sprintf("%s: unsupported LD runtime_format %s", path, sQuote(as.character(manifest$runtime_format))), call. = FALSE)
  }
  .validate_plain_relative_filename(manifest$reference_cache, sprintf("%s: reference_cache", path))
  .validate_md5_hex(manifest$reference_cache_md5, sprintf("%s: reference_cache_md5", path))
  has_r_reference <- !is.null(manifest$r_reference_cache)
  has_r_md5 <- !is.null(manifest$r_reference_cache_md5)
  if (xor(has_r_reference, has_r_md5)) {
    stop(sprintf("%s: r_reference_cache and r_reference_cache_md5 must be present together", path), call. = FALSE)
  }
  if (isTRUE(require_r_reference) && !has_r_reference) {
    stop(sprintf("%s: missing r_reference_cache/r_reference_cache_md5 for R LD loading", path), call. = FALSE)
  }
  if (has_r_reference) {
    .validate_plain_relative_filename(manifest$r_reference_cache, sprintf("%s: r_reference_cache", path))
    .validate_md5_hex(manifest$r_reference_cache_md5, sprintf("%s: r_reference_cache_md5", path))
  }
  if (!is.list(manifest$shards) || !length(manifest$shards)) {
    stop(sprintf("%s: shards must be a non-empty list", path), call. = FALSE)
  }
  seen <- character()
  reference_identity_by_chr <- list()
  for (i in seq_along(manifest$shards)) {
    entry <- manifest$shards[[i]]
    .validate_ld_manifest_entry(entry, sprintf("%s:shards[%d]", path, i - 1L))
    key <- paste(entry$chr, if (is.null(entry$sex)) "<null>" else entry$sex, sep = "\r")
    if (key %in% seen) {
      stop(sprintf("%s: duplicate manifest entry for chr %s sex %s", path, entry$chr, sQuote(as.character(entry$sex))), call. = FALSE)
    }
    seen <- c(seen, key)
    identity <- list(
      reference_bim = entry$reference_bim,
      num_snp = as.integer(entry$num_snp),
      reference_checksum = entry$reference_checksum
    )
    prior <- reference_identity_by_chr[[entry$chr]]
    if (is.null(prior)) {
      reference_identity_by_chr[[entry$chr]] <- identity
    } else if (!identical(prior, identity)) {
      stop(
        sprintf(
          "%s: manifest entries for chr %s must share reference_bim, num_snp, and reference_checksum",
          path, entry$chr
        ),
        call. = FALSE
      )
    }
  }
  manifest
}

.validate_ld_manifest_entry <- function(entry, where) {
  required <- c("chr", "sex", "file", "file_md5", "num_snp", "nnz", "reference_checksum", "reference_bim")
  missing <- setdiff(required, names(entry))
  if (length(missing)) {
    stop(sprintf("%s: missing required fields: %s", where, paste(missing, collapse = ", ")), call. = FALSE)
  }
  .validate_chr_sex(entry$chr, entry$sex, where)
  .validate_relative_path(entry$file, sprintf("%s: file", where))
  .validate_md5_hex(entry$file_md5, sprintf("%s: file_md5", where))
  .validate_positive_int(entry$num_snp, sprintf("%s: num_snp", where))
  .validate_nonnegative_int(entry$nnz, sprintf("%s: nnz", where))
  if (!is.character(entry$reference_checksum) || length(entry$reference_checksum) != 1L || identical(entry$reference_checksum, "")) {
    stop(sprintf("%s: reference_checksum must be a non-empty string", where), call. = FALSE)
  }
  .validate_plain_relative_filename(entry$reference_bim, sprintf("%s: reference_bim", where))
  invisible(NULL)
}

.validate_ld_shard_metadata <- function(meta, path, expected_format) {
  required <- c(
    "object_type", "schema_version", "format", "chr", "sex", "num_snp", "nnz",
    "matrix", "diagonal", "value", "reference_checksum", "reference_bim",
    "num_monomorphic_snps"
  )
  missing <- setdiff(required, names(meta))
  if (length(missing)) {
    stop(sprintf("%s: metadata missing required fields: %s", path, paste(missing, collapse = ", ")), call. = FALSE)
  }
  if (!identical(meta$object_type, "ld_shard")) {
    stop(sprintf("%s: metadata object_type must be 'ld_shard'", path), call. = FALSE)
  }
  if (!identical(meta$schema_version, .ld_manifest_schema)) {
    stop(sprintf("%s: unsupported metadata schema_version %s", path, sQuote(as.character(meta$schema_version))), call. = FALSE)
  }
  if (!identical(meta$format, expected_format)) {
    stop(sprintf("%s: metadata format must be %s", path, sQuote(expected_format)), call. = FALSE)
  }
  .validate_chr_sex(meta$chr, meta$sex, sprintf("%s: metadata", path))
  .validate_positive_int(meta$num_snp, sprintf("%s: metadata num_snp", path))
  .validate_nonnegative_int(meta$nnz, sprintf("%s: metadata nnz", path))
  .validate_nonnegative_int(meta$num_monomorphic_snps, sprintf("%s: metadata num_monomorphic_snps", path))
  if (as.integer(meta$num_monomorphic_snps) > as.integer(meta$num_snp)) {
    stop(sprintf("%s: metadata num_monomorphic_snps must not exceed num_snp", path), call. = FALSE)
  }
  if (as.numeric(meta$nnz) >= 2^31) {
    stop(sprintf("%s: CSC32 metadata nnz must be < 2^31", path), call. = FALSE)
  }
  if (!identical(meta$matrix, "symmetric") || !identical(meta$diagonal, "explicit_unit") || !identical(meta$value, "r")) {
    stop(sprintf("%s: metadata matrix/diagonal/value fields are invalid", path), call. = FALSE)
  }
  if (!is.character(meta$reference_checksum) || length(meta$reference_checksum) != 1L || identical(meta$reference_checksum, "")) {
    stop(sprintf("%s: metadata reference_checksum must be a non-empty string", path), call. = FALSE)
  }
  .validate_plain_relative_filename(meta$reference_bim, sprintf("%s: metadata reference_bim", path))
  if (!is.null(meta$sparse_layout) && !identical(meta$sparse_layout, "csc")) {
    stop(sprintf("%s: metadata sparse_layout must be 'csc'", path), call. = FALSE)
  }
  if (!is.null(meta$index_base) && !identical(as.integer(meta$index_base), 0L)) {
    stop(sprintf("%s: metadata index_base must be 0", path), call. = FALSE)
  }
  invisible(NULL)
}

.validate_manifest_entry_agreement <- function(entry, meta, path) {
  for (key in c("chr", "sex", "num_snp", "nnz", "reference_checksum", "reference_bim")) {
    if (!identical(entry[[key]], meta[[key]])) {
      stop(sprintf("%s: manifest/per-file metadata mismatch for %s", path, key), call. = FALSE)
    }
  }
  invisible(NULL)
}

.validate_chr_sex <- function(chr_label, sex, where) {
  if (!is.character(chr_label) || length(chr_label) != 1L || !(chr_label %in% .canonical_chr_order)) {
    stop(sprintf("%s: chr must be one of 1-22 or X", where), call. = FALSE)
  }
  if (identical(chr_label, "X")) {
    .validate_chrx_sex(sex, sprintf("%s: sex", where))
  } else if (!is.null(sex)) {
    stop(sprintf("%s: autosomal LD shards must have sex null", where), call. = FALSE)
  }
  invisible(NULL)
}

.validate_chrx_sex <- function(sex, where) {
  if (!is.character(sex) || length(sex) != 1L || is.na(sex) || !(sex %in% .ld_valid_chrx_sex)) {
    stop(sprintf("%s: chrX sex must be one of female, male, combined", where), call. = FALSE)
  }
  sex
}

.validate_positive_int <- function(value, where) {
  value_num <- suppressWarnings(as.numeric(value))
  if (length(value_num) != 1L || is.na(value_num) || !is.finite(value_num) || floor(value_num) != value_num || value_num <= 0) {
    stop(sprintf("%s must be a positive integer", where), call. = FALSE)
  }
  invisible(NULL)
}

.validate_nonnegative_int <- function(value, where) {
  value_num <- suppressWarnings(as.numeric(value))
  if (length(value_num) != 1L || is.na(value_num) || !is.finite(value_num) || floor(value_num) != value_num || value_num < 0) {
    stop(sprintf("%s must be a non-negative integer", where), call. = FALSE)
  }
  invisible(NULL)
}

.validate_plain_relative_filename <- function(value, where) {
  if (!is.character(value) || length(value) != 1L || is.na(value) || identical(value, "")) {
    stop(sprintf("%s must be a non-empty filename", where), call. = FALSE)
  }
  if (grepl("[/\\\\]", value) || grepl("(^|[.])[.]", value) || value %in% c(".", "..")) {
    stop(sprintf("%s must be a plain relative filename", where), call. = FALSE)
  }
  value
}

.validate_relative_path <- function(value, where) {
  if (!is.character(value) || length(value) != 1L || is.na(value) || identical(value, "")) {
    stop(sprintf("%s must be a non-empty relative path", where), call. = FALSE)
  }
  if (grepl("^(/|[A-Za-z]:)", value) || any(strsplit(value, "[/\\\\]")[[1L]] == "..")) {
    stop(sprintf("%s must be a relative path inside the LD distribution", where), call. = FALSE)
  }
  value
}

.validate_md5_hex <- function(value, where) {
  if (!is.character(value) || length(value) != 1L || !grepl("^[0-9a-f]{32}$", value)) {
    stop(sprintf("%s must be a lowercase MD5 hex string", where), call. = FALSE)
  }
  invisible(NULL)
}

.validate_ld_root <- function(path, where) {
  root <- .validate_path_scalar(path, "path")
  if (!dir.exists(root)) {
    stop(sprintf("%s: path must identify a panel root directory", where), call. = FALSE)
  }
  root
}

.require_ld_file <- function(path) {
  if (!file.exists(path) || dir.exists(path)) {
    stop(sprintf("LD file not found: %s", path), call. = FALSE)
  }
  invisible(NULL)
}

.md5_file <- function(path) {
  digest::digest(path, algo = "md5", file = TRUE)
}

.write_ld_manifest <- function(manifest, path) {
  text <- jsonlite::toJSON(manifest, auto_unbox = TRUE, null = "null", pretty = TRUE)
  tmp <- tempfile(".ld_manifest_", tmpdir = dirname(path))
  on.exit(unlink(tmp), add = TRUE)
  writeLines(text, tmp, useBytes = TRUE)
  if (!file.rename(tmp, path)) {
    stop(sprintf("Failed to replace LD manifest: %s", path), call. = FALSE)
  }
  invisible(NULL)
}

reference <- function(x, ...) UseMethod("reference")
default_chrX_sex <- function(x, ...) UseMethod("default_chrX_sex")
a1freq <- function(x, chrX_sex = NULL, ...) UseMethod("a1freq")
multiply_r2 <- function(ld_panel, M, chrX_sex = NULL, ...) UseMethod("multiply_r2")

load_ld <- function(path, shards = NULL, default_chrX_sex = NULL, retain_ld_r = TRUE) {
  root <- .validate_ld_root(path, "load_ld")
  manifest <- .read_ld_manifest(file.path(root, "ld_manifest.json"), require_r_reference = TRUE)
  default_chrX_sex <- .validate_chrx_sex(
    if (is.null(default_chrX_sex)) "female" else default_chrX_sex,
    "default_chrX_sex"
  )
  reference_panel <- .load_ld_manifest_reference(root, manifest, shards)
  ref_shards <- shards(reference_panel)
  groups <- .load_ld_shard_groups(root, manifest, ref_shards, retain_ld_r = isTRUE(retain_ld_r))
  .new_ld_panel(groups, default_chrX_sex = default_chrX_sex, reference = reference_panel)
}

load_ld_reference <- function(path, shards = NULL) {
  root <- .validate_ld_root(path, "load_ld_reference")
  manifest <- .read_ld_manifest(file.path(root, "ld_manifest.json"), require_r_reference = TRUE)
  .load_ld_manifest_reference(root, manifest, shards)
}

validate_ld_distribution <- function(path, check_payload_structure = FALSE) {
  root <- .validate_ld_root(path, "validate_ld_distribution")
  manifest <- .read_ld_manifest(file.path(root, "ld_manifest.json"), require_r_reference = TRUE)
  python_reference_cache_path <- file.path(root, manifest$reference_cache)
  .require_ld_file(python_reference_cache_path)
  if (!identical(.md5_file(python_reference_cache_path), manifest$reference_cache_md5)) {
    stop(sprintf("%s: reference_cache_md5 does not match manifest", python_reference_cache_path), call. = FALSE)
  }
  reference_cache_path <- .ld_r_reference_cache_path(root, manifest)
  .require_ld_file(reference_cache_path)
  if (!identical(.md5_file(reference_cache_path), manifest$r_reference_cache_md5)) {
    stop(sprintf("%s: r_reference_cache_md5 does not match manifest", reference_cache_path), call. = FALSE)
  }
  reference_panel <- load_reference_cache(reference_cache_path)
  ref_by_label <- stats::setNames(shards(reference_panel), vapply(shards(reference_panel), function(s) s$label, character(1)))
  seen_reference_bim <- character()

  for (entry in manifest$shards) {
    shard_path <- file.path(root, entry$file)
    .require_ld_file(shard_path)
    if (!identical(.md5_file(shard_path), entry$file_md5)) {
      stop(sprintf("%s: file_md5 does not match manifest", shard_path), call. = FALSE)
    }
    loaded <- .read_ld_npz_shard(
      shard_path,
      check_payload_structure = isTRUE(check_payload_structure),
      retain_ld_r = TRUE,
      expected_file_md5 = entry$file_md5
    )
    .validate_manifest_entry_agreement(entry, loaded$metadata, shard_path)
    ref_shard <- ref_by_label[[entry$chr]]
    if (is.null(ref_shard)) {
      stop(sprintf("%s: reference cache missing shard %s", reference_cache_path, entry$chr), call. = FALSE)
    }
    .validate_ld_reference_entry(ref_shard, entry, reference_cache_path)

    reference_key <- entry$chr
    if (!(reference_key %in% seen_reference_bim)) {
      .validate_bundled_reference_bim(file.path(root, entry$reference_bim), entry)
      seen_reference_bim <- c(seen_reference_bim, reference_key)
    }
  }
  list(ok = TRUE)
}

prepare_ld_npz_for_r <- function(npz_root, extract_npz = FALSE) {
  npz_root <- .validate_ld_root(npz_root, "prepare_ld_npz_for_r")
  if (!is.logical(extract_npz) || length(extract_npz) != 1L || is.na(extract_npz)) {
    stop("extract_npz must be a scalar logical value", call. = FALSE)
  }
  manifest_path <- file.path(npz_root, "ld_manifest.json")
  manifest <- .read_ld_manifest(manifest_path, require_r_reference = FALSE)
  labels <- .canonical_chr_order[.canonical_chr_order %in% unique(vapply(manifest$shards, function(e) e$chr, character(1)))]
  if (!length(labels)) {
    stop("prepare_ld_npz_for_r: manifest has no supported LD shards", call. = FALSE)
  }

  reference_cache <- if (is.null(manifest$r_reference_cache)) {
    "reference_cache.rds"
  } else {
    .validate_plain_relative_filename(manifest$r_reference_cache, sprintf("%s: r_reference_cache", manifest_path))
  }
  reference_panel <- .load_reference_from_ld_bims(npz_root, manifest$shards, labels)
  .validate_ld_manifest_reference_panel(reference_panel, manifest$shards, npz_root)
  reference_cache_path <- file.path(npz_root, reference_cache)
  .save_reference_cache_atomic(reference_panel, reference_cache_path)

  manifest$r_reference_cache <- reference_cache
  manifest$r_reference_cache_md5 <- .md5_file(reference_cache_path)
  if (isTRUE(extract_npz)) {
    for (entry in manifest$shards) {
      .extract_npz_ld_payload_cache(file.path(npz_root, entry$file), entry$file_md5)
    }
  }
  .write_ld_manifest(manifest, manifest_path)
  invisible(manifest)
}

num_snp.LDShard <- function(x, ...) x$num_snp
num_snp.LDPanel <- function(x, ...) x$num_snp
shards.LDPanel <- function(x, ...) .ld_panel_selected_shards(x, NULL)
shard_offsets.LDPanel <- function(x, ...) x$shard_offsets
reference.LDPanel <- function(x, ...) x$reference
default_chrX_sex.LDPanel <- function(x, ...) x$default_chrX_sex

a1freq.LDShard <- function(x, chrX_sex = NULL, ...) x$a1freq
a1freq.LDPanel <- function(x, chrX_sex = NULL, ...) {
  parts <- lapply(.ld_panel_selected_shards(x, chrX_sex), function(s) s$a1freq)
  unlist(parts, use.names = FALSE)
}

select_shards.LDPanel <- function(x, shards, ...) {
  available <- vapply(x$shard_groups, function(g) g[[1]]$label, character(1))
  selected <- .validate_requested_shards(shards, available, "LDPanel.select_shards")
  by_label <- stats::setNames(x$shard_groups, available)
  .new_ld_panel(
    unname(by_label[selected]),
    default_chrX_sex = x$default_chrX_sex,
    reference = if (is.null(x$reference)) NULL else select_shards(x$reference, selected)
  )
}

multiply_r2.LDPanel <- function(ld_panel, M, chrX_sex = NULL, ...) {
  if (inherits(M, "sparseMatrix")) {
    return(.multiply_r2_sparse(ld_panel, M, chrX_sex))
  }
  arr <- as.matrix(M)
  vector_input <- is.null(dim(M))
  if (vector_input) {
    if (length(M) != ld_panel$num_snp) {
      stop(sprintf("multiply_r2: vector length mismatch: expected %d, got %d", ld_panel$num_snp, length(M)), call. = FALSE)
    }
    arr <- matrix(as.numeric(M), ncol = 1L)
  } else if (length(dim(M)) == 2L) {
    if (nrow(arr) != ld_panel$num_snp) {
      stop(sprintf("multiply_r2: matrix row count mismatch: expected %d, got %d", ld_panel$num_snp, nrow(arr)), call. = FALSE)
    }
    storage.mode(arr) <- "numeric"
  } else {
    stop("multiply_r2: M must be a vector or 2D matrix", call. = FALSE)
  }

  out <- matrix(0, nrow = nrow(arr), ncol = ncol(arr))
  for (entry in .ld_panel_selected_shards_with_offsets(ld_panel, chrX_sex)) {
    rows <- seq.int(entry$start0 + 1L, entry$stop0)
    out[rows, ] <- as.matrix(entry$shard$ld_r2 %*% arr[rows, , drop = FALSE])
  }
  if (vector_input) {
    return(as.numeric(out[, 1L]))
  }
  out
}

fast_prune <- function(logpvec, ld_panel, r2_threshold = 0.2, chrX_sex = NULL) {
  if (!inherits(ld_panel, "LDPanel")) {
    stop("fast_prune: ld_panel must be an LDPanel", call. = FALSE)
  }
  threshold <- as.numeric(r2_threshold)
  if (length(threshold) != 1L || is.na(threshold) || !is.finite(threshold) || threshold < 0) {
    stop("fast_prune: r2_threshold must be a finite non-negative number", call. = FALSE)
  }
  dims <- dim(logpvec)
  if (!is.null(dims) && !(length(dims) == 2L && (dims[[1L]] == 1L || dims[[2L]] == 1L))) {
    stop("fast_prune: logpvec must be a vector", call. = FALSE)
  }
  vec <- as.numeric(logpvec)
  if (length(vec) != ld_panel$num_snp) {
    stop(sprintf("fast_prune: vector length mismatch: expected %d, got %d", ld_panel$num_snp, length(vec)), call. = FALSE)
  }
  out <- vec
  for (entry in .ld_panel_selected_shards_with_offsets(ld_panel, chrX_sex)) {
    rows <- seq.int(entry$start0 + 1L, entry$stop0)
    out[rows] <- .fast_prune_shard(out[rows], entry$shard$ld_r2, threshold)
  }
  if (!is.null(dims)) {
    return(matrix(out, nrow = dims[[1L]], ncol = dims[[2L]]))
  }
  out
}

print.LDPanel <- function(x, ...) {
  labels <- vapply(x$shard_groups, function(g) g[[1]]$label, character(1))
  cat(sprintf("<LDPanel: %d SNPs across %d shard group(s): %s>\n", x$num_snp, length(labels), paste(labels, collapse = ", ")))
  invisible(x)
}

.new_ld_shard <- function(chr, sex, num_snp, ld_r, a1freq, reference_checksum,
                          num_monomorphic_snps = 0L, retain_ld_r = TRUE) {
  if (!inherits(ld_r, "dgCMatrix")) {
    ld_r <- methods::as(ld_r, "dgCMatrix")
  }
  ld_r2 <- .build_ld_r2(ld_r)
  structure(
    list(
      label = as.character(chr),
      chr = as.character(chr),
      sex = sex,
      num_snp = as.integer(num_snp),
      ld_r = if (isTRUE(retain_ld_r)) ld_r else NULL,
      ld_r2 = ld_r2,
      a1freq = as.numeric(a1freq),
      reference_checksum = as.character(reference_checksum),
      num_monomorphic_snps = as.integer(num_monomorphic_snps)
    ),
    class = "LDShard"
  )
}

.new_ld_panel <- function(shard_groups, default_chrX_sex = "female", reference = NULL) {
  if (!length(shard_groups)) {
    stop("LDPanel requires at least one shard group", call. = FALSE)
  }
  default_chrX_sex <- .validate_chrx_sex(default_chrX_sex, "default_chrX_sex")
  pos <- 0L
  offsets <- data.frame(
    shard_label = character(length(shard_groups)),
    start0 = integer(length(shard_groups)),
    stop0 = integer(length(shard_groups)),
    stringsAsFactors = FALSE
  )
  for (i in seq_along(shard_groups)) {
    group <- shard_groups[[i]]
    if (!length(group)) {
      stop("LDPanel shard groups must be non-empty", call. = FALSE)
    }
    first <- group[[1L]]
    for (shard in group) {
      if (!identical(shard$chr, first$chr)) {
        stop("all LD shards in a group must share chr", call. = FALSE)
      }
      if (!identical(as.integer(shard$num_snp), as.integer(first$num_snp))) {
        stop("all LD shards in a group must share num_snp", call. = FALSE)
      }
    }
    if (identical(first$chr, "X")) {
      present <- vapply(group, function(s) s$sex, character(1))
      if (!(default_chrX_sex %in% present)) {
        stop(sprintf(
          "default_chrX_sex must name a loaded chrX LD shard; got %s, present %s",
          sQuote(default_chrX_sex), paste(sQuote(sort(present)), collapse = ", ")
        ), call. = FALSE)
      }
    }
    offsets$shard_label[[i]] <- first$label
    offsets$start0[[i]] <- pos
    pos <- pos + first$num_snp
    offsets$stop0[[i]] <- pos
  }
  if (!is.null(reference)) {
    ref_shards <- shards(reference)
    if (length(ref_shards) != length(shard_groups)) {
      stop("LDPanel reference shard count must match LD shard groups", call. = FALSE)
    }
    for (i in seq_along(ref_shards)) {
      if (!identical(ref_shards[[i]]$label, shard_groups[[i]][[1]]$label) ||
          !identical(as.integer(ref_shards[[i]]$num_snp), as.integer(shard_groups[[i]][[1]]$num_snp))) {
        stop("LDPanel reference shards must match LD shard groups", call. = FALSE)
      }
    }
  }
  structure(
    list(
      shard_groups = shard_groups,
      num_snp = pos,
      shard_offsets = offsets,
      default_chrX_sex = default_chrX_sex,
      reference = reference
    ),
    class = "LDPanel"
  )
}

.build_ld_r2 <- function(ld_r) {
  .new_dgCMatrix_from_slots(
    i = ld_r@i,
    p = ld_r@p,
    x = ld_r@x * ld_r@x,
    dim = ld_r@Dim,
    dimnames = ld_r@Dimnames
  )
}

.multiply_r2_sparse <- function(ld_panel, M, chrX_sex) {
  if (length(dim(M)) != 2L) {
    stop("multiply_r2: sparse M must be a 2D matrix", call. = FALSE)
  }
  if (nrow(M) != ld_panel$num_snp) {
    stop(sprintf("multiply_r2: matrix row count mismatch: expected %d, got %d", ld_panel$num_snp, nrow(M)), call. = FALSE)
  }
  parts <- vector("list", length(ld_panel$shard_groups))
  k <- 0L
  for (entry in .ld_panel_selected_shards_with_offsets(ld_panel, chrX_sex)) {
    k <- k + 1L
    rows <- seq.int(entry$start0 + 1L, entry$stop0)
    parts[[k]] <- entry$shard$ld_r2 %*% M[rows, , drop = FALSE]
  }
  do.call(rbind, parts)
}

.fast_prune_shard <- function(values, ld_r2, threshold) {
  out <- as.numeric(values)
  finite_idx <- which(is.finite(out))
  if (!length(finite_idx)) {
    return(out)
  }
  order_idx <- finite_idx[order(-abs(out[finite_idx]), finite_idx)]
  pruned <- rep.int(FALSE, length(out))
  retained <- rep.int(FALSE, length(out))
  # PERF: loop retained because greedy pruning is order-dependent; vectorization
  # is not used because each retained SNP changes which later SNPs are eligible.
  for (idx in order_idx) {
    if (pruned[[idx]]) {
      next
    }
    retained[[idx]] <- TRUE
    start <- ld_r2@p[[idx]] + 1L
    stop <- ld_r2@p[[idx + 1L]]
    if (start > stop) {
      next
    }
    row_idx <- ld_r2@i[start:stop] + 1L
    values_r2 <- ld_r2@x[start:stop]
    neighbors <- row_idx[values_r2 >= threshold]
    neighbors <- neighbors[neighbors != idx & !retained[neighbors]]
    if (length(neighbors)) {
      pruned[neighbors] <- TRUE
      out[neighbors] <- NaN
    }
  }
  out
}

.ld_panel_selected_shards <- function(panel, chrX_sex) {
  lapply(.ld_panel_selected_shards_with_offsets(panel, chrX_sex), function(entry) entry$shard)
}

.ld_panel_selected_shards_with_offsets <- function(panel, chrX_sex) {
  out <- vector("list", length(panel$shard_groups))
  for (i in seq_along(panel$shard_groups)) {
    group <- panel$shard_groups[[i]]
    first <- group[[1L]]
    if (!identical(first$chr, "X")) {
      selected <- first
    } else {
      sex <- if (is.null(chrX_sex)) panel$default_chrX_sex else .validate_chrx_sex(chrX_sex, "chrX_sex")
      by_sex <- stats::setNames(group, vapply(group, function(s) s$sex, character(1)))
      selected <- by_sex[[sex]]
      if (is.null(selected)) {
        present <- names(by_sex)
        stop(sprintf(
          "chrX_sex must name a loaded chrX LD shard; got %s, present %s",
          sQuote(sex), paste(sQuote(sort(present)), collapse = ", ")
        ), call. = FALSE)
      }
    }
    out[[i]] <- list(
      shard = selected,
      start0 = panel$shard_offsets$start0[[i]],
      stop0 = panel$shard_offsets$stop0[[i]]
    )
  }
  out
}

.load_ld_shard_groups <- function(root, manifest, ref_shards, retain_ld_r) {
  entries <- manifest$shards
  groups <- vector("list", length(ref_shards))
  for (i in seq_along(ref_shards)) {
    ref_shard <- ref_shards[[i]]
    if (identical(ref_shard$label, "X")) {
      selected <- Filter(function(e) identical(e$chr, "X"), entries)
      if (!length(selected)) {
        stop("LD manifest has no chrX shard for reference shard X", call. = FALSE)
      }
    } else {
      selected <- Filter(function(e) identical(e$chr, ref_shard$label) && is.null(e$sex), entries)
      if (length(selected) != 1L) {
        stop(sprintf("LD manifest must contain exactly one autosomal shard for chr %s", ref_shard$label), call. = FALSE)
      }
    }
    group <- vector("list", length(selected))
    seen_sex <- character()
    for (j in seq_along(selected)) {
      entry <- selected[[j]]
      shard_path <- file.path(root, entry$file)
      suffix <- if (is.null(entry$sex)) "" else sprintf(" (%s)", entry$sex)
      .ld_info(sprintf("statgen.load_ld: loading shard %s%s from %s", entry$chr, suffix, shard_path))
      loaded <- .read_ld_npz_shard(
        shard_path,
        check_payload_structure = FALSE,
        retain_ld_r = retain_ld_r,
        expected_file_md5 = entry$file_md5
      )
      .validate_manifest_entry_agreement(entry, loaded$metadata, shard_path)
      .validate_ld_reference_compatibility(loaded$shard, ref_shard, shard_path)
      key <- if (is.null(loaded$shard$sex)) "<null>" else loaded$shard$sex
      if (key %in% seen_sex) {
        stop(sprintf("LD manifest has duplicate shard for chr %s sex %s", loaded$shard$chr, sQuote(key)), call. = FALSE)
      }
      seen_sex <- c(seen_sex, key)
      group[[j]] <- loaded$shard
    }
    groups[[i]] <- group
  }
  groups
}

.ld_info <- function(message_text) {
  if (identical(get_verbosity(), "info")) {
    message(message_text)
  }
  invisible(NULL)
}

copy_ld_fixture <- function() {
  src <- system.file("extdata", "ld", package = "statgen", mustWork = TRUE)
  dst <- tempfile("ld_npz_")
  dir.create(dst)
  ok <- file.copy(list.files(src, full.names = TRUE), dst, recursive = TRUE)
  stopifnot(all(ok))
  unlink(file.path(dst, "reference_cache.rds"))
  manifest_path <- file.path(dst, "ld_manifest.json")
  manifest <- jsonlite::fromJSON(manifest_path, simplifyVector = FALSE)
  manifest$r_reference_cache <- NULL
  manifest$r_reference_cache_md5 <- NULL
  writeLines(jsonlite::toJSON(manifest, auto_unbox = TRUE, pretty = TRUE, null = "null"), manifest_path, useBytes = TRUE)
  dst
}

test_that("bundled prepared LD fixture loads without in-place preparation", {
  ld_root <- system.file("extdata", "ld", package = "statgen", mustWork = TRUE)

  report <- validate_ld_distribution(ld_root)
  expect_true(report$ok)

  ld <- load_ld(ld_root)
  expect_s3_class(ld, "LDPanel")
  expect_equal(num_snp(ld), 8L)
  expect_equal(
    round(a1freq(ld), 6),
    c(0.30, 0.40, 0.20, 0.35, 0.15, 0.28, 0.32, 0.22)
  )
})

test_that("prepare_ld_npz_for_r enables LD loading and operations", {
  old_verbosity <- get_verbosity()
  on.exit(set_verbosity(old_verbosity), add = TRUE)
  set_verbosity("quiet")

  ld_root <- copy_ld_fixture()

  expect_error(load_ld(ld_root), "missing r_reference_cache")

  manifest <- prepare_ld_npz_for_r(ld_root, extract_npz = TRUE)
  expect_identical(manifest$runtime_format, "python_npz_csc32")
  expect_identical(manifest$r_reference_cache, "reference_cache.rds")
  expect_match(manifest$r_reference_cache_md5, "^[0-9a-f]{32}$")
  expect_true(file.exists(file.path(ld_root, manifest$r_reference_cache)))
  cache_dir <- file.path(ld_root, paste0(manifest$shards[[1]]$file, ".d"))
  expect_true(dir.exists(cache_dir))
  expect_true(file.exists(file.path(cache_dir, "data.npy")))
  marker <- jsonlite::fromJSON(file.path(cache_dir, ".statgen-extracted-npz.json"))
  expect_identical(marker$source_file, manifest$shards[[1]]$file)
  expect_identical(marker$source_file_md5, manifest$shards[[1]]$file_md5)

  manifest2 <- prepare_ld_npz_for_r(ld_root)
  expect_identical(manifest2$r_reference_cache, manifest$r_reference_cache)
  expect_match(manifest2$r_reference_cache_md5, "^[0-9a-f]{32}$")

  report_fast <- validate_ld_distribution(ld_root)
  expect_true(report_fast$ok)
  report_full <- validate_ld_distribution(ld_root, check_payload_structure = TRUE)
  expect_true(report_full$ok)

  ref <- load_ld_reference(ld_root)
  ld <- load_ld(ld_root)
  expect_s3_class(ld, "LDPanel")
  expect_equal(num_snp(ld), 8L)
  expect_equal(num_snp(ref), 8L)
  expect_equal(vapply(shards(ld), function(s) s$label, character(1)), c("1", "X"))
  expect_equal(default_chrX_sex(ld), "female")

  expect_equal(
    round(a1freq(ld), 6),
    c(0.30, 0.40, 0.20, 0.35, 0.15, 0.28, 0.32, 0.22)
  )
  expect_equal(
    round(a1freq(ld, chrX_sex = "combined"), 6),
    c(0.30, 0.40, 0.20, 0.35, 0.15, 0.25, 0.30, 0.20)
  )

  expect_equal(
    round(multiply_r2(ld, seq_len(num_snp(ld))), 6),
    c(2.89, 3.81, 3.09, 6.95, 6.96, 8.9575, 10.515, 8.8575)
  )
  expect_equal(
    round(fast_prune(c(9, 9, 7, 1, 2, 6, 5, 4), ld, 0.35), 6),
    c(9, NaN, 7, NaN, 2, 6, NaN, 4)
  )

  ld_no_r <- load_ld(ld_root, retain_ld_r = FALSE)
  expect_true(is.null(ld_no_r$shard_groups[[1]][[1]]$ld_r))
  expect_equal(round(multiply_r2(ld_no_r, seq_len(num_snp(ld_no_r)))[[1]], 6), 2.89)
})

test_that("R LD loader rejects non-NPZ runtime manifests", {
  ld_root <- copy_ld_fixture()
  manifest_path <- file.path(ld_root, "ld_manifest.json")
  manifest <- jsonlite::fromJSON(manifest_path, simplifyVector = FALSE)
  manifest$runtime_format <- "matlab_mat_sparse_double"
  writeLines(jsonlite::toJSON(manifest, auto_unbox = TRUE), manifest_path)

  expect_error(load_ld(ld_root), "unsupported LD runtime_format")
})

.sumstats_cache_schema <- "sumstats_cache/0.1"
.sumstats_required_cols <- c("chr", "bp", "a1", "a2", "p")
.sumstats_optional_cols <- c("z", "n", "beta", "se", "eaf", "info")
.sumstats_allowed_cols <- c(.sumstats_required_cols, .sumstats_optional_cols, "snp")
.sumstats_col_map <- c(pos = "bp", effectallele = "a1", otherallele = "a2")

logpvec <- function(x, ...) UseMethod("logpvec")
zvec <- function(x, ...) UseMethod("zvec")
nvec <- function(x, ...) UseMethod("nvec")
beta_vec <- function(x, ...) UseMethod("beta_vec")
se_vec <- function(x, ...) UseMethod("se_vec")
eaf_vec <- function(x, ...) UseMethod("eaf_vec")
info_vec <- function(x, ...) UseMethod("info_vec")

load_sumstats <- function(path, reference) {
  path <- .validate_path_scalar(path, "path")
  if (!inherits(reference, "ReferencePanel")) {
    stop("reference must be a ReferencePanel", call. = FALSE)
  }
  df <- .parse_sumstats_tsv(path)
  .build_sumstats_panel(df, reference, path)
}

create_sumstats <- function(reference, p, z = NULL, n = NULL,
                            beta = NULL, se = NULL,
                            eaf = NULL, info = NULL) {
  if (!inherits(reference, "ReferencePanel")) {
    stop("reference must be a ReferencePanel", call. = FALSE)
  }
  num_snp_value <- num_snp(reference)
  p <- .coerce_aligned_numeric_vector(p, num_snp_value, "p", allow_inf = FALSE)
  .validate_p_values(p, "p")
  aligned <- list(
    z = .coerce_optional_aligned_numeric_vector(z, num_snp_value, "z"),
    n = .coerce_optional_aligned_numeric_vector(n, num_snp_value, "n"),
    beta = .coerce_optional_aligned_numeric_vector(beta, num_snp_value, "beta"),
    se = .coerce_optional_aligned_numeric_vector(se, num_snp_value, "se"),
    eaf = .coerce_optional_aligned_numeric_vector(eaf, num_snp_value, "eaf"),
    info = .coerce_optional_aligned_numeric_vector(info, num_snp_value, "info")
  )
  .build_sumstats_from_aligned(reference, .derive_logp(p), aligned)
}

save_sumstats_cache <- function(sumstats, path) {
  if (!inherits(sumstats, "Sumstats")) {
    stop("sumstats must be a Sumstats object", call. = FALSE)
  }
  path <- .validate_path_scalar(path, "path")
  offsets <- shard_offsets(sumstats)
  metadata <- list(
    schema = .sumstats_cache_schema,
    n_shards = length(sumstats$shards),
    shard_labels = vapply(sumstats$shards, function(s) s$label, character(1)),
    shard_checksums = vapply(sumstats$shards, function(s) s$reference_checksum, character(1)),
    shard_start0 = offsets$start0,
    shard_stop0 = offsets$stop0,
    has_z = !is.null(zvec(sumstats)),
    has_n = !is.null(nvec(sumstats)),
    has_beta = !is.null(beta_vec(sumstats)),
    has_se = !is.null(se_vec(sumstats)),
    has_eaf = !is.null(eaf_vec(sumstats)),
    has_info = !is.null(info_vec(sumstats))
  )
  payload <- list(
    metadata = metadata,
    logpvec = logpvec(sumstats),
    zvec = zvec(sumstats),
    nvec = nvec(sumstats),
    beta_vec = beta_vec(sumstats),
    se_vec = se_vec(sumstats),
    eaf_vec = eaf_vec(sumstats),
    info_vec = info_vec(sumstats)
  )
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  saveRDS(payload, path)
  invisible(NULL)
}

load_sumstats_cache <- function(path, shards = NULL) {
  path <- .validate_path_scalar(path, "path")
  payload <- readRDS(path)
  .validate_sumstats_cache_payload(payload)
  meta <- payload$metadata
  selected <- .validate_requested_shards(shards, meta$shard_labels, "load_sumstats_cache")
  label_to_index <- stats::setNames(seq_along(meta$shard_labels), meta$shard_labels)

  out <- vector("list", length(selected))
  for (i in seq_along(selected)) {
    label <- selected[[i]]
    idx <- label_to_index[[label]]
    start <- meta$shard_start0[[idx]] + 1L
    stop <- meta$shard_stop0[[idx]]
    out[[i]] <- .new_sumstats_shard(
      label = label,
      reference_checksum = meta$shard_checksums[[idx]],
      logpvec = payload$logpvec[start:stop],
      zvec = if (isTRUE(meta$has_z)) payload$zvec[start:stop] else NULL,
      nvec = if (isTRUE(meta$has_n)) payload$nvec[start:stop] else NULL,
      beta_vec = if (isTRUE(meta$has_beta)) payload$beta_vec[start:stop] else NULL,
      se_vec = if (isTRUE(meta$has_se)) payload$se_vec[start:stop] else NULL,
      eaf_vec = if (isTRUE(meta$has_eaf)) payload$eaf_vec[start:stop] else NULL,
      info_vec = if (isTRUE(meta$has_info)) payload$info_vec[start:stop] else NULL
    )
  }
  panel <- .new_sumstats_panel(out)
  .warn_optional_zn_completeness(zvec(panel), nvec(panel), logpvec(panel), "load_sumstats_cache")
  panel
}

num_snp.SumstatsShard <- function(x, ...) x$num_snp
num_snp.Sumstats <- function(x, ...) x$num_snp
shards.Sumstats <- function(x, ...) x$shards
shard_offsets.Sumstats <- function(x, ...) x$shard_offsets

logpvec.SumstatsShard <- function(x, ...) x$logpvec
logpvec.Sumstats <- function(x, ...) unlist(lapply(x$shards, logpvec), use.names = FALSE)

zvec.SumstatsShard <- function(x, ...) x$zvec
zvec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "zvec")

nvec.SumstatsShard <- function(x, ...) x$nvec
nvec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "nvec")

is_present.SumstatsShard <- function(x, ...) !is.nan(x$logpvec)
is_present.Sumstats <- function(x, ...) !is.nan(logpvec(x))

beta_vec.SumstatsShard <- function(x, ...) x$beta_vec
beta_vec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "beta_vec")

se_vec.SumstatsShard <- function(x, ...) x$se_vec
se_vec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "se_vec")

eaf_vec.SumstatsShard <- function(x, ...) x$eaf_vec
eaf_vec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "eaf_vec")

info_vec.SumstatsShard <- function(x, ...) x$info_vec
info_vec.Sumstats <- function(x, ...) .optional_sumstats_concat(x, "info_vec")

select_shards.Sumstats <- function(x, shards, ...) {
  available <- vapply(x$shards, function(s) s$label, character(1))
  selected <- .validate_requested_shards(shards, available, "Sumstats.select_shards")
  by_label <- stats::setNames(x$shards, available)
  .new_sumstats_panel(unname(by_label[selected]))
}

save_cache.Sumstats <- function(x, path, ...) save_sumstats_cache(x, path)

print.Sumstats <- function(x, ...) {
  labels <- vapply(x$shards, function(s) s$label, character(1))
  cat(sprintf("<Sumstats: %d SNPs across %d shard(s): %s>\n", x$num_snp, length(labels), paste(labels, collapse = ", ")))
  invisible(x)
}

.optional_sumstats_concat <- function(x, field) {
  present <- vapply(x$shards, function(s) !is.null(s[[field]]), logical(1))
  if (!any(present)) {
    return(NULL)
  }
  if (!all(present)) {
    stop(sprintf("Invalid Sumstats object: inconsistent %s presence across shards", field), call. = FALSE)
  }
  unlist(lapply(x$shards, function(s) s[[field]]), use.names = FALSE)
}

.new_sumstats_shard <- function(label, reference_checksum, logpvec,
                                zvec = NULL, nvec = NULL, beta_vec = NULL,
                                se_vec = NULL, eaf_vec = NULL, info_vec = NULL) {
  logpvec <- as.numeric(logpvec)
  n <- length(logpvec)
  structure(
    list(
      label = as.character(label),
      reference_checksum = as.character(reference_checksum),
      num_snp = as.integer(n),
      logpvec = logpvec,
      zvec = .coerce_optional_shard_numeric_vector(zvec, n, "zvec"),
      nvec = .coerce_optional_shard_numeric_vector(nvec, n, "nvec"),
      beta_vec = .coerce_optional_shard_numeric_vector(beta_vec, n, "beta_vec"),
      se_vec = .coerce_optional_shard_numeric_vector(se_vec, n, "se_vec"),
      eaf_vec = .coerce_optional_shard_numeric_vector(eaf_vec, n, "eaf_vec"),
      info_vec = .coerce_optional_shard_numeric_vector(info_vec, n, "info_vec")
    ),
    class = "SumstatsShard"
  )
}

.new_sumstats_panel <- function(shard_list) {
  if (!length(shard_list)) {
    stop("Sumstats requires at least one shard", call. = FALSE)
  }
  for (field in c("zvec", "nvec", "beta_vec", "se_vec", "eaf_vec", "info_vec")) {
    present <- vapply(shard_list, function(s) !is.null(s[[field]]), logical(1))
    if (any(present) && !all(present)) {
      stop(sprintf("Sumstats shards must have consistent %s presence", field), call. = FALSE)
    }
  }
  pos <- 0L
  offsets <- data.frame(
    shard_label = character(length(shard_list)),
    start0 = integer(length(shard_list)),
    stop0 = integer(length(shard_list)),
    stringsAsFactors = FALSE
  )
  for (i in seq_along(shard_list)) {
    offsets$shard_label[[i]] <- shard_list[[i]]$label
    offsets$start0[[i]] <- pos
    pos <- pos + shard_list[[i]]$num_snp
    offsets$stop0[[i]] <- pos
  }
  structure(list(shards = shard_list, num_snp = pos, shard_offsets = offsets), class = "Sumstats")
}

.parse_sumstats_tsv <- function(path) {
  if (!file.exists(path)) {
    stop(sprintf("File not found: %s", path), call. = FALSE)
  }
  raw_names <- .read_sumstats_header(path)
  canonical_names <- .canonicalize_sumstats_columns(raw_names, path)
  missing <- setdiff(.sumstats_required_cols, canonical_names)
  if (length(missing)) {
    stop(sprintf("%s: missing required columns: %s", path, paste(missing, collapse = ", ")), call. = FALSE)
  }
  selected <- canonical_names %in% c(.sumstats_required_cols, .sumstats_optional_cols)
  selected_raw <- raw_names[selected]
  selected_names <- canonical_names[selected]
  char_cols <- selected_raw[selected_names %in% c("chr", "a1", "a2")]
  numeric_cols <- selected_raw[!(selected_names %in% c("chr", "a1", "a2"))]

  fread_args <- c(
    .sumstats_fread_input(path),
    list(
      select = selected_raw,
      colClasses = list(character = char_cols, numeric = numeric_cols),
      data.table = FALSE,
      showProgress = FALSE
    )
  )
  df <- tryCatch(
    do.call(data.table::fread, fread_args),
    error = function(e) {
      stop(sprintf("%s: malformed TSV", path), call. = FALSE)
    }
  )
  names(df) <- selected_names

  missing <- setdiff(.sumstats_required_cols, names(df))
  if (length(missing)) {
    stop(sprintf("%s: missing required columns: %s", path, paste(missing, collapse = ", ")), call. = FALSE)
  }

  bad_chr <- is.na(df$chr) | df$chr == ""
  if (any(bad_chr)) {
    stop(sprintf("%s: row %d: chr must be non-empty", path, which(bad_chr)[[1]] + 1L), call. = FALSE)
  }
  .validate_source_chr_labels(df$chr, path)
  df <- df[df$chr %in% .canonical_chr_order, , drop = FALSE]

  for (field in c("a1", "a2")) {
    bad <- is.na(df[[field]]) | df[[field]] == ""
    if (any(bad)) {
      row <- as.integer(rownames(df)[which(bad)[[1]]]) + 1L
      stop(sprintf("%s: row %d: %s must be non-empty", path, row, field), call. = FALSE)
    }
  }
  bad_a1 <- !grepl("^[ACGT]+$", df$a1)
  if (any(bad_a1)) {
    idx <- which(bad_a1)[[1]]
    row <- as.integer(rownames(df)[idx]) + 1L
    stop(sprintf("%s: row %d: a1 must be uppercase DNA bases (A/C/G/T): %s", path, row, sQuote(df$a1[[idx]])), call. = FALSE)
  }
  bad_a2 <- !grepl("^[ACGT]+$", df$a2)
  if (any(bad_a2)) {
    idx <- which(bad_a2)[[1]]
    row <- as.integer(rownames(df)[idx]) + 1L
    stop(sprintf("%s: row %d: a2 must be uppercase DNA bases (A/C/G/T): %s", path, row, sQuote(df$a2[[idx]])), call. = FALSE)
  }
  same_allele <- df$a1 == df$a2
  if (any(same_allele)) {
    row <- as.integer(rownames(df)[which(same_allele)[[1]]]) + 1L
    stop(sprintf("%s: row %d: a1 and a2 must differ", path, row), call. = FALSE)
  }

  bp <- suppressWarnings(as.numeric(df$bp))
  bad_bp <- is.na(bp) | !is.finite(bp) | floor(bp) != bp
  if (any(bad_bp)) {
    idx <- which(bad_bp)[[1]]
    row <- as.integer(rownames(df)[idx]) + 1L
    stop(sprintf("%s: row %d: bp is not an integer: %s", path, row, sQuote(df$bp[[idx]])), call. = FALSE)
  }
  df$bp <- as.integer(bp)

  p <- suppressWarnings(as.numeric(df$p))
  bad_p <- is.na(p) | !is.finite(p) | p < 0 | p > 1
  if (any(bad_p)) {
    idx <- which(bad_p)[[1]]
    row <- as.integer(rownames(df)[idx]) + 1L
    stop(sprintf("%s: row %d: p must be finite numeric in [0, 1]: %s", path, row, sQuote(df$p[[idx]])), call. = FALSE)
  }
  df$p <- p

  for (field in .sumstats_optional_cols) {
    if (field %in% names(df)) {
      value <- suppressWarnings(as.numeric(df[[field]]))
      value[!is.finite(value)] <- NaN
      df[[field]] <- value
    }
  }
  df
}

.read_sumstats_header <- function(path) {
  con <- if (grepl("\\.gz$", path, ignore.case = TRUE)) gzfile(path, "rt") else file(path, "rt")
  on.exit(close(con), add = TRUE)
  header <- tryCatch(readLines(con, n = 1L, warn = FALSE), error = function(e) character())
  if (!length(header)) {
    stop(sprintf("%s: malformed TSV", path), call. = FALSE)
  }
  strsplit(header[[1]], "\t", fixed = TRUE)[[1]]
}

.sumstats_fread_input <- function(path) {
  if (!grepl("\\.gz$", path, ignore.case = TRUE) || requireNamespace("R.utils", quietly = TRUE)) {
    return(list(input = path))
  }
  gzip <- Sys.which("gzip")
  if (!nzchar(gzip)) {
    stop(sprintf("%s: gzip input requires R.utils or a gzip executable", path), call. = FALSE)
  }
  cmd <- paste(shQuote(gzip), "-dc", shQuote(path))
  list(cmd = cmd)
}

.canonicalize_sumstats_columns <- function(columns, path) {
  out <- tolower(as.character(columns))
  mapped <- .sumstats_col_map[out]
  out[!is.na(mapped)] <- unname(mapped[!is.na(mapped)])
  duplicates <- unique(out[duplicated(out)])
  if (length(duplicates)) {
    stop(sprintf("%s: duplicate columns after column normalization: %s", path, paste(sort(duplicates), collapse = ", ")), call. = FALSE)
  }
  out
}

.validate_source_chr_labels <- function(chr_values, path) {
  chr_style <- startsWith(tolower(chr_values), "chr")
  if (any(chr_style)) {
    stop(sprintf("%s: row %d: chr-style labels (e.g., chr1/chrX) are not allowed", path, which(chr_style)[[1]] + 1L), call. = FALSE)
  }
  known <- chr_values %in% c(.canonical_chr_order, .ignored_chr)
  if (!all(known)) {
    idx <- which(!known)[[1]]
    stop(sprintf("%s: row %d: unsupported chr label %s; expected 1-22, X (Y/MT are ignored)", path, idx + 1L, sQuote(chr_values[[idx]])), call. = FALSE)
  }
}

.build_sumstats_panel <- function(df, reference, source_path) {
  n <- num_snp(reference)
  aligned_p <- rep.int(NaN, n)
  aligned_optional <- lapply(.sumstats_optional_cols, function(field) {
    if (field %in% names(df)) rep.int(NaN, n) else NULL
  })
  names(aligned_optional) <- .sumstats_optional_cols

  src_chr <- as.character(df$chr)
  keep_supported <- src_chr %in% .canonical_chr_order
  source_rows_by_chr <- split(which(keep_supported), src_chr[keep_supported])
  src_a1_hash64 <- .allele_hash64(df$a1)
  src_a2_hash64 <- .allele_hash64(df$a2)

  ref_shards <- shards(reference)
  offsets <- shard_offsets(reference)
  for (i in seq_along(ref_shards)) {
    ref_shard <- ref_shards[[i]]
    start <- offsets$start0[[i]] + 1L
    stop <- offsets$stop0[[i]]
    src_idx <- source_rows_by_chr[[ref_shard$label]]
    if (is.null(src_idx)) {
      src_idx <- integer()
    }
    if (!length(src_idx)) {
      warning(sprintf("%s: shard %s has no matching source rows; representing as all missing", source_path, ref_shard$label), call. = FALSE)
      next
    }
    local_match <- .match_variant_keys(
      bp(ref_shard), a1_hash64(ref_shard), a2_hash64(ref_shard),
      df$bp[src_idx], src_a1_hash64[src_idx], src_a2_hash64[src_idx],
      ref_shard$label, "sumstats"
    )
    has_match <- !is.na(local_match)
    matched_src <- src_idx[local_match[has_match]]
    aligned_idx <- seq.int(start, stop)[has_match]
    aligned_p[aligned_idx] <- df$p[matched_src]
    for (field in .sumstats_optional_cols) {
      if (!is.null(aligned_optional[[field]])) {
        aligned_optional[[field]][aligned_idx] <- df[[field]][matched_src]
      }
    }

    matched_source_mask <- rep.int(FALSE, length(src_idx))
    matched_source_mask[local_match[has_match]] <- TRUE
    unmatched_src <- src_idx[!matched_source_mask]
    swapped <- .count_variant_key_intersections(
      bp(ref_shard), a1_hash64(ref_shard), a2_hash64(ref_shard),
      df$bp[unmatched_src], src_a2_hash64[unmatched_src], src_a1_hash64[unmatched_src]
    )
    if (swapped > 0L) {
      warning(
        sprintf(
          "%s: shard %s: %d unmatched sumstats variant(s) would match the reference if a1/a2 were swapped; variants remain unmatched",
          source_path, ref_shard$label, swapped
        ),
        call. = FALSE
      )
    }
  }

  logp <- .derive_logp(aligned_p)
  .warn_optional_zn_completeness(aligned_optional$z, aligned_optional$n, logp, "load_sumstats")
  .build_sumstats_from_aligned(reference, logp, aligned_optional)
}

.build_sumstats_from_aligned <- function(reference, aligned_logp, aligned_optional) {
  ref_shards <- shards(reference)
  offsets <- shard_offsets(reference)
  out <- vector("list", length(ref_shards))
  for (i in seq_along(ref_shards)) {
    ref_shard <- ref_shards[[i]]
    start <- offsets$start0[[i]] + 1L
    stop <- offsets$stop0[[i]]
    out[[i]] <- .new_sumstats_shard(
      label = ref_shard$label,
      reference_checksum = ref_shard$checksum,
      logpvec = aligned_logp[start:stop],
      zvec = if (is.null(aligned_optional$z)) NULL else aligned_optional$z[start:stop],
      nvec = if (is.null(aligned_optional$n)) NULL else aligned_optional$n[start:stop],
      beta_vec = if (is.null(aligned_optional$beta)) NULL else aligned_optional$beta[start:stop],
      se_vec = if (is.null(aligned_optional$se)) NULL else aligned_optional$se[start:stop],
      eaf_vec = if (is.null(aligned_optional$eaf)) NULL else aligned_optional$eaf[start:stop],
      info_vec = if (is.null(aligned_optional$info)) NULL else aligned_optional$info[start:stop]
    )
  }
  .new_sumstats_panel(out)
}

.derive_logp <- function(p) {
  out <- rep.int(NaN, length(p))
  zero <- !is.na(p) & p == 0
  positive <- !is.na(p) & p > 0
  out[zero] <- Inf
  out[positive] <- -log10(p[positive])
  out
}

.validate_p_values <- function(p, name) {
  bad <- is.na(p) | !is.finite(p) | p < 0 | p > 1
  if (any(bad)) {
    stop(sprintf("%s[%d] must be finite numeric in [0, 1]", name, which(bad)[[1]]), call. = FALSE)
  }
}

.coerce_aligned_numeric_vector <- function(vec, n, name, allow_inf = FALSE) {
  arr <- as.numeric(vec)
  if (length(arr) != n) {
    stop(sprintf("%s length mismatch: expected %d, got %d", name, n, length(arr)), call. = FALSE)
  }
  if (!allow_inf) {
    bad <- is.infinite(arr)
    if (any(bad)) {
      stop(sprintf("%s[%d] must be finite numeric or NaN", name, which(bad)[[1]]), call. = FALSE)
    }
  }
  arr
}

.coerce_optional_aligned_numeric_vector <- function(vec, n, name) {
  if (is.null(vec)) {
    return(NULL)
  }
  .coerce_aligned_numeric_vector(vec, n, name, allow_inf = FALSE)
}

.coerce_optional_shard_numeric_vector <- function(vec, n, name) {
  if (is.null(vec)) {
    return(NULL)
  }
  arr <- as.numeric(vec)
  if (length(arr) != n) {
    stop(sprintf("%s length mismatch: expected %d, got %d", name, n, length(arr)), call. = FALSE)
  }
  arr
}

.warn_optional_zn_completeness <- function(zvec, nvec, logpvec, context) {
  present <- !is.nan(logpvec)
  for (entry in list(list(name = "zvec", value = zvec), list(name = "nvec", value = nvec))) {
    if (is.null(entry$value)) {
      warning(sprintf("%s: %s is absent", context, entry$name), call. = FALSE)
    } else if (any(is.nan(entry$value[present]))) {
      warning(sprintf("%s: %s has missing values among present sumstats variants", context, entry$name), call. = FALSE)
    }
  }
}

.validate_sumstats_cache_payload <- function(payload) {
  cache <- .validate_cache_payload_metadata(payload, .sumstats_cache_schema, "sumstats")
  meta <- cache$meta
  total <- cache$total
  if (is.null(payload$logpvec) || length(payload$logpvec) != total) {
    stop("Invalid sumstats cache: logpvec length mismatch", call. = FALSE)
  }
  for (field in c("z", "n", "beta", "se", "eaf", "info")) {
    flag <- paste0("has_", field)
    vector_name <- if (field %in% c("z", "n")) paste0(field, "vec") else paste0(field, "_vec")
    if (isTRUE(meta[[flag]]) && (is.null(payload[[vector_name]]) || length(payload[[vector_name]]) != total)) {
      stop(sprintf("Invalid sumstats cache: %s length mismatch", vector_name), call. = FALSE)
    }
  }
}

.phase2_reference_template <- function() {
  ref_path <- system.file("extdata", "reference_chr1.bim", package = "statgen", mustWork = TRUE)
  file.path(dirname(ref_path), "reference_chr@.bim")
}

.phase2_capture_warnings <- function(expr) {
  warnings <- character()
  value <- withCallingHandlers(
    expr,
    warning = function(w) {
      warnings <<- c(warnings, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  list(value = value, warnings = warnings)
}

test_that("load_sumstats aligns TSV rows and cache round-trips", {
  sum_path <- system.file("extdata", "traits.tsv.gz", package = "statgen", mustWork = TRUE)
  ref <- load_reference(.phase2_reference_template())

  expect_warning(s <- load_sumstats(sum_path, ref), "zvec has missing values")
  expect_s3_class(s, "Sumstats")
  expect_equal(num_snp(s), 8L)
  expect_equal(is_present(s), c(TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, TRUE, FALSE))
  expect_equal(zvec(s), c(2.5, NaN, 1.8, -1.2, NaN, 3.0, 0.5, NaN))
  expect_equal(nvec(s), c(1000, 1000, 1000, 1000, NaN, 500, 500, NaN))
  expect_true(is.infinite(logpvec(s)[[3]]))
  expect_true(is.null(beta_vec(s)))

  cache <- tempfile(fileext = ".rds")
  save_sumstats_cache(s, cache)
  expect_warning(loaded <- load_sumstats_cache(cache), "zvec has missing values")
  expect_equal(logpvec(loaded), logpvec(s))
  expect_equal(nvec(loaded), nvec(s))
  expect_equal(is_present(loaded), is_present(s))

  x_only <- select_shards(loaded, "X")
  expect_equal(num_snp(x_only), 3L)
  expect_equal(is_present(x_only), c(TRUE, TRUE, FALSE))
})

test_that("load_sumstats accepts documented column aliases", {
  ref <- load_reference(.phase2_reference_template(), shards = "1")
  path <- tempfile(fileext = ".tsv")
  writeLines(c(
    "CHR\tPOS\tSNP\tEffectAllele\tOtherAllele\tP\tZ\tN",
    "1\t100\trs1001\tA\tG\t0\t2\t100"
  ), path)

  captured <- .phase2_capture_warnings(load_sumstats(path, ref))
  s <- captured$value
  expect_true(is.infinite(logpvec(s)[[1]]))
  expect_equal(zvec(s)[[1]], 2)
  expect_equal(nvec(s)[[1]], 100)
  expect_true(is_present(s)[[1]])
})

test_that("create_sumstats preserves optional fields and save_cache convenience", {
  ref <- load_reference(.phase2_reference_template())
  n <- num_snp(ref)
  s <- create_sumstats(
    ref,
    p = rep(0.5, n),
    z = seq_len(n),
    n = rep(100, n),
    beta = rep(0.1, n),
    se = rep(0.2, n),
    eaf = rep(0.3, n),
    info = rep(0.9, n)
  )

  expect_equal(nvec(s), rep(100, n))
  expect_equal(beta_vec(s), rep(0.1, n))
  expect_equal(se_vec(s), rep(0.2, n))
  expect_equal(eaf_vec(s), rep(0.3, n))
  expect_equal(info_vec(s), rep(0.9, n))

  cache <- tempfile(fileext = ".rds")
  save_cache(s, cache)
  loaded <- load_sumstats_cache(cache, shards = "X")
  expect_equal(num_snp(loaded), 3L)
  expect_equal(nvec(loaded), rep(100, 3L))
})

test_that("load_sumstats warns for absent zvec and nvec", {
  ref <- load_reference(.phase2_reference_template(), shards = "1")
  path <- tempfile(fileext = ".tsv")
  writeLines(c(
    "chr\tbp\ta1\ta2\tp",
    "1\t100\tA\tG\t0.1"
  ), path)

  captured <- .phase2_capture_warnings(load_sumstats(path, ref))
  expect_true(any(grepl("zvec is absent", captured$warnings)))
  expect_true(any(grepl("nvec is absent", captured$warnings)))
  expect_null(zvec(captured$value))
  expect_null(nvec(captured$value))

  cache <- tempfile(fileext = ".rds")
  save_sumstats_cache(captured$value, cache)
  loaded <- .phase2_capture_warnings(load_sumstats_cache(cache))
  expect_true(any(grepl("zvec is absent", loaded$warnings)))
  expect_true(any(grepl("nvec is absent", loaded$warnings)))
})

test_that("load_annotations paints BED intervals and cache round-trips", {
  bed1 <- system.file("extdata", "anno1.bed", package = "statgen", mustWork = TRUE)
  bed2 <- system.file("extdata", "anno2.bed", package = "statgen", mustWork = TRUE)
  ref <- load_reference(.phase2_reference_template())
  ann <- load_annotations(c(bed1, bed2), ref)

  expect_s3_class(ann, "AnnotationPanel")
  expect_equal(num_snp(ann), 8L)
  expect_equal(num_annot(ann), 2L)
  expect_equal(annonames(ann), c("anno1", "anno2"))
  expect_s4_class(annomat(ann), "lgCMatrix")
  ann_dense <- as.matrix(annomat(ann))
  expect_equal(colnames(annomat(ann)), c("anno1", "anno2"))
  expect_equal(as.vector(ann_dense[, 1L]), c(TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE))
  expect_equal(as.vector(ann_dense[, 2L]), c(FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE))

  cache <- tempfile(fileext = ".rds")
  save_cache(ann, cache)
  loaded <- load_annotations_cache(cache, shards = "X")
  expect_equal(annonames(loaded), c("anno1", "anno2"))
  expect_equal(colnames(annomat(loaded)), c("anno1", "anno2"))
  expect_equal(as.matrix(annomat(loaded)), as.matrix(annomat(ann))[6:8, ])
})

test_that("load_annotations accepts scalar path and leading comments", {
  ref <- load_reference(.phase2_reference_template(), shards = "1")
  bed <- tempfile(fileext = ".bed")
  writeLines(c("# leading comment", "", "1\t99\t200"), bed)

  ann <- load_annotations(bed, ref)
  expect_equal(num_annot(ann), 1L)
  expect_equal(as.vector(as.matrix(annomat(ann))), c(TRUE, TRUE, FALSE, FALSE, FALSE))
})

test_that("annotation factories, selection, and union work", {
  ref <- load_reference(.phase2_reference_template())
  n <- num_snp(ref)
  a <- create_annotation(ref, rep(c(1, 0), length.out = n), "a")
  b <- create_annotations(
    ref,
    annotation_matrix = matrix(rep(c(0, 1), length.out = n), ncol = 1L),
    annotation_names = "b"
  )
  u <- union_annotations(a, b)

  expect_equal(annonames(u), c("a", "b"))
  expect_equal(colnames(annomat(u)), c("a", "b"))
  selected <- select_annotations(u, names = c("b", "a"))
  expect_equal(annonames(selected), c("b", "a"))
  expect_equal(colnames(annomat(selected)), c("b", "a"))
  expect_error(select_annotations(u, "missing"), "unknown annotation")
})

test_that("annotations cache validation rejects bad schema and zero shards", {
  ref <- load_reference(.phase2_reference_template(), shards = "1")
  ann <- create_annotation(ref, rep(TRUE, num_snp(ref)), "a")
  cache <- tempfile(fileext = ".rds")
  save_annotations_cache(ann, cache)
  payload <- readRDS(cache)

  bad_schema <- tempfile(fileext = ".rds")
  payload$metadata$schema <- "annotations_cache/bad"
  saveRDS(payload, bad_schema)
  expect_error(load_annotations_cache(bad_schema), "Unsupported annotations cache schema")

  zero_shards <- tempfile(fileext = ".rds")
  saveRDS(
    list(
      metadata = list(
        schema = "annotations_cache/0.1",
        n_shards = 0L,
        shard_labels = character(),
        shard_checksums = character(),
        shard_start0 = numeric(),
        shard_stop0 = numeric()
      ),
      annomat = Matrix::Matrix(matrix(logical(), nrow = 0L, ncol = 1L), sparse = TRUE),
      annonames = "a"
    ),
    zero_shards
  )
  expect_error(load_annotations_cache(zero_shards), "n_shards must be at least 1")
})

.check_unique_variant_keys <- function(bp, a1_hash64, a2_hash64, where, object_name = "variant") {
  .validate_equal_lengths(
    list(bp = bp, a1_hash64 = a1_hash64, a2_hash64 = a2_hash64),
    "numeric variant key vector lengths mismatch"
  )
  if (length(bp) < 2L) {
    return(invisible(NULL))
  }
  ord <- order(bp, a1_hash64, a2_hash64)
  duplicate <- bp[ord][-1L] == bp[ord][-length(ord)] &
    a1_hash64[ord][-1L] == a1_hash64[ord][-length(ord)] &
    a2_hash64[ord][-1L] == a2_hash64[ord][-length(ord)]
  if (any(duplicate)) {
    stop(sprintf("%s: duplicate %s matching key", where, object_name), call. = FALSE)
  }
  invisible(NULL)
}

.match_variant_keys <- function(ref_bp, ref_a1_hash64, ref_a2_hash64,
                                src_bp, src_a1_hash64, src_a2_hash64,
                                shard_label, source_name) {
  ref_a1_hash64 <- bit64::as.integer64(ref_a1_hash64)
  ref_a2_hash64 <- bit64::as.integer64(ref_a2_hash64)
  src_a1_hash64 <- bit64::as.integer64(src_a1_hash64)
  src_a2_hash64 <- bit64::as.integer64(src_a2_hash64)

  .check_unique_variant_keys(
    ref_bp, ref_a1_hash64, ref_a2_hash64,
    sprintf("reference shard %s", shard_label),
    "reference"
  )
  .check_unique_variant_keys(
    src_bp, src_a1_hash64, src_a2_hash64,
    sprintf("source shard %s", shard_label),
    source_name
  )

  match_src <- rep.int(NA_integer_, length(ref_bp))
  if (!length(ref_bp) || !length(src_bp)) {
    return(match_src)
  }

  pairs <- .variant_key_pair_indices(
    ref_bp, ref_a1_hash64, ref_a2_hash64,
    src_bp, src_a1_hash64, src_a2_hash64
  )
  if (length(pairs$ref_index)) {
    match_src[pairs$ref_index] <- pairs$src_index
  }
  match_src
}

.count_variant_key_intersections <- function(ref_bp, ref_a1_hash64, ref_a2_hash64,
                                             src_bp, src_a1_hash64, src_a2_hash64) {
  ref_a1_hash64 <- bit64::as.integer64(ref_a1_hash64)
  ref_a2_hash64 <- bit64::as.integer64(ref_a2_hash64)
  src_a1_hash64 <- bit64::as.integer64(src_a1_hash64)
  src_a2_hash64 <- bit64::as.integer64(src_a2_hash64)

  .check_unique_variant_keys(
    ref_bp, ref_a1_hash64, ref_a2_hash64,
    "reference",
    "reference"
  )
  .check_unique_variant_keys(
    src_bp, src_a1_hash64, src_a2_hash64,
    "source",
    "variant"
  )

  if (!length(ref_bp) || !length(src_bp)) {
    return(0L)
  }
  length(.variant_key_pair_indices(
    ref_bp, ref_a1_hash64, ref_a2_hash64,
    src_bp, src_a1_hash64, src_a2_hash64
  )$ref_index)
}

.variant_key_pair_indices <- function(ref_bp, ref_a1_hash64, ref_a2_hash64,
                                      src_bp, src_a1_hash64, src_a2_hash64) {
  n_ref <- length(ref_bp)
  n_src <- length(src_bp)
  if (!n_ref || !n_src) {
    return(list(ref_index = integer(), src_index = integer()))
  }

  bp_all <- c(as.integer(ref_bp), as.integer(src_bp))
  a1_all <- c(bit64::as.integer64(ref_a1_hash64), bit64::as.integer64(src_a1_hash64))
  a2_all <- c(bit64::as.integer64(ref_a2_hash64), bit64::as.integer64(src_a2_hash64))
  is_source <- c(rep.int(FALSE, n_ref), rep.int(TRUE, n_src))
  local_index <- c(seq_len(n_ref), seq_len(n_src))

  # Sort one combined table and find adjacent equal tuple keys. This keeps the
  # allele hashes as bit64 integer64 values throughout and avoids string keys.
  ord <- order(bp_all, a1_all, a2_all)
  if (length(ord) < 2L) {
    return(list(ref_index = integer(), src_index = integer()))
  }

  bp_sorted <- bp_all[ord]
  a1_sorted <- a1_all[ord]
  a2_sorted <- a2_all[ord]
  source_sorted <- is_source[ord]
  index_sorted <- local_index[ord]

  left <- seq_len(length(ord) - 1L)
  right <- left + 1L
  same_key <- bp_sorted[left] == bp_sorted[right] &
    a1_sorted[left] == a1_sorted[right] &
    a2_sorted[left] == a2_sorted[right]
  opposite_side <- source_sorted[left] != source_sorted[right]
  pair_left <- left[same_key & opposite_side]
  if (!length(pair_left)) {
    return(list(ref_index = integer(), src_index = integer()))
  }
  pair_right <- pair_left + 1L
  left_is_source <- source_sorted[pair_left]
  list(
    ref_index = ifelse(left_is_source, index_sorted[pair_right], index_sorted[pair_left]),
    src_index = ifelse(left_is_source, index_sorted[pair_left], index_sorted[pair_right])
  )
}

.read_npz_ld_payload <- function(path, scratch_parent = NULL, expected_file_md5 = NULL) {
  extracted_dir <- .ld_npz_extracted_dir(path)
  if (.is_valid_extracted_npz_cache(path, extracted_dir, expected_file_md5)) {
    return(.read_extracted_npz_ld_payload(path, extracted_dir))
  }

  listing <- utils::unzip(path, list = TRUE)
  names_in_zip <- listing$Name
  required_files <- paste0(.ld_required_arrays, ".npy")
  missing <- setdiff(required_files, names_in_zip)
  if (length(missing)) {
    stop(sprintf("%s: missing required arrays: %s", path, paste(sub("\\.npy$", "", missing), collapse = ", ")), call. = FALSE)
  }
  if (is.null(scratch_parent)) {
    scratch_parent <- .npz_scratch_parent(path)
  }
  dir.create(scratch_parent, recursive = TRUE, showWarnings = FALSE)
  tmp <- tempfile("statgen_npz_", tmpdir = scratch_parent)
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  utils::unzip(path, files = required_files, exdir = tmp)
  arrays <- lapply(.ld_required_arrays, function(name) .read_npy_array(file.path(tmp, paste0(name, ".npy")), path, name))
  .npz_arrays_to_payload(arrays, path)
}

.ld_npz_extracted_dir <- function(path) {
  paste0(path, ".d")
}

.ld_npz_extracted_marker <- function(cache_dir) {
  file.path(cache_dir, ".statgen-extracted-npz.json")
}

.is_valid_extracted_npz_cache <- function(path, cache_dir, expected_file_md5 = NULL) {
  if (!dir.exists(cache_dir)) {
    return(FALSE)
  }
  required_files <- file.path(cache_dir, paste0(.ld_required_arrays, ".npy"))
  if (any(!file.exists(required_files) | dir.exists(required_files))) {
    return(FALSE)
  }
  marker_path <- .ld_npz_extracted_marker(cache_dir)
  if (!file.exists(marker_path) || dir.exists(marker_path)) {
    return(FALSE)
  }
  marker <- tryCatch(
    jsonlite::fromJSON(marker_path, simplifyVector = FALSE),
    error = function(e) NULL
  )
  if (!is.list(marker) ||
      !identical(marker$format, "statgen_extracted_npz") ||
      !identical(marker$source_file, basename(path))) {
    return(FALSE)
  }
  source_md5 <- if (is.null(expected_file_md5)) .md5_file(path) else expected_file_md5
  identical(marker$source_file_md5, source_md5)
}

.read_extracted_npz_ld_payload <- function(path, cache_dir) {
  arrays <- lapply(.ld_required_arrays, function(name) .read_npy_array(file.path(cache_dir, paste0(name, ".npy")), path, name))
  .npz_arrays_to_payload(arrays, path)
}

.npz_arrays_to_payload <- function(arrays, path) {
  names(arrays) <- .ld_required_arrays
  metadata_raw <- as.raw(arrays$metadata$data)
  metadata <- tryCatch(
    jsonlite::fromJSON(rawToChar(metadata_raw), simplifyVector = FALSE),
    error = function(e) stop(sprintf("%s: metadata must be UTF-8 JSON bytes", path), call. = FALSE)
  )
  list(
    data = arrays$data$data,
    indices = arrays$indices$data,
    indptr = arrays$indptr$data,
    shape = arrays$shape$data,
    a1freq = arrays$a1freq$data,
    metadata = metadata
  )
}

.extract_npz_ld_payload_cache <- function(path, expected_file_md5) {
  .require_ld_file(path)
  source_md5 <- .md5_file(path)
  if (!identical(source_md5, expected_file_md5)) {
    stop(sprintf("%s: file_md5 does not match manifest", path), call. = FALSE)
  }
  listing <- utils::unzip(path, list = TRUE)
  required_files <- paste0(.ld_required_arrays, ".npy")
  missing <- setdiff(required_files, listing$Name)
  if (length(missing)) {
    stop(sprintf("%s: missing required arrays: %s", path, paste(sub("\\.npy$", "", missing), collapse = ", ")), call. = FALSE)
  }

  cache_dir <- .ld_npz_extracted_dir(path)
  tmp <- tempfile(".statgen_npz_d_", tmpdir = dirname(path))
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  utils::unzip(path, files = required_files, exdir = tmp)
  extracted_files <- file.path(tmp, required_files)
  if (any(!file.exists(extracted_files) | dir.exists(extracted_files))) {
    stop(sprintf("%s: failed to extract required LD arrays", path), call. = FALSE)
  }
  marker <- list(
    format = "statgen_extracted_npz",
    schema_version = "1.0",
    source_file = basename(path),
    source_file_md5 = source_md5
  )
  marker_text <- jsonlite::toJSON(marker, auto_unbox = TRUE, pretty = TRUE)
  writeLines(marker_text, .ld_npz_extracted_marker(tmp), useBytes = TRUE)

  if (file.exists(cache_dir)) {
    unlink(cache_dir, recursive = TRUE)
  }
  if (!file.rename(tmp, cache_dir)) {
    stop(sprintf("Failed to replace extracted LD cache: %s", cache_dir), call. = FALSE)
  }
  invisible(cache_dir)
}

.npz_scratch_parent <- function(path) {
  env_root <- Sys.getenv("STATGEN_SCRATCH", unset = "")
  if (nzchar(env_root)) {
    dir.create(env_root, recursive = TRUE, showWarnings = FALSE)
    if (dir.exists(env_root)) {
      return(env_root)
    }
    stop(sprintf("Cannot create STATGEN_SCRATCH directory: %s", env_root), call. = FALSE)
  }
  root <- dirname(path)
  if (dir.exists(root) && file.access(root, 2L) == 0L) {
    return(root)
  }
  warning("STATGEN_SCRATCH is unset and LD .npz directory is not writable; using system temporary directory for extraction", call. = FALSE)
  tempdir()
}

.read_npy_array <- function(file, archive_path, array_name) {
  con <- file(file, "rb")
  on.exit(close(con), add = TRUE)
  magic <- readBin(con, "raw", n = 6L)
  if (!identical(magic, as.raw(c(0x93, charToRaw("NUMPY"))))) {
    stop(sprintf("%s: %s is not an NPY array", archive_path, array_name), call. = FALSE)
  }
  version <- readBin(con, "integer", n = 2L, size = 1L, signed = FALSE, endian = "little")
  if (length(version) != 2L || !(version[[1L]] %in% c(1L, 2L, 3L))) {
    stop(sprintf("%s: unsupported NPY version for %s", archive_path, array_name), call. = FALSE)
  }
  header_len <- if (version[[1L]] == 1L) {
    readBin(con, "integer", n = 1L, size = 2L, signed = FALSE, endian = "little")
  } else {
    readBin(con, "integer", n = 1L, size = 4L, signed = FALSE, endian = "little")
  }
  header_raw <- readBin(con, "raw", n = header_len)
  header <- rawToChar(header_raw)
  descr <- .parse_npy_header_value(header, "descr")
  fortran <- .parse_npy_header_value(header, "fortran_order")
  shape <- .parse_npy_shape(header)
  if (!identical(fortran, "False")) {
    stop(sprintf("%s: %s must not be Fortran-ordered", archive_path, array_name), call. = FALSE)
  }
  expected_descr <- switch(
    array_name,
    data = "<f4",
    indices = "<i4",
    indptr = "<i4",
    shape = "<i8",
    a1freq = "<f4",
    metadata = "|u1",
    stop(sprintf("%s: unsupported array %s", archive_path, array_name), call. = FALSE)
  )
  if (!identical(descr, expected_descr)) {
    stop(sprintf("%s: %s must have dtype %s, got %s", archive_path, array_name, expected_descr, descr), call. = FALSE)
  }
  count <- prod(shape)
  if (!is.finite(count) || count < 0 || count > .Machine$integer.max) {
    stop(sprintf("%s: %s shape is unsupported", archive_path, array_name), call. = FALSE)
  }
  data <- switch(
    descr,
    "<f4" = readBin(con, "numeric", n = as.integer(count), size = 4L, endian = "little"),
    "<i4" = readBin(con, "integer", n = as.integer(count), size = 4L, signed = TRUE, endian = "little"),
    "|u1" = readBin(con, "integer", n = as.integer(count), size = 1L, signed = FALSE, endian = "little"),
    "<i8" = .read_npy_int64_as_integer(con, as.integer(count), archive_path, array_name)
  )
  if (length(data) != count) {
    stop(sprintf("%s: short read for %s", archive_path, array_name), call. = FALSE)
  }
  list(data = data, shape = shape, descr = descr)
}

.parse_npy_header_value <- function(header, key) {
  pattern <- sprintf("'%s': ([^,}]+)", key)
  hit <- regmatches(header, regexpr(pattern, header, perl = TRUE))
  if (!length(hit) || hit == "") {
    stop(sprintf("Malformed NPY header: missing %s", key), call. = FALSE)
  }
  value <- sub(pattern, "\\1", hit, perl = TRUE)
  value <- trimws(value)
  sub("^'([^']*)'$", "\\1", value)
}

.parse_npy_shape <- function(header) {
  hit <- regmatches(header, regexpr("'shape': \\(([^)]*)\\)", header, perl = TRUE))
  if (!length(hit) || hit == "") {
    stop("Malformed NPY header: missing shape", call. = FALSE)
  }
  inside <- sub(".*'shape': \\(([^)]*)\\).*", "\\1", hit, perl = TRUE)
  parts <- trimws(strsplit(inside, ",", fixed = TRUE)[[1L]])
  parts <- parts[nzchar(parts)]
  if (!length(parts)) {
    return(integer())
  }
  out <- suppressWarnings(as.integer(parts))
  if (anyNA(out) || any(out < 0L)) {
    stop("Malformed NPY header: invalid shape", call. = FALSE)
  }
  out
}

.read_npy_int64_as_integer <- function(con, n, archive_path, array_name) {
  if (!n) {
    return(integer())
  }
  raw <- readBin(con, "raw", n = n * 8L)
  if (length(raw) != n * 8L) {
    stop(sprintf("%s: short read for %s", archive_path, array_name), call. = FALSE)
  }
  bytes <- matrix(as.integer(raw), nrow = 8L)
  negative <- bytes[8L, ] >= 128L
  high <- bytes[5L, ] != 0L | bytes[6L, ] != 0L | bytes[7L, ] != 0L | bytes[8L, ] != 0L
  exceeds_r_integer <- bytes[4L, ] >= 128L
  if (any(negative | high | exceeds_r_integer)) {
    stop(sprintf("%s: %s int64 values exceed the supported R integer range", archive_path, array_name), call. = FALSE)
  }
  values <- bytes[1L, ] + bytes[2L, ] * 256L + bytes[3L, ] * 65536L + bytes[4L, ] * 16777216L
  as.integer(values)
}

.reference_cache_schema <- "reference_cache/0.1"

num_snp <- function(x, ...) UseMethod("num_snp")
shards <- function(x, ...) UseMethod("shards")
chr <- function(x, ...) UseMethod("chr")
snp <- function(x, ...) UseMethod("snp")
bp <- function(x, ...) UseMethod("bp")
a1 <- function(x, ...) UseMethod("a1")
a2 <- function(x, ...) UseMethod("a2")
a1_hash64 <- function(x, ...) UseMethod("a1_hash64")
a2_hash64 <- function(x, ...) UseMethod("a2_hash64")
is_single_nucleotide_variant <- function(x, ...) UseMethod("is_single_nucleotide_variant")
is_strand_ambiguous <- function(x, ...) UseMethod("is_strand_ambiguous")
shard_offsets <- function(x, ...) UseMethod("shard_offsets")
select_shards <- function(x, shards, ...) UseMethod("select_shards")
is_object_compatible <- function(reference, object, ...) UseMethod("is_object_compatible")
validate_checksums <- function(x, ...) UseMethod("validate_checksums")
save_cache <- function(x, path, ...) UseMethod("save_cache")

load_reference <- function(path, shards = NULL) {
  path <- .validate_path_scalar(path, "path")
  if (grepl("@", path, fixed = TRUE)) {
    available_labels <- character()
    available_paths <- character()
    for (label in .canonical_chr_order) {
      candidate <- gsub("@", label, path, fixed = TRUE)
      if (file.exists(candidate)) {
        available_labels <- c(available_labels, label)
        available_paths <- c(available_paths, candidate)
      }
    }
    if (!length(available_labels)) {
      stop(sprintf("No BIM shards found matching template: %s", path), call. = FALSE)
    }
    selected <- .validate_requested_shards(shards, available_labels, "load_reference")
    out <- vector("list", length(selected))
    for (i in seq_along(selected)) {
      label <- selected[[i]]
      bim <- .parse_bim(available_paths[[match(label, available_labels)]])
      out[[i]] <- .new_reference_shard(
        label, bim$chr, bim$snp, bim$bp, bim$a1, bim$a2,
        bim$a1_hash64, bim$a2_hash64
      )
    }
    return(.new_reference_panel(out))
  }

  bim <- .parse_bim(path)
  available_labels <- .canonical_chr_order[vapply(
    .canonical_chr_order,
    function(label) any(bim$chr == label),
    logical(1)
  )]
  selected <- .validate_requested_shards(shards, available_labels, "load_reference")
  out <- vector("list", length(selected))
  for (i in seq_along(selected)) {
    label <- selected[[i]]
    keep <- bim$chr == label
    out[[i]] <- .new_reference_shard(
      label, bim$chr[keep], bim$snp[keep], bim$bp[keep], bim$a1[keep], bim$a2[keep],
      bim$a1_hash64[keep], bim$a2_hash64[keep]
    )
  }
  .new_reference_panel(out)
}

save_reference_cache <- function(panel, path) {
  if (!inherits(panel, "ReferencePanel")) {
    stop("panel must be a ReferencePanel", call. = FALSE)
  }
  path <- .validate_path_scalar(path, "path")
  offsets <- shard_offsets(panel)
  metadata <- list(
    schema = .reference_cache_schema,
    n_shards = length(panel$shards),
    shard_labels = vapply(panel$shards, function(s) s$label, character(1)),
    shard_checksums = vapply(panel$shards, function(s) s$checksum, character(1)),
    shard_start0 = offsets$start0,
    shard_stop0 = offsets$stop0
  )
  payload <- list(
    metadata = metadata,
    bp = bp(panel),
    snp_text_by_shard = vapply(panel$shards, function(s) .encode_text_payload(snp(s)), character(1)),
    a1_text_by_shard = vapply(panel$shards, function(s) .encode_text_payload(a1(s)), character(1)),
    a2_text_by_shard = vapply(panel$shards, function(s) .encode_text_payload(a2(s)), character(1)),
    a1_hash64 = a1_hash64(panel),
    a2_hash64 = a2_hash64(panel)
  )
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  saveRDS(payload, path)
  invisible(NULL)
}

load_reference_cache <- function(path, shards = NULL) {
  path <- .validate_path_scalar(path, "path")
  payload <- readRDS(path)
  .validate_reference_cache_payload(payload)
  meta <- payload$metadata
  selected <- .validate_requested_shards(shards, meta$shard_labels, "load_reference_cache")
  label_to_index <- stats::setNames(seq_along(meta$shard_labels), meta$shard_labels)

  out <- vector("list", length(selected))
  for (i in seq_along(selected)) {
    label <- selected[[i]]
    idx <- label_to_index[[label]]
    start <- meta$shard_start0[[idx]] + 1L
    stop <- meta$shard_stop0[[idx]]
    row_count <- stop - meta$shard_start0[[idx]]
    out[[i]] <- .new_reference_shard_from_cache(
      label = label,
      num_snp = row_count,
      bp = payload$bp[start:stop],
      snp_text = payload$snp_text_by_shard[[idx]],
      a1_text = payload$a1_text_by_shard[[idx]],
      a2_text = payload$a2_text_by_shard[[idx]],
      checksum = meta$shard_checksums[[idx]],
      a1_hash64 = payload$a1_hash64[start:stop],
      a2_hash64 = payload$a2_hash64[start:stop]
    )
  }
  .new_reference_panel(out)
}

num_snp.ReferenceShard <- function(x, ...) x$num_snp
num_snp.ReferencePanel <- function(x, ...) x$num_snp
shards.ReferencePanel <- function(x, ...) x$shards
shards.default <- function(x, ...) {
  stop("object has no shards accessor", call. = FALSE)
}

chr.ReferenceShard <- function(x, ...) rep.int(x$label, x$num_snp)
chr.ReferencePanel <- function(x, ...) unlist(lapply(x$shards, chr), use.names = FALSE)

snp.ReferenceShard <- function(x, ...) .shard_string_vector(x, "snp")
snp.ReferencePanel <- function(x, ...) unlist(lapply(x$shards, snp), use.names = FALSE)

bp.ReferenceShard <- function(x, ...) x$bp
bp.ReferencePanel <- function(x, ...) unlist(lapply(x$shards, bp), use.names = FALSE)

a1.ReferenceShard <- function(x, ...) .shard_string_vector(x, "a1")
a1.ReferencePanel <- function(x, ...) unlist(lapply(x$shards, a1), use.names = FALSE)

a2.ReferenceShard <- function(x, ...) .shard_string_vector(x, "a2")
a2.ReferencePanel <- function(x, ...) unlist(lapply(x$shards, a2), use.names = FALSE)

a1_hash64.ReferenceShard <- function(x, ...) x$a1_hash64
a1_hash64.ReferencePanel <- function(x, ...) do.call(c, lapply(x$shards, a1_hash64))

a2_hash64.ReferenceShard <- function(x, ...) x$a2_hash64
a2_hash64.ReferencePanel <- function(x, ...) do.call(c, lapply(x$shards, a2_hash64))

is_single_nucleotide_variant.ReferenceShard <- function(x, ...) {
  .is_single_nucleotide_variant_hash(a1_hash64(x), a2_hash64(x))
}

is_single_nucleotide_variant.ReferencePanel <- function(x, ...) {
  unlist(lapply(x$shards, is_single_nucleotide_variant), use.names = FALSE)
}

is_strand_ambiguous.ReferenceShard <- function(x, ...) {
  .is_strand_ambiguous_hash(a1_hash64(x), a2_hash64(x))
}

is_strand_ambiguous.ReferencePanel <- function(x, ...) {
  unlist(lapply(x$shards, is_strand_ambiguous), use.names = FALSE)
}

shard_offsets.ReferencePanel <- function(x, ...) x$shard_offsets

select_shards.ReferencePanel <- function(x, shards, ...) {
  available <- vapply(x$shards, function(s) s$label, character(1))
  selected <- .validate_requested_shards(shards, available, "ReferencePanel.select_shards")
  by_label <- stats::setNames(x$shards, available)
  .new_reference_panel(unname(by_label[selected]))
}

validate_checksums.ReferencePanel <- function(x, ...) {
  for (shard in x$shards) {
    computed <- .checksum_from_arrays(chr(shard), bp(shard), a1(shard), a2(shard))
    if (!identical(computed, shard$checksum)) {
      stop(sprintf("Reference checksum mismatch for shard %s", shard$label), call. = FALSE)
    }
  }
  TRUE
}

is_object_compatible.ReferencePanel <- function(reference, object, ...) {
  obj_shards <- tryCatch(shards(object), error = function(e) NULL)
  if (is.null(obj_shards)) {
    warning("statgen: is_object_compatible: object has no shards accessor", call. = FALSE)
    return(FALSE)
  }
  if (length(obj_shards) != length(reference$shards)) {
    warning(
      sprintf(
        "statgen: is_object_compatible: shard count mismatch: reference has %d, object has %d",
        length(reference$shards), length(obj_shards)
      ),
      call. = FALSE
    )
    return(FALSE)
  }

  ok <- TRUE
  for (i in seq_along(reference$shards)) {
    ref_s <- reference$shards[[i]]
    obj_s <- obj_shards[[i]]
    obj_label <- obj_s$label
    if (is.null(obj_label) || !identical(obj_label, ref_s$label)) {
      warning(
        sprintf(
          "statgen: is_object_compatible: shard label mismatch: reference %s, object %s",
          ref_s$label,
          if (is.null(obj_label)) "<missing>" else as.character(obj_label)
        ),
        call. = FALSE
      )
      ok <- FALSE
      next
    }
    obj_num <- tryCatch(num_snp(obj_s), error = function(e) obj_s$num_snp)
    if (is.null(obj_num) || !identical(as.integer(obj_num), as.integer(ref_s$num_snp))) {
      warning(
        sprintf("statgen: is_object_compatible: shard %s: row count mismatch", ref_s$label),
        call. = FALSE
      )
      ok <- FALSE
      next
    }
    obj_checksum <- obj_s$reference_checksum
    if (is.null(obj_checksum)) {
      obj_checksum <- obj_s$checksum
    }
    if (!is.null(obj_checksum) && !identical(obj_checksum, ref_s$checksum)) {
      warning(
        sprintf("statgen: is_object_compatible: shard %s: reference_checksum mismatch", ref_s$label),
        call. = FALSE
      )
      ok <- FALSE
    }
  }
  ok
}

is_object_compatible.default <- function(reference, object, ...) {
  warning("statgen: is_object_compatible: reference must be a ReferencePanel", call. = FALSE)
  FALSE
}

save_cache.ReferencePanel <- function(x, path, ...) save_reference_cache(x, path)

print.ReferenceShard <- function(x, ...) {
  cat(sprintf("<ReferenceShard %s: %d SNPs>\n", x$label, x$num_snp))
  invisible(x)
}

print.ReferencePanel <- function(x, ...) {
  labels <- vapply(x$shards, function(s) s$label, character(1))
  cat(sprintf("<ReferencePanel: %d SNPs across %d shard(s): %s>\n", x$num_snp, length(labels), paste(labels, collapse = ", ")))
  invisible(x)
}

.new_reference_shard <- function(label, chr, snp, bp, a1, a2, a1_hash64 = NULL, a2_hash64 = NULL) {
  chr <- as.character(chr)
  snp <- as.character(snp)
  bp <- as.integer(bp)
  a1 <- as.character(a1)
  a2 <- as.character(a2)
  .validate_equal_lengths(list(chr = chr, snp = snp, bp = bp, a1 = a1, a2 = a2), "ReferenceShard vector lengths must match")
  if (length(chr) && any(chr != label)) {
    stop(sprintf("Reference shard %s contains multiple chr labels", label), call. = FALSE)
  }
  if (is.null(a1_hash64)) {
    a1_hash64 <- .allele_hash64(a1)
  } else {
    a1_hash64 <- bit64::as.integer64(a1_hash64)
  }
  if (is.null(a2_hash64)) {
    a2_hash64 <- .allele_hash64(a2)
  } else {
    a2_hash64 <- bit64::as.integer64(a2_hash64)
  }
  .validate_equal_lengths(
    list(bp = bp, a1_hash64 = a1_hash64, a2_hash64 = a2_hash64),
    "ReferenceShard allele hash vector lengths must match"
  )
  structure(
    list(
      label = label,
      num_snp = length(chr),
      snp = snp,
      bp = bp,
      a1 = a1,
      a2 = a2,
      a1_hash64 = a1_hash64,
      a2_hash64 = a2_hash64,
      checksum = .checksum_from_arrays(chr, bp, a1, a2)
    ),
    class = "ReferenceShard"
  )
}

.new_reference_shard_from_cache <- function(label, num_snp, bp, snp_text, a1_text, a2_text, checksum, a1_hash64, a2_hash64) {
  bp <- as.integer(bp)
  a1_hash64 <- bit64::as.integer64(a1_hash64)
  a2_hash64 <- bit64::as.integer64(a2_hash64)
  if (length(bp) != num_snp || length(a1_hash64) != num_snp || length(a2_hash64) != num_snp) {
    stop("Invalid reference cache: shard vector lengths mismatch", call. = FALSE)
  }
  structure(
    list(
      label = label,
      num_snp = as.integer(num_snp),
      snp_text = snp_text,
      bp = bp,
      a1_text = a1_text,
      a2_text = a2_text,
      a1_hash64 = a1_hash64,
      a2_hash64 = a2_hash64,
      checksum = checksum
    ),
    class = "ReferenceShard"
  )
}

.new_reference_panel <- function(shard_list) {
  if (!length(shard_list)) {
    stop("ReferencePanel requires at least one shard", call. = FALSE)
  }
  pos <- 0L
  offsets <- data.frame(
    shard_label = character(length(shard_list)),
    start0 = integer(length(shard_list)),
    stop0 = integer(length(shard_list)),
    stringsAsFactors = FALSE
  )
  for (i in seq_along(shard_list)) {
    offsets$shard_label[[i]] <- shard_list[[i]]$label
    offsets$start0[[i]] <- pos
    pos <- pos + shard_list[[i]]$num_snp
    offsets$stop0[[i]] <- pos
  }
  structure(
    list(
      shards = shard_list,
      num_snp = pos,
      shard_offsets = offsets
    ),
    class = "ReferencePanel"
  )
}

.validate_reference_cache_payload <- function(payload) {
  cache <- .validate_cache_payload_metadata(payload, .reference_cache_schema, "reference")
  meta <- cache$meta
  n_shards <- cache$n_shards
  total <- cache$total
  payload_fields <- c("bp", "snp_text_by_shard", "a1_text_by_shard", "a2_text_by_shard", "a1_hash64", "a2_hash64")
  for (field in payload_fields) {
    if (is.null(payload[[field]])) {
      stop(sprintf("Invalid reference cache: missing %s", field), call. = FALSE)
    }
  }
  if (length(payload$snp_text_by_shard) != n_shards || length(payload$a1_text_by_shard) != n_shards || length(payload$a2_text_by_shard) != n_shards) {
    stop("Invalid reference cache: string payload shard count mismatch", call. = FALSE)
  }
  if (length(payload$bp) != total || length(payload$a1_hash64) != total || length(payload$a2_hash64) != total) {
    stop("Invalid reference cache: panel-wide vector lengths mismatch", call. = FALSE)
  }
}

.checksum_from_arrays <- function(chr, bp, a1, a2) {
  text <- paste0(chr, ":", as.integer(bp), ":", a1, ":", a2, "\n", collapse = "")
  digest::digest(enc2utf8(text), algo = "md5", serialize = FALSE)
}

.encode_text_payload <- function(values) {
  paste(as.character(values), collapse = "\n")
}

.decode_text_payload <- function(payload, expected_len, field) {
  if (!is.character(payload) || length(payload) != 1L || is.na(payload)) {
    stop(sprintf("Invalid reference cache: %s text payload must be a character scalar", field), call. = FALSE)
  }
  values <- if (identical(payload, "")) character() else strsplit(payload, "\n", fixed = TRUE)[[1]]
  if (length(values) != expected_len) {
    stop(sprintf("Invalid reference cache: decoded %s length mismatch", field), call. = FALSE)
  }
  values
}

.shard_string_vector <- function(x, field) {
  direct <- x[[field]]
  if (!is.null(direct)) {
    return(direct)
  }
  .decode_text_payload(x[[paste0(field, "_text")]], x$num_snp, field)
}

test_that("load_reference reads bundled BIM fixture", {
  path <- system.file("extdata", "reference_chr1.bim", package = "statgen", mustWork = TRUE)
  ref <- load_reference(path)

  expect_s3_class(ref, "ReferencePanel")
  expect_equal(num_snp(ref), 5L)
  expect_equal(vapply(shards(ref), function(s) s$label, character(1)), "1")
  expect_equal(bp(ref), c(100L, 200L, 300L, 400L, 500L))
  expect_equal(snp(ref)[1:2], c("rs1001", "rs1002"))
  expect_equal(a1(ref)[1], "A")
  expect_equal(a2(ref)[1], "G")
  expect_true(inherits(a1_hash64(ref), "integer64"))
  expect_true(inherits(a2_hash64(ref), "integer64"))
})

test_that("reference cache round-trips and subsets", {
  path <- system.file("extdata", "reference_chr1.bim", package = "statgen", mustWork = TRUE)
  ref <- load_reference(path)
  cache <- tempfile(fileext = ".rds")

  save_reference_cache(ref, cache)
  loaded <- load_reference_cache(cache)

  expect_equal(num_snp(loaded), num_snp(ref))
  expect_equal(chr(loaded), chr(ref))
  expect_equal(snp(loaded), snp(ref))
  expect_equal(bp(loaded), bp(ref))
  expect_equal(a1(loaded), a1(ref))
  expect_equal(a2(loaded), a2(ref))
  expect_equal(as.character(a1_hash64(loaded)), as.character(a1_hash64(ref)))
  expect_true(validate_checksums(loaded))

  subset <- load_reference_cache(cache, shards = "1")
  expect_equal(num_snp(subset), 5L)
})

---
title: "Using statgen in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using statgen in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

`statgen` stores each analysis object on a shared reference SNP axis. The R
package loads reference panels, summary statistics, annotations, genotypes, and
LD panels into those coordinates.

```{r}
library(statgen)
set_verbosity("quiet")
```

## Reference, Sumstats, And Annotations

The bundled package fixtures are intentionally tiny so that examples and CRAN
checks do not require external tools or large data.

This vignette reads those installed fixtures with
`system.file(..., package = "statgen")`. They are demonstration inputs only. In
real analyses, replace these paths with post-harmonization reference, summary
statistics, annotation, genotype, and LD files prepared for one analysis input
set. For an end-to-end repository workflow that prepares source inputs and then
runs the same R analysis pattern on those files, see
`docs/TUTORIAL_1_PREPARE_DATA.md` and `docs/TUTORIAL_4_R.md`.

```{r}
reference_template <- file.path(
  dirname(system.file("extdata", "reference_chr1.bim", package = "statgen")),
  "reference_chr@.bim"
)
reference <- load_reference(reference_template)

sumstats_path <- system.file("extdata", "traits_complete.tsv.gz", package = "statgen")
sumstats <- load_sumstats(sumstats_path, reference)

annotation_paths <- system.file(
  "extdata",
  c("anno1.bed", "anno2.bed"),
  package = "statgen"
)
annotations <- load_annotations(annotation_paths, reference)

num_snp(reference)
head(logpvec(sumstats))
colnames(annomat(annotations))
```

R-native object caches are RDS files. They are useful for repeated analyses in
the same R runtime.

```{r}
sumstats_cache <- tempfile(fileext = ".rds")
save_sumstats_cache(sumstats, sumstats_cache)
sumstats_cached <- load_sumstats_cache(sumstats_cache)
identical(is_present(sumstats_cached), is_present(sumstats))
```

## LD Panels

Python builds the canonical LD `.npz` distribution. For package examples,
`ld_root` points to a bundled toy distribution that already includes the R
reference-cache sidecar. For real analyses, run `prepare_ld_npz_for_r()` once
on the Python-built `ld_npz/` directory for the same harmonized input set, then
load that directory with `load_ld()`.

```{r}
ld_root <- system.file("extdata", "ld", package = "statgen")
ld <- load_ld(ld_root)

num_snp(ld)
a1freq(ld)
```

`multiply_r2()` multiplies aligned vectors or matrices by LD `r^2`.
`fast_prune()` applies greedy LD pruning to aligned scores.

```{r}
scores <- seq_len(num_snp(ld))
multiply_r2(ld, scores)

logp <- rev(seq_len(num_snp(ld)))
fast_prune(logp, ld, r2_threshold = 0.35)
```

## Genotype Metadata And Fetching

Genotype panels keep PLINK metadata and fetch hardcalls on demand. The original
BED files must remain available unless a replacement `bed_path` is supplied.

```{r}
ref_chr1 <- load_reference(system.file("extdata", "reference_chr1.bim", package = "statgen"))
bed_chr1 <- system.file("extdata", "genotype_1.bed", package = "statgen")
genotype <- load_genotype(sub("\\.bed$", "", bed_chr1), ref_chr1)

geno <- fetch_genotypes_int8(genotype, c(1, 3))
dim(geno)
head(geno)
```

\name{genotype}
\alias{load_genotype}
\alias{save_genotype_cache}
\alias{load_genotype_cache}
\alias{num_sample}
\alias{fid}
\alias{iid}
\alias{father_id}
\alias{mother_id}
\alias{sex}
\alias{is_male}
\alias{is_female}
\alias{ploidy_male}
\alias{ploidy_female}
\alias{source_layout}
\alias{source_row0}
\alias{is_subject_present}
\alias{fetch_genotypes_int8}
\alias{fetch_genotypes}
\title{GenotypePanel objects}
\description{
Load PLINK bfile-backed genotype metadata, inspect reference-aligned genotype
panels, cache metadata, and fetch hardcall genotype slices from BED files.
}
\usage{
load_genotype(bfile_prefix, reference)
save_genotype_cache(panel, path)
load_genotype_cache(path, shards = NULL)

num_sample(x, ...)
fid(x, ...)
iid(x, ...)
father_id(x, ...)
mother_id(x, ...)
sex(x, ...)
is_male(x, ...)
is_female(x, ...)
ploidy_male(x, ...)
ploidy_female(x, ...)
source_layout(x, ...)
source_row0(x, ...)
is_subject_present(x, shard, ...)

fetch_genotypes_int8(x, snp_indices, bed_path = NULL, ...)
fetch_genotypes(x, snp_indices, bed_path = NULL,
  haploid_mode = NULL, ...)
}
\arguments{
\item{bfile_prefix}{Character scalar PLINK bfile prefix, without suffixes.
Use \code{@} as a shard-label placeholder for sharded input.}
\item{reference}{A \code{ReferencePanel}.}
\item{panel, x}{A \code{GenotypePanel}.}
\item{path}{Character scalar path to an RDS genotype metadata cache.}
\item{shards}{Optional character vector of shard labels in canonical order.}
\item{shard}{A loaded shard label for subject-presence lookup.}
\item{snp_indices}{One-based panel-global SNP indices.}
\item{bed_path}{Optional BED path override. A sharded override may contain
\code{@}.}
\item{haploid_mode}{Optional output mode, \code{"raw"} or
\code{"ploidy_scaled"}. The default is \code{"raw"}.}
\item{...}{Reserved for S3 methods.}
}
\value{
\code{load_genotype()} and \code{load_genotype_cache()} return
\code{GenotypePanel} S3 objects. Metadata accessors return ordinary R vectors.
\code{fetch_genotypes_int8()} returns an integer matrix with samples as rows
and requested SNPs as columns, using \code{-1} for missing hardcalls.
\code{fetch_genotypes()} returns a numeric matrix with missing calls as
\code{NaN}.
}
\details{
Genotype panels are reference-aligned metadata handles for PLINK BED payloads.
Hardcalls are decoded on demand; caches store only metadata needed to locate
and map BED rows.
}
\examples{
ref_path <- system.file("extdata", "reference_chr1.bim", package = "statgen")
geno_bed <- system.file("extdata", "genotype_1.bed", package = "statgen")
geno_prefix <- sub("\\\\.bed$", "", geno_bed)
ref <- load_reference(ref_path)
g <- load_genotype(geno_prefix, ref)
geno <- fetch_genotypes(g, c(1, 2))
dim(geno)
head(geno)

cache <- tempfile(fileext = ".rds")
save_genotype_cache(g, cache)
g_cached <- load_genotype_cache(cache)
num_sample(g_cached)
sum(is_present(g_cached))
}

\name{version}
\alias{version}
\title{Return the statgen Package Version}
\usage{
version()
}
\value{
A character scalar containing the installed package version.
}
\description{
Returns the installed statgen R package version.
}
\examples{
version()
}

\name{reference}
\alias{load_reference}
\alias{save_reference_cache}
\alias{load_reference_cache}
\alias{num_snp}
\alias{shards}
\alias{chr}
\alias{snp}
\alias{bp}
\alias{a1}
\alias{a2}
\alias{a1_hash64}
\alias{a2_hash64}
\alias{is_single_nucleotide_variant}
\alias{is_strand_ambiguous}
\alias{shard_offsets}
\alias{select_shards}
\alias{is_object_compatible}
\alias{validate_checksums}
\alias{save_cache}
\title{ReferencePanel objects}
\description{
Load PLINK BIM reference panels as statgen \code{ReferencePanel} S3 objects,
inspect aligned reference vectors, and save or load R-native reference caches.
}
\usage{
# Constructors and cache I/O
load_reference(path, shards = NULL)
load_reference_cache(path, shards = NULL)
save_reference_cache(panel, path)

# Accessors
num_snp(x, ...)
shards(x, ...)
chr(x, ...)
snp(x, ...)
bp(x, ...)
a1(x, ...)
a2(x, ...)
a1_hash64(x, ...)
a2_hash64(x, ...)
is_single_nucleotide_variant(x, ...)
is_strand_ambiguous(x, ...)
shard_offsets(x, ...)

# Operations
select_shards(x, shards, ...)
is_object_compatible(reference, object, ...)
validate_checksums(x, ...)
save_cache(x, path, ...)
}
\arguments{
\item{path}{Character scalar path. For \code{load_reference()}, this is a BIM
file or sharded BIM template containing \code{@}. For
\code{load_reference_cache()}, this is an RDS reference cache. For cache
writers, this is the output cache path.}
\item{shards}{Optional character vector of shard labels in canonical order.}
\item{panel, reference, x, object}{Reference panel or compatible statgen object.}
\item{...}{Reserved for S3 methods.}
}
\details{
Reference panels are represented as \code{ReferencePanel} S3 objects backed by
an ordered list of reference shards.

Most exported functions in this API are S3 generics operating on
\code{ReferencePanel} objects and other reference-aligned statgen objects.
Users should access panel data through accessor functions such as \code{bp(x)}
and \code{a1(x)} rather than directly accessing internal fields.

\code{is_object_compatible(reference, object)} dispatches on the first argument,
which should be a \code{ReferencePanel}. Other reference-aligned panel classes
participate by providing \code{shards()} methods that return their ordered shard
objects.

\code{shards(x)} returns the ordered shard objects. \code{save_cache()} is the
cross-object cache-writing generic; \code{save_reference_cache()} is the
explicit reference-panel cache writer.
}
\value{
\code{load_reference()} and \code{load_reference_cache()} return a
\code{ReferencePanel} S3 object. Accessors return ordinary R vectors or data
frames; allele hashes are \code{bit64::integer64} vectors.
}
\examples{
path <- system.file("extdata", "reference_chr1.bim", package = "statgen")
ref <- load_reference(path)
num_snp(ref)
head(snp(ref))
sum(is_strand_ambiguous(ref))

cache <- tempfile(fileext = ".rds")
save_reference_cache(ref, cache)
ref_cached <- load_reference_cache(cache)
num_snp(ref_cached)

save_cache(ref, cache)
}

\name{statgen-package}
\alias{statgen-package}
\docType{package}
\title{Statistical Genetics Data Objects and Loaders}
\description{
CRAN-compatible R runtime for statgen statistical genetics data objects.
}
\details{
\pkg{statgen} provides R objects and loaders for reference-aligned statistical
genetics data: reference variants, genotypes, linkage disequilibrium (LD),
annotations, and GWAS summary statistics. The package is designed for analysis
workflows where genome build, contig labels, and allele orientation have already
been harmonized upstream.

The central object is a \code{ReferencePanel}, which defines the ordered SNP
axis used by the other panel types. \code{LDPanel}, \code{GenotypePanel},
\code{AnnotationPanel}, and \code{Sumstats} objects expose data in those
reference coordinates, so operations can combine objects without silently
renaming contigs, swapping alleles, or dropping unmatched variants.

Typical workflows start by loading or caching a reference panel with
\code{\link{load_reference}}, then loading aligned data with functions such as
\code{\link{load_ld}}, \code{\link{load_genotype}},
\code{\link{load_annotations}}, or \code{\link{load_sumstats}}. Panel objects
provide accessors for aligned vectors and matrices, \code{\link{select_shards}}
for canonical chromosome subsetting, and cache helpers for repeated analyses.

LD distributions are directory-based runtime artifacts discovered through
\code{ld_manifest.json}; they are loaded with \code{\link{load_ld}} rather than
as individual shard files. R reads the Python \code{.npz} LD shard format
directly after \code{\link{prepare_ld_npz_for_r}} adds an R reference-cache
sidecar. R-native object caches use RDS files, while portable source inputs
such as PLINK BIM/BED/FAM, BED annotations, and summary-statistics TSV files
remain external inputs.
}
\seealso{
\code{\link{load_reference}}, \code{\link{load_ld}},
\code{\link{load_genotype}}, \code{\link{load_annotations}},
\code{\link{load_sumstats}}, \code{\link{version}},
\code{\link{set_verbosity}}
}

\name{verbosity}
\alias{get_verbosity}
\alias{set_verbosity}
\title{Get or Set Runtime Verbosity}
\usage{
get_verbosity()
set_verbosity(level)
}
\arguments{
\item{level}{Character scalar, either \code{"quiet"} or \code{"info"}.}
}
\value{
\code{get_verbosity()} returns the current verbosity level. \code{set_verbosity()}
returns the selected level invisibly.
}
\description{
Controls the package-wide statgen runtime verbosity setting.
}
\examples{
old <- get_verbosity()
set_verbosity("quiet")
get_verbosity()
set_verbosity(old)
}

\name{ld}
\alias{load_ld}
\alias{load_ld_reference}
\alias{validate_ld_distribution}
\alias{prepare_ld_npz_for_r}
\alias{a1freq}
\alias{multiply_r2}
\alias{fast_prune}
\alias{reference}
\alias{default_chrX_sex}
\title{LD panel loading, preparation, and operations}
\usage{
load_ld(path, shards = NULL, default_chrX_sex = NULL, retain_ld_r = TRUE)
load_ld_reference(path, shards = NULL)
validate_ld_distribution(path, check_payload_structure = FALSE)
prepare_ld_npz_for_r(npz_root, extract_npz = FALSE)
a1freq(x, chrX_sex = NULL, ...)
multiply_r2(ld_panel, M, chrX_sex = NULL, ...)
fast_prune(logpvec, ld_panel, r2_threshold = 0.2, chrX_sex = NULL)
reference(x, ...)
default_chrX_sex(x, ...)
}
\arguments{
\item{path}{Path to a Python \code{.npz} LD distribution root prepared for R loading.}
\item{shards}{Optional character vector of canonical shard labels to load.}
\item{default_chrX_sex}{Default chrX LD shard selector, one of \code{"female"}, \code{"male"}, or \code{"combined"}.}
\item{retain_ld_r}{Logical; retain raw signed LD \code{r} matrices in loaded shards. Operations keep working when this is \code{FALSE}.}
\item{check_payload_structure}{Logical; perform expensive sparse payload checks during distribution validation.}
\item{npz_root}{Path to a Python \code{.npz} LD distribution root.}
\item{extract_npz}{Logical; when \code{TRUE}, create or refresh sibling
\code{*.npz.d} directories containing extracted \code{.npy} arrays for faster
R loading.}
\item{x}{An LD panel or shard object.}
\item{ld_panel}{An \code{LDPanel}.}
\item{M}{A numeric vector or matrix with rows aligned to the LD panel reference.}
\item{logpvec}{Numeric vector of signed or unsigned log-p values aligned to the LD panel reference.}
\item{r2_threshold}{Finite non-negative pruning threshold; defaults to \code{0.2}.}
\item{chrX_sex}{Optional chrX LD shard selector for operations. Ignored when chrX is absent.}
\item{...}{Reserved for S3 method dispatch.}
}
\description{
Prepare Python LD \code{.npz} distributions for R loading, validate their
manifests and sparse CSC payloads, load LD panels, and run the core LD
operations.
}
\details{
R reads Python LD \code{.npz} shard files directly. R-loadable distributions use
\code{ld_manifest.json}, bundled reference BIM files, Python \code{.npz} LD
shards, and an R reference cache sidecar recorded in manifest fields
\code{r_reference_cache} and \code{r_reference_cache_md5}. \code{load_ld()}
constructs \code{Matrix::dgCMatrix} objects in memory and derives the
elementwise squared LD matrix used by \code{multiply_r2()} and
\code{fast_prune()}.

\code{prepare_ld_npz_for_r()} builds or refreshes the R reference cache sidecar
from the bundled BIM files for all shards described in the existing Python
manifest and updates that manifest in place. With \code{extract_npz = TRUE}, it
also creates or refreshes extracted \code{*.npz.d} cache directories next to the
manifest-declared shard files. The operation is idempotent.
}
\value{
\code{load_ld()} returns an \code{LDPanel}. \code{load_ld_reference()} returns
the paired \code{ReferencePanel}. \code{validate_ld_distribution()} returns a
list with \code{ok = TRUE} on success. \code{prepare_ld_npz_for_r()} returns the
updated manifest invisibly.
\code{a1freq()} returns aligned allele frequencies, \code{multiply_r2()} returns
a vector or matrix with the same shape as \code{M}, and \code{fast_prune()}
returns a numeric vector or one-dimensional matrix with pruned entries set to
\code{NaN}.
}
\examples{
ld_root <- system.file("extdata", "ld", package = "statgen")
ld <- load_ld(ld_root)

num_snp(ld)
head(a1freq(ld))

scores <- seq_len(num_snp(ld))
multiply_r2(ld, scores)

logp <- rev(seq_len(num_snp(ld)))
fast_prune(logp, ld, r2_threshold = 0.35)
}
\keyword{datasets}

\name{annotations}
\alias{load_annotations}
\alias{create_annotations}
\alias{create_annotation}
\alias{save_annotations_cache}
\alias{load_annotations_cache}
\alias{annomat}
\alias{annonames}
\alias{num_annot}
\alias{select_annotations}
\alias{union_annotations}
\title{AnnotationPanel objects}
\description{
Paint BED interval annotations onto a reference panel, inspect sparse
annotation matrices, and save or load R-native annotation caches.
}
\usage{
load_annotations(bed_paths, reference)
create_annotations(reference, annotation_matrix, annotation_names)
create_annotation(reference, annovec, annoname)
save_annotations_cache(panel, path)
load_annotations_cache(path, shards = NULL)

annomat(x, ...)
annonames(x, ...)
num_annot(x, ...)
select_annotations(x, names, ...)
union_annotations(x, other, mode = "by_name", ...)
}
\arguments{
\item{bed_paths}{Character path or character vector of BED files.}
\item{reference}{A \code{ReferencePanel}.}
\item{annotation_matrix}{A binary matrix with SNPs as rows and annotations as columns.}
\item{annotation_names, names}{Character annotation names.}
\item{annovec}{Binary annotation vector aligned to \code{reference}.}
\item{annoname}{Character scalar annotation name.}
\item{panel, x, other}{Annotation panel objects.}
\item{path}{Character scalar cache path.}
\item{shards}{Optional character vector of shard labels in canonical order.}
\item{mode}{Union mode. Currently only \code{"by_name"} is supported.}
\item{...}{Reserved for S3 methods.}
}
\value{
\code{load_annotations()}, \code{create_annotations()},
\code{create_annotation()}, and \code{load_annotations_cache()} return
\code{AnnotationPanel} S3 objects. \code{annomat(x)} returns a sparse
\code{Matrix::lgCMatrix} with column names equal to \code{annonames(x)}.
}
\details{
BED intervals are interpreted as 0-based half-open intervals and are matched to
reference chromosome labels exactly. The annotation matrix is binary and
reference-aligned.
}
\examples{
ref_path <- system.file("extdata", "reference_chr1.bim", package = "statgen")
bed_path <- system.file("extdata", "anno1.bed", package = "statgen")
ref <- load_reference(ref_path)
ann <- load_annotations(bed_path, ref)
dim(annomat(ann))

cache <- tempfile(fileext = ".rds")
save_annotations_cache(ann, cache)
ann_cached <- load_annotations_cache(cache)
dim(annomat(ann_cached))
}

\name{sumstats}
\alias{load_sumstats}
\alias{create_sumstats}
\alias{save_sumstats_cache}
\alias{load_sumstats_cache}
\alias{logpvec}
\alias{zvec}
\alias{nvec}
\alias{is_present}
\alias{beta_vec}
\alias{se_vec}
\alias{eaf_vec}
\alias{info_vec}
\title{Sumstats objects}
\description{
Load, create, inspect, and cache reference-aligned GWAS summary statistics.
}
\usage{
load_sumstats(path, reference)
create_sumstats(reference, p, z = NULL, n = NULL,
  beta = NULL, se = NULL, eaf = NULL, info = NULL)
save_sumstats_cache(sumstats, path)
load_sumstats_cache(path, shards = NULL)

logpvec(x, ...)
zvec(x, ...)
nvec(x, ...)
is_present(x, ...)
beta_vec(x, ...)
se_vec(x, ...)
eaf_vec(x, ...)
info_vec(x, ...)
}
\arguments{
\item{path}{Character scalar path to a summary-statistics TSV or RDS cache.}
\item{reference}{A \code{ReferencePanel}.}
\item{sumstats, x}{A \code{Sumstats} object.}
\item{p, z, n, beta, se, eaf, info}{Aligned
numeric vectors. Optional vectors may be \code{NULL}.}
\item{shards}{Optional character vector of shard labels in canonical order.}
\item{...}{Reserved for S3 methods.}
}
\value{
\code{load_sumstats()}, \code{create_sumstats()}, and
\code{load_sumstats_cache()} return \code{Sumstats} S3 objects. Accessors
return ordinary R vectors, with \code{NULL} for absent optional fields.
}
\details{
Rows are projected onto the supplied reference by exact chromosome, position,
and allele-hash matching. Missing reference variants are represented by
\code{NaN} in \code{logpvec(x)} and by \code{is_present(x) == FALSE}.
Gzipped TSV input requires the suggested \pkg{R.utils} package so
\code{data.table::fread()} can read compressed files portably.
}
\examples{
ref_path <- system.file("extdata", "reference_chr1.bim", package = "statgen")
sum_path <- system.file("extdata", "traits_complete.tsv.gz", package = "statgen")
ref <- load_reference(ref_path)
s <- load_sumstats(sum_path, ref)
head(logpvec(s))

cache <- tempfile(fileext = ".rds")
save_sumstats_cache(s, cache)
s_cached <- load_sumstats_cache(cache)
head(logpvec(s_cached))
}
